#ifndef __HashTable_h
#define __HashTable_h

#include <assert.h>
#include <memory.h>
#include <stdlib.h>

#define NULL 0

/*
There must be the following functions
class F must have the following static functions

unsigned int GetHashValue1(K);
unsigned int GetHashValue2(K);
K GetHashKey(V);
V GetUnusedValue1();
V GetUnusedValue2();
*/

class CHTVirtualAlloc;

#ifdef WIN32
#include <windows.h>

class CHTVirtualAlloc
{
public:
	static void *Allocate(unsigned int numBytes)
	{
		return VirtualAlloc(NULL, numBytes, MEM_COMMIT, PAGE_READWRITE);
	}
	static void Release(void *toRelease)
	{
		VirtualFree(toRelease, 0, MEM_RELEASE);
	}

	static unsigned int GetMinimumSize()
	{
		SYSTEM_INFO si;
		GetSystemInfo(&si);

		return si.dwPageSize;
	}
};
#endif

class CHTNew
{
public:
	static void *Allocate(unsigned int numBytes)
	{
		return new char[numBytes];
	}
	static void Release(void *toRelease)
	{
		delete[] ((char*) toRelease);
	}

	static unsigned int GetMinimumSize()
	{
		return 32;
	}
};


template <class T> class CHFIntToPointer
{
public:
	static unsigned int GetHashValue1(int id){return id;};
	static unsigned int GetHashValue2(int id){return id ^ 0xaaaaaaaa;};
	static T* GetUnusedValue1(){return (T*) 0xffffffff;};
	static T* GetUnusedValue2(){return (T*) 0xfffffffe;};
};

template <class K, class V, class F, bool AutoInit, class A = CHTVirtualAlloc>
class CHashTable
{
public:
	CHashTable();
	~CHashTable();

	void Init();
	void UnInit();

	//If the key is already there, returns 1.
	int Add(V value);

	//If the key was there, returns 1.
	int Remove(K key);

	//If the key is there, returns 1.
	int Get(K key, V* value);

	int Enumerate(int(f)(V value, void *handle), void *handle);

	void RemoveAll();

	unsigned int GetNumElements();

	void BeginGetAll();
	int GetNext(V* value);
	void EndGetAll();

	//this interface allows multiple iterators
	void* BeginIteration();
	int GetNextValue(void **pos, V* value);
	void EndIteration(void *pos);

private:
	unsigned int MinNumElements;
	V* FindHashElement(K, V**);
	void ReAllocate();

	V* Elements;
	unsigned int NumElements;
	unsigned int NumUsedElements;
	unsigned int NumDeletedElements;

	//iteration variables
	void* Iter;
	unsigned int NumOutstandingIterators;
};

template <class K, class V, class F, bool AutoInit, class A>
unsigned int
CHashTable<K, V, F, AutoInit, A>::GetNumElements()
{
	return NumUsedElements;
}

template <class K, class V, class F, bool AutoInit, class A>
CHashTable<K, V, F, AutoInit, A>::CHashTable()
{
	if (AutoInit)
		Init();
}

template <class K, class V, class F, bool AutoInit, class A>
void
CHashTable<K, V, F, AutoInit, A>::Init()
{
	Elements = NULL;
	NumElements = 0;
	NumUsedElements = 0;
	Iter = NULL;
	NumOutstandingIterators = 0;
	MinNumElements = __min(A::GetMinimumSize() / sizeof(V), 8);;
}

template <class K, class V, class F, bool AutoInit, class A>
void
CHashTable<K, V, F, AutoInit, A>::RemoveAll()
{
	assert(NumOutstandingIterators == 0);
	UnInit();
	Init();
}

template <class K, class V, class F, bool AutoInit, class A>
int 
CHashTable<K, V, F, AutoInit, A>::Enumerate(int(f)(V value, void *handle), void *handle)
{
	unsigned int i;
	for (i=0; i<NumElements; i++)
	{
		if (Elements[i] != F::GetUnusedValue1())
		{
			if (Elements[i] != F::GetUnusedValue2())
			{
				int shouldContinue = f(Elements[i], handle);
				if (shouldContinue == 0)
					return 0;

			}
		}
	}
	return 1;
}

template <class K, class V, class F, bool AutoInit, class A>
void
CHashTable<K, V, F, AutoInit, A>::UnInit()
{
	assert(NumOutstandingIterators == 0);
		
	if (NumElements != 0)
		A::Release(Elements);
	NumElements = 0;
	NumUsedElements = 0;
	Elements = NULL;
}


template <class K, class V, class F, bool AutoInit, class A>
CHashTable<K, V, F, AutoInit, A>::~CHashTable()
{	
	if (AutoInit)
		UnInit();
}

template <class K, class V, class F, bool AutoInit, class A>
V*
CHashTable<K, V, F, AutoInit, A>::FindHashElement(K key, V** firstDeletedElementOut)
{
	unsigned int mask = NumElements - 1;
	unsigned int hash = F::GetHashValue1(key) & mask;
	unsigned int hashI = (F::GetHashValue2(key) * 2 + 1);

	V* firstDeletedElement = NULL;
	
	while (1)
	{
		V v = Elements[hash];

		//Check if we found an empty element
		if (v == F::GetUnusedValue1())
		{
			*firstDeletedElementOut = firstDeletedElement;
			return Elements + hash;
		}

		//Make sure we are not looking at a deleted element
		if (v != F::GetUnusedValue2())
		{
			if (F::GetHashKey(v) == key)
			{
				*firstDeletedElementOut = firstDeletedElement;
				return Elements + hash;
			}
		}
		else
		{
			if (firstDeletedElement == NULL)
				firstDeletedElement = Elements + hash;
		}


		hash = (hash + hashI) & mask;
	}
}

template <class K, class V, class F, bool AutoInit, class A>
int
CHashTable<K, V, F, AutoInit, A>::Add(V val)
{
	assert(NumOutstandingIterators == 0);
	
	ReAllocate();
	K key = F::GetHashKey(val);
	V* firstDeletedElement;
	V* hashEl = FindHashElement(key, &firstDeletedElement);

	//Check if the space was free
	if (*hashEl == F::GetUnusedValue1())
	{
		NumUsedElements++;
		if (firstDeletedElement == NULL)
		{
			*hashEl = val;
		}
		else
		{
			NumDeletedElements--;
			*firstDeletedElement = val;
		}

		ReAllocate();

		return 0;
	}

	if (firstDeletedElement == NULL)
	{
		*hashEl = val;
	}
	else
	{
		*firstDeletedElement = val;
		*hashEl = F::GetUnusedValue2();
	}

	return 1;
}


template <class K, class V, class F, bool AutoInit, class A>
int
CHashTable<K, V, F, AutoInit, A>::Remove(K key)
{
	assert(NumOutstandingIterators == 0);

	V* firstDeletedElement;
	V* hashEl = FindHashElement(key, &firstDeletedElement);

	//Check if the space was free
	if (*hashEl == F::GetUnusedValue1())
	{
		return 0;
	}
	
	
	*hashEl = F::GetUnusedValue2();
	NumDeletedElements++;
	NumUsedElements--;

	ReAllocate();
	return 1;
}

template <class K, class V, class F, bool AutoInit, class A>
int
CHashTable<K, V, F, AutoInit, A>::Get(K key, V* value)
{
	if (NumElements == 0)
		return 0;

	V* firstDeleted;
	V* hashEl = FindHashElement(key, &firstDeleted);
	if (*hashEl == F::GetUnusedValue1())
	{
		return 0;
	}

	*value = *hashEl;

	if (firstDeleted != NULL)
	{
		*firstDeleted = *hashEl;
		*hashEl = F::GetUnusedValue2();
	}
		
	return 1;
}

template <class K, class V, class F, bool AutoInit, class A>
void
CHashTable<K, V, F, AutoInit, A>::ReAllocate()
{
	assert(NumOutstandingIterators == 0);

	unsigned int newNumElements = 0;
	int needToReAllocate;
	if (NumElements == 0)
	{
		newNumElements = MinNumElements;
		needToReAllocate = 1;
	}
	else
	{
		needToReAllocate = 0;
		//If there is less that 50% free space, reallocate
		if (((NumUsedElements + NumDeletedElements) * 2) > NumElements)
		{
			needToReAllocate = 1;
			newNumElements = MinNumElements;
			while (newNumElements < (NumUsedElements * 2))
				newNumElements += newNumElements;
		}
	}

	if (needToReAllocate)
	{
		unsigned int oldNumElements = NumElements;
		V* oldElements = Elements;

		NumUsedElements = 0;
		NumDeletedElements = 0;
		NumElements = newNumElements;
		Elements = (V*) A::Allocate(sizeof(V) * NumElements);
		unsigned int i;
		for (i=0; i<NumElements; i++)
		{
			Elements[i] = F::GetUnusedValue1();
		}


		for (i=0; i<oldNumElements; i++)
		{
			if (oldElements[i] != F::GetUnusedValue1())
				if (oldElements[i] != F::GetUnusedValue2())
					Add(oldElements[i]);
		}

		if (oldNumElements != 0)
			A::Release(oldElements);
	}
}

template <class K, class V, class F, bool AutoInit, class A>
void
CHashTable<K, V, F, AutoInit, A>::BeginGetAll()
{
	assert(Iter == NULL);
	Iter = BeginIteration();
}

template <class K, class V, class F, bool AutoInit, class A>
int
CHashTable<K, V, F, AutoInit, A>::GetNext(V* value)
{
	return GetNextValue(&Iter, value);
}

template <class K, class V, class F, bool AutoInit, class A>
void
CHashTable<K, V, F, AutoInit, A>::EndGetAll()
{
	EndIteration(Iter);
	Iter = NULL;
}

template <class K, class V, class F, bool AutoInit, class A>
void* 
CHashTable<K, V, F, AutoInit, A>::BeginIteration()
{
	NumOutstandingIterators ++;
	return (void*) Elements;
}

template <class K, class V, class F, bool AutoInit, class A>
int 
CHashTable<K, V, F, AutoInit, A>::GetNextValue(void **pos, V* value)
{
	V** current = (V**) pos;
	if ((*current) == NULL)
	{
		return 0;
	}

	while (1)
	{
		if ((*current) == (Elements + NumElements))
		{
			(*current) = NULL;
			return 0;
		}

		if ((**current) != F::GetUnusedValue1())
		{
			if ((**current) != F::GetUnusedValue2())
			{
				*value = (**current);
				(*current)++;
				return 1;
			}
		}

		(*current)++;
	}
}

template <class K, class V, class F, bool AutoInit, class A>
void 
CHashTable<K, V, F, AutoInit, A>::EndIteration(void *pos)
{
	NumOutstandingIterators --;
}

/*
class CIntHash
{
public:
	static unsigned int GetHashValue1(int x){return x;};
	static unsigned int GetHashValue2(int x){return x ^ 0xaaaaaaaa;};
	static int GetHashKey(int x){return x*2;};
	static int GetUnusedValue1(){return 0xffffffff;};
	static int GetUnusedValue2(){return 0xfffffffe;};
};


int
test(int numArgs, char**args, char**envp) 
{ 
	//Key is twice value

	CHashTable<int, int, CIntHash> hash;
	int val;
	for (val=0; val<100000; val++)
	{
		hash.Add(val);

		int key = val*2;
		int getVal;
		assert(hash.Get(key, &getVal));
		assert(getVal == val);

		if (val & 1)
		{
			assert(hash.Remove(key));
		}

		assert(hash.Get(key+1, NULL) == 0);
	}

	for (val = 0; val<100000; val++)
	{
		int key = val*2;
		if (val & 1)
		{
			assert(hash.Get(key, NULL) == 0);
		}
		else
		{
			int getVal;
			assert(hash.Get(key, &getVal));
			assert(getVal == val);
		}
	}

	for (val = 0; val<100000; val++)
	{
		int key = val*2;

		if ((val & 1) == 0)
			hash.Remove(key);

		assert(hash.Get(key, NULL) == 0);
		assert(hash.Get(key+1, NULL) == 0);
	}
}
	
*/		





#endif