#ifndef __ResizeableArray_h
#define __ResizeableArray_h

#include <assert.h>
#include <stdlib.h>
#include <new.h>


#define NULL 0



template <class T>
class CResizeableArray
{
public:
	CResizeableArray();
	CResizeableArray(const CResizeableArray<T>& arr);
	~CResizeableArray();

	void SetShouldAutomaticallyShrink(bool shouldShrink);

	void SetSize(unsigned int size);
	void SetExactSize(unsigned int size);

	void Set(unsigned int numValues, const T* values);

	//These references are only valid until on of the following functions is called:
	//SetSize, SetExactSize, ~CResizableArray, AddElement, RemoveElement
	T* GetElementRef(unsigned int index);
	T* GetLastElementRef();
	const T* GetConstElementRef(unsigned int index) const;
	const T* GetConstLastElementRef() const;
	void RemoveElement(unsigned int index);
	void RemoveLastElement();
	T* AddElement();
	
	void Prepend(const CResizeableArray<T>& rhs);
	void Concat(const CResizeableArray<T>& rhs) { Append(rhs); };
	void Append(const CResizeableArray<T>& rhs);
	
	void Append(unsigned int arrSize, const T* arr);

	unsigned int GetNumElements() const;

	//These functions only work correctly if T can be sucessfully passed by value
	T GetElement(unsigned int index) const;
	T GetLastElement() const;
	void SetElement(unsigned int index, const T& value);
	void AddElement(T value);

	//Operators

	CResizeableArray<T>& operator=(const CResizeableArray<T>& rhs);
	bool operator==(const CResizeableArray<T>& rhs) const;
	bool operator!=(const CResizeableArray<T>& rhs) const;
	const T& operator[](unsigned int) const;
	T& operator[](unsigned int);
	
private:
	void ResizeArray(unsigned int newSize);
	
	unsigned int NumAllocated;
	unsigned int NumElements;
	T* Elements;

	bool ShouldAutomaticallyShrink;
};

/*
Invariants:
	if (NumElements == 0)
		Elements = NULL
	else
		Elements points to an array of size NumAllocated.

	NumElements <= NumAllocated

	If (Elements != NULL)
		We have exactly one block of memory allocated from the global heap

Requirements:
	T can be copied by memcpy

Abstraction Function
	NumElements is the number of elements in the array.
	Elements points to the first element
*/

template <class T>
CResizeableArray<T>::CResizeableArray()
{	
	//Create an array with no elements
	NumElements = 0;
	NumAllocated = 0;
	Elements = NULL;
	ShouldAutomaticallyShrink = true;
}

template <class T>
CResizeableArray<T>::CResizeableArray(const CResizeableArray<T>& arr)
{
	//Create an array from another array
	ShouldAutomaticallyShrink = arr.ShouldAutomaticallyShrink;

	NumAllocated = arr.NumElements;
	NumElements = arr.NumElements;

	Elements = (T*) new char[sizeof(T)*NumElements];

	unsigned int i;
	for (i=0; i<NumElements; i++)
		Elements[i] = arr[i];
}

template <class T>
CResizeableArray<T>::~CResizeableArray()
{
	unsigned int i;
	for (i=0; i<NumElements; i++)
		Elements[i].T::~T();

	if (Elements != NULL)
		delete[] ((char*) Elements);

//Invariant is broken on deconstuction
}


template <class T>
void 
CResizeableArray<T>::SetShouldAutomaticallyShrink(bool shouldShrink)
{
	ShouldAutomaticallyShrink = shouldShrink;
}

template<class T>
void
CResizeableArray<T>::SetSize(unsigned int newSize)
{
//This sets the size of the array to newSize.
//If the array grows, elements with indices greater or equal to the previous size
//	are garbage.
//If it shrinks, elements with indices less than the newSize are deleted.


	//This satisfies the conditions of the ResizeArray function.  
	//The value of NumElements is reset later


	//Destruct any elements that will no longer be valid;
	unsigned int i;
	for (i=newSize; i<NumElements; i++)
		Elements[i].T::~T();


	NumElements = __min(NumElements, newSize);

	if (NumAllocated < newSize)
	{
		unsigned int sizeToAllocate = NumAllocated * 2;
		sizeToAllocate = __max(sizeToAllocate, 16);
		sizeToAllocate = __max(sizeToAllocate, newSize);

		ResizeArray(sizeToAllocate);
	} 
	else if (ShouldAutomaticallyShrink && (NumAllocated > newSize * 2))
	{
		//shrink our array if we have too many allocated objects and we're supposed
		//to automatically shrink.
		ResizeArray(newSize);
	}

	//Construct any new elements
	for (i=NumElements; i<newSize; i++)
		new (&Elements[i]) T;
		
	NumElements = newSize;
}


template<class T>
void
CResizeableArray<T>::RemoveElement(unsigned int toRemove)
{
	//This removes elements from the array and slides all elements with indices greater 
	//than i one index lower
	
	assert(toRemove < NumElements);

	unsigned int i;
	for (i=toRemove; (i+1) < NumElements; i++)
	if (i < (NumElements - 1))
		Elements[i] = Elements[i+1];

	NumElements--;
	Elements[NumElements].T::~T();

	//shrink our array if we have too many allocated objects and we're supposed
	//to automatically shrink.
	if (ShouldAutomaticallyShrink && (NumAllocated > NumElements * 2))
		ResizeArray(__max(NumElements, 16));
}

template<class T>
void
CResizeableArray<T>::RemoveLastElement()
{
	assert(NumElements > 0);
	RemoveElement(NumElements-1);
}
template<class T>
void
CResizeableArray<T>::SetExactSize(unsigned int newSize)
{
	//Destruct any elements that will no longer be valid;
	unsigned int i;
	for (i=newSize; i<NumElements; i++)
		Elements[i].T::~T();

	NumElements = __min(newSize, NumElements);
	ResizeArray(newSize);

	//Construct any new elements
	for (i=NumElements; i<newSize; i++)
		new (&Elements[i]) T;

	NumElements = newSize;
}


template<class T>
void
CResizeableArray<T>::ResizeArray(unsigned int newSize)
{
/*
This changes the amount of memory allocated to newSize without
affecting the abstract state of the array

newSize must be greater than or equal to NumElements
*/
	assert(newSize >= NumElements);

	if (newSize == NumAllocated)
		return;

	NumAllocated = newSize;
	if (newSize == 0)
	{
		delete[] ((char*)Elements);
		Elements = NULL;
		return;
	}
		
	
	T* newElements = (T*) new char[sizeof(T)*newSize];
	if (Elements != NULL)
	{
		unsigned int i;
		for (i=0; i<__min(NumElements, newSize); i++)
		{
			new (&newElements[i]) T(Elements[i]);
			Elements[i].T::~T();
		}

		delete[] ((char*) Elements);
	}
	Elements = newElements;
}

template<class T>
unsigned int
CResizeableArray<T>::GetNumElements() const
{
	return NumElements;
}

template<class T>
T*
CResizeableArray<T>::GetElementRef(unsigned int i)
{
	assert(i < NumElements);
	return Elements + i;
}

template<class T>
T*
CResizeableArray<T>::GetLastElementRef()
{
	//Requires that there is at least one element in the array
	assert(NumElements != 0);
	return Elements + NumElements - 1;
}

template<class T>
const T*
CResizeableArray<T>::GetConstElementRef(unsigned int i) const
{
	assert(i < NumElements);
	return Elements + i;
}

template<class T>
const T*
CResizeableArray<T>::GetConstLastElementRef() const
{
	//Requires that there is at least one element in the array
	assert(NumElements != 0);
	return Elements + NumElements - 1;
}

template<class T>
T*
CResizeableArray<T>::AddElement()
{
	SetSize(GetNumElements()+1);
	return GetLastElementRef();
}

/******************************************************************8
These functions only work if T can be passed by value
*/

template<class T>
void 
CResizeableArray<T>::SetElement(unsigned int index, const T& value)
{
	assert(index < NumElements);

	Elements[index] = value;
}

template<class T>
T
CResizeableArray<T>::GetLastElement() const
{
	//Requires: There is at least one element in the array.
	assert(NumElements != 0);
	return Elements[NumElements-1];
}

template<class T>
T
CResizeableArray<T>::GetElement(unsigned int i) const
{
	assert(i < NumElements);
	return Elements[i];
}

template<class T>
void 
CResizeableArray<T>::AddElement(T value)
{
	SetSize(GetNumElements()+1);
	SetElement(GetNumElements()-1, value);
}

template<class T>
void 
CResizeableArray<T>::Append(const CResizeableArray<T>& rhs)
{
	if (&rhs == this)
	{
		unsigned int originalSize = NumElements;
		SetSize(originalSize * 2);
		unsigned int i;
		for (i=0; i<originalSize; i++)
			Elements[i+originalSize] = Elements[i];
	}
	else
	{
		unsigned int thisSize = GetNumElements();
		SetSize(thisSize + rhs.GetNumElements());
		unsigned int i;

		for (i=0; i<rhs.GetNumElements(); i++)
			Elements[thisSize + i] = rhs.Elements[i];
	}
}
	
template<class T>
void
CResizeableArray<T>::Prepend(const CResizeableArray<T>& rhs)
{
	if (&rhs == this)
	{
		Append(rhs);
	}
	else
	{
		unsigned int thisSize = GetNumElements();
		SetSize(thisSize + rhs.GetNumElements());

		unsigned int i;
		for (i=0; i<thisSize; i++)
		{
			unsigned int backI = thisSize - 1 - i;
			Elements[rhs.GetNumElements()+backI] = Elements[backI];
		}

		for (i=0; i<rhs.GetNumElements(); i++)
			Elements[i] = rhs.Elements[i];
	}
}

template<class T>
void 
CResizeableArray<T>::Set(unsigned int numElements, const T* elements)
{
	SetSize(numElements);
	unsigned int i;
	for (i=0; i<numElements; i++)
		Elements[i] = elements[i];
}

template<class T>
void 
CResizeableArray<T>::Append(unsigned int numNewElements, const T* newElements)
{
	if (numNewElements == 0)
		return;

	unsigned int prevNumElements = GetNumElements();
	SetSize(prevNumElements + numNewElements);

	T* elements = Elements + prevNumElements;

	unsigned int i;
	for (i=0; i<numNewElements; i++)
		elements[i] = newElements[i];
}

template<class T>
CResizeableArray<T>& 
CResizeableArray<T>::operator=(const CResizeableArray<T>& rhs)
{
	SetSize(rhs.GetNumElements());
	unsigned int i;
	for (i=0; i<rhs.GetNumElements(); i++)
		Elements[i] = rhs[i];

	return *this;
}

template<class T>
bool 
CResizeableArray<T>::operator==(const CResizeableArray<T>& rhs) const
{
	if( GetNumElements() != rhs.GetNumElements())
		return false;

	unsigned int i;
	for (i=0; i<NumElements; i++)
		if (Elements[i] != rhs.Elements[i])
			return false;
	
	return true;
}

template<class T>
bool 
CResizeableArray<T>::operator!=(const CResizeableArray<T>& rhs) const
{
	return !operator==(rhs);
}

template<class T>
const T& 
CResizeableArray<T>::operator [](unsigned int index) const
{
	assert(index < NumElements);
	return Elements[index];
}

template<class T>
T& 
CResizeableArray<T>::operator [](unsigned int index)
{
	assert(index < NumElements);
	return Elements[index];
}

#endif