#ifndef __VVHashTable_h
#define __VVHashTable_h

#include "assert.h"

/*
class F must have the following static functions:

	unsigned int GetHash1ForValue(const V* value);	
	unsigned int GetHash2ForValue(const V* value); 
	bool GetDoKeysMatch(const V*, const V*);
	void *AllocateAndZeroMemory(unsigned int numBytes);
	void ReleaseMemory(void *toRelease);
	enum {MinNumElements = power of 2};

The following should be true of class F, where x and y are V*'s

	if GetDoKeysMatch(x, y)
	{
		GetHash1ForValue(x) = GetHash1ForValue(y);
		GetHash2ForValue(x) = GetHasy2ForValue(y);
	}
*/

class CVVHTVirtualAlloc
{
public:
	static void *AllocateAndZeroMemory(unsigned int numBytes)
	{
		return VirtualAlloc(NULL, numBytes, MEM_COMMIT, PAGE_READWRITE);
	};

	static void ReleaseMemory(void *toRelease)
	{
		VirtualFree(toRelease, 0, MEM_RELEASE);
	}

	enum {MinNumElements = 1024};
};

class CVVHTNewDelete
{
public:
	static void *AllocateAndZeroMemory(unsigned int numBytes)
	{
		void *ret = new char[numBytes];
		memset(ret, 0, numBytes);
		return ret;
	};

	static void ReleaseMemory(void *toRelease)
	{
		delete[] toRelease;
	}

	enum {MinNumElements = 64};
};


template <class V, class F>
class CVVHashTable
{
/*
VVHashTable is a hash table that uses referenced values as both the Keys and the Values.

Its abstract state is a set of pointers to valid values such that no F::GetDoKeysMatch
returns false all elements in the set.
*/
public:
	CVVHashTable();
	/*
	Initializes the hash table to contain the empty set of pointers
	It remains initialized until it is deleted.
	*/

	~CVVHashTable();
	
	void RemoveAll();
	/*
	Sets the set of pointers to the empty set.

	There must be no active iteration when this function is called.
	*/
	
	V* Add(V* value);
	/*
	Adds a value to the set.  If there is a value already in the set that
	matches the value, that value is removed from the set and returned.  Otherwise,
	NULL is returned.

	The objects pointed to by all pointers that are passed into this function
	must remain valid, until those pointers are removed from the set.
	The values that are tested in the GetDoKeyMatch function must not be changed while
	the values are in the set.  Consequently, the hash values must remain unchanged.

	There must be no active iteration when this function is called.
	*/

	bool Remove(const V* value);
	/*
	If there is a pointer in the set that matches value, that pointer is removed from
	the set and true is returned.  Otherwise, false is returned.

	There must be no active iteration when this function is called.
	*/

	V* GetMatchingValue(const V* valueToMatch);
	/*
	If there is a pointer in the set that matches value to match, that pointer is returned.
	Otherwise, NULL is returned.
	*/

	unsigned int GetNumValues();
	/*
	Returns that number of elements in the set of pointers.
	*/


	void BeginIterationValues();
	/*
	Creates an active iteration for the set of pointers.  Initially, all objects that are in
	the set are in the uniterated set.

	There must be no active iteration when this function is called.
	*/

	V* IterateValue();
	/*
	If there are elements in the uniterated set, removes one from the uniterated set and returns it.
	Otherwise, returns NULL

	There must be an active iteration when this function is called.
	*/

	void EndIterateValues();
	/*
	Destroys the active iteration for this object.  

  	There must be an active iteration when this function is called.
	*/

	//this interface allows multiple iterators
	void* BeginIteration();
	V* GetNextValue(void **pos);
	void EndIteration(void *pos);

	void Init();
	void UnInit();

private:
	V** FindHashElement(const V*, V*** firstDeleted);
	void ReAllocate();

	//iteration variables
	void* Iter;
	unsigned int NumOutstandingIterators;

	V** Elements;
	unsigned int NumElements;
	unsigned int NumUsedElements;
	unsigned int NumDeletedElements;
};

/*
Invariants:
If NumElements == 0, Elements is NULL;
	Else Elements points to a block of memory that will hold NumElements type V*'s

There are NumDeletedElements values in Elements that are equal to 0xffffffff
 
There are NumUsedElements values in Elements that point to valid values.

There are NumElements - NumDeletedElements - NumUsedElements that are NULL.

There are no (non 0 or 0xffffffff) values in Elements that match.

NumElements is either zero or a power of two.
  
The probe seqence for a value is p[n] = (hash1 + (hash2*2+1) * n) mod NumElements;
The shortest cycle in this sequence is of length NumElements because (hash2*2+1) is odd
and NumElements is a power of two, so they are relatively prime.

If a value is stored at Elements[p[n]], then for all m less than n, Elements[p[m]] != NULL

There is always at least one NULL element in Elements

Abstraction Function:
The set of pointers that this set contains are all values in Elements[0..NumElements] that 
are neither NULL, or 0xffffffff
*/



template <class V, class F>
CVVHashTable<V, F>::CVVHashTable()
{
	Init();
}

template <class V, class F>
CVVHashTable<V, F>::~CVVHashTable()
{
	UnInit();
}

template <class V, class F>
void
CVVHashTable<V, F>::Init()
{
	NumElements = 0;
	NumUsedElements = 0;
	Elements = NULL;
	Iter = NULL;
	NumOutstandingIterators = 0;
}

template <class V, class F>
void
CVVHashTable<V, F>::UnInit()
{
	assert(NumOutstandingIterators == 0);

	if (NumElements != 0)
		F::ReleaseMemory(Elements);
	NumElements = 0;
	NumUsedElements = 0;
	Elements = NULL;
	Iter = NULL;
}

template <class V, class F>
void
CVVHashTable<V, F>::RemoveAll()
{
	assert(NumOutstandingIterators == 0);
	UnInit();
}

template <class V, class F>
V**
CVVHashTable<V, F>::FindHashElement(const V* value, V*** firstDeletedOut)
{
	assert(NumElements > 0);
	assert(Elements != NULL);

	unsigned int mask = NumElements - 1;
	unsigned int hash = (F::GetHash1ForValue(value) * 15401) & mask;
	unsigned int hashI = (F::GetHash2ForValue(value) * 2 + 1);

	V** firstDeleted = NULL;

	//hash = value.p[0]

	//this loop will always return b/c the probe sequence will hit every element,
	//and there is always at least one NULL element.
	while (1)
	{
		V* v = Elements[hash];

		if (v == NULL)
		{
			*firstDeletedOut = firstDeleted;
			return Elements + hash;
		}

		if (v != ((V*)0xffffffff))
		{
			if (F::GetDoKeysMatch(value, Elements[hash]))
			{
				*firstDeletedOut = firstDeleted;
				return Elements + hash;
			}
		}
		else
		{
			if (firstDeleted == NULL)
				firstDeleted = Elements + hash;
		}

		//Compute the next index in the probe sequence
		hash = (hash + hashI) & mask;
	}
}

template <class V, class F>
unsigned int 
CVVHashTable<V, F>::GetNumValues()
{
	return NumUsedElements;
}

template <class V, class F>
V*
CVVHashTable<V, F>::Add(V* val)
{
	assert(NumOutstandingIterators == 0);

	//Make sure the NumElements != 0
	ReAllocate();
	
	V** firstDeleted;

	V** hashEl = FindHashElement(val, &firstDeleted);

	V* ret = *hashEl;

	if (firstDeleted == NULL)
	{
		//We are storing over another element or a blank
		//If we are storing over a blank, we must count it
		//Otherwise, we do not need to update any counts

		*hashEl = val;
		if (ret == NULL)
		{
			NumUsedElements++;
			ReAllocate();
		}
	}
	else
	{
		//We can store over a deleted element
		if (ret == NULL)
		{
			*firstDeleted = val;
			NumUsedElements++;
			NumDeletedElements--;
			ReAllocate();
		}
		else
		{
			*firstDeleted = val;
			*hashEl = (V*) 0xffffffff;
		}
	}



	return ret;
}


template <class V, class F>
bool
CVVHashTable<V, F>::Remove(const V* val)
{
	assert(NumOutstandingIterators == 0);

	if (NumElements == 0)
		return false;
	
	V** firstDeleted;
	V** hashEl = FindHashElement(val, &firstDeleted);

	//Check if the space was free, which means that nothing in the set matches val
	if (*hashEl == NULL)
		return false;
		
	*hashEl = (V*) 0xffffffff;
	NumDeletedElements++;
	NumUsedElements--;

	ReAllocate();
	return true;
}

template <class V, class F>
V*
CVVHashTable<V, F>::GetMatchingValue(const V* toMatch)
{
	if (NumElements == 0)
		return NULL;

	V** firstDeleted;
	V** hashEl = FindHashElement(toMatch, &firstDeleted);

	V* ret = *hashEl;

	if (ret != NULL)
	{
		if (firstDeleted != NULL)
		{
			*firstDeleted = ret;
			*hashEl = (V*) 0xffffffff;
		}
	}

	return ret;
}

template <class V, class F>
void
CVVHashTable<V, F>::ReAllocate()
{
	assert(NumOutstandingIterators == 0);

	unsigned int newNumElements = 0;
	bool needToReAllocate = false;
	if (NumElements == 0)
	{
		newNumElements = F::MinNumElements;
		needToReAllocate = true;
	}
	else
	{
		//If there is less than 25% free space or 
		//more than 75% free space and NumElements is greater than MinNumElements
		//then reallocate
		if ((((NumUsedElements + NumDeletedElements) * 4) > NumElements * 3) ||
			((NumUsedElements * 4 < NumElements) && (NumElements > F::MinNumElements)))
		{
			needToReAllocate = true;
			newNumElements = F::MinNumElements;
			while (newNumElements < (NumUsedElements * 2))
				newNumElements += newNumElements;
		}
	}

	if (needToReAllocate)
	{
		unsigned int oldNumElements = NumElements;
		V** oldElements = Elements;

		NumDeletedElements = 0;
		NumElements = newNumElements;
		Elements = (V**) F::AllocateAndZeroMemory(sizeof(V*) * NumElements);
		unsigned int i;
		for (i=0; i<oldNumElements; i++)
		{
			if (oldElements[i] != ((V*)0xffffffff))
			{
				if (oldElements[i] != 0x00000000)
				{
					//rehash the element
					V** firstDeleted;
					V** hashEl = FindHashElement(oldElements[i], &firstDeleted);
					if (*hashEl != NULL)
					{
#ifdef _DEBUG
						//we should never have two elements with matching keys,
						//but the following are debugging statements to help
						//with that erroneous situation.
						FindHashElement(oldElements[i], &firstDeleted);
						F::GetDoKeysMatch(*hashEl, oldElements[i]);
						assert(0);
#endif
					}
					//store the old element in the new hash table
					*hashEl = oldElements[i];
				}
			}
		}

		if (oldNumElements != 0)
			F::ReleaseMemory(oldElements);
	}
}

template <class V, class F>
void
CVVHashTable<V, F>::BeginIterationValues()
{
	assert(Iter == NULL);
	Iter = BeginIteration();
}

template <class V, class F>
V*
CVVHashTable<V, F>::IterateValue()
{
	return GetNextValue(&Iter);
}

template <class V, class F>
void
CVVHashTable<V, F>::EndIterateValues()
{
	EndIteration(Iter);
	Iter = NULL;
}

template <class V, class F>
void* 
CVVHashTable<V, F>::BeginIteration()
{
	NumOutstandingIterators ++;
	return (void*) Elements;
}

template <class V, class F>
V* 
CVVHashTable<V, F>::GetNextValue(void **pos)
{
	V*** current = (V***) pos;

	while (1)
	{
		if ((*current) == (Elements + NumElements))
			return NULL;

		if ((**current) != NULL)
		{
			if ((**current) != ((V*) 0xffffffff))
			{
				V* value = **current;
				(*current)++;
				return value;
			}
		}

		(*current)++;
	}
}

template <class V, class F>
void
CVVHashTable<V, F>::EndIteration(void *pos)
{
	NumOutstandingIterators --;
}


#endif