#ifndef __XGLImportCallback_h
#define __XGLImportCallback_h

class CXGLMesh;
class CXGLData;
class CXGLObject;
class CXGLBackground;
class CXGLLighting;
class CXGLDataHandle;
class CXGLWorld;
class CXGLTexture;
class CXGLTextureBuffer;
class CXGLExtgeom;

/**************************************

The following set of callback functions
should be implemented in order to import
XGL into your progam.

**************************************/
class CXGLImportCallback
{
public:

/*
WorldCallback()

This callback is made when all of the data in the world
has been imported.  This is the last callback that will be 
made.  The ownership of this world is now yours.  You must delete this
world when you are through with it.
*/
	virtual void WorldCallback(CXGLWorld *world) = 0;

	
/*
MeshCallback(CXGLMesh *mesh)

This callback is made when a mesh is encountered
in the XGL file.  The CXGLMesh structure contains the data for the
mesh.  The ownership of this mesh is now yours.  You must delete this mesh
when you are through with it.

The implementation of this function should return a handle so that the
XGLImportCallback can identify this mesh by that handle at a later point.
For example when you later receive an object callback you can do the following:

	int meshIndex = object->GetMeshCallbackHandle();

This meshIndex is the same value that you passed back in this callback and indicates
that the above object contains this mesh.
*/
	virtual int MeshCallback(CXGLMesh* mesh) = 0;

/*
ObjectCallback(CXGLObject *object)

This callback is made when an instance of a object at a particular
location in the world is encountered in an XGL file.  The transform
specifies the location in the object's parent space where the mesh should
appear.  The ownership of this object is now yours.  You must delete this object
when you are through with it.

The implementation of this function should return a handle so that the XGLImportCallback
can identify this object by that handle at a later point.
For example when you later receive a world callback you can do the following:

	int i;
	for (i = 0; i < world->NumObjects; i++)
	{
		int objectIndex = world->GetObjectCallbackHandle(i);
	}

This objectIndex(ices) is the same value you passed back in this callback and indicates that
the above world contains this object.  The pathID is an attribute of the object which is unique only to 
this object and it's siblings.  The pathID is a user-defined tag typically used to define
a unique path through the object tree.
*/
	virtual int ObjectCallback(CXGLObject *object) = 0;

/*
DataCallback(CXGLData *data)

This callback is made when a block of data is encountered in an XGL file.  Data can be
associated either with a patch or an object.  The ownership of this data is now yours.  You
must delete this object when you are through with it.

The implementation of this function should return a handle so that the XGLImportCallback
can identify this data by that handle at a later point.  For example, when you later receive an
object you can do the following:

	int i;
	for (i = 0; i < object->NumDatas; i++)
	{
		int dataIndex = object->GetDataCallbackHandle(i);
	}

This dataIndex(ices) is the same value you passed back in this callback and indicates that
the above object contains this data.
*/

	virtual int DataCallback(CXGLData *data) = 0;

	
/*
TextureCallback(CXGLTexture* texture)

This callback is made when a texture is encountered in an XGL file.  Textures are associated
with meshes.  The ownership of this data is now yours.  You must delete this object when you are
through with it.

The implementation of this function should return a handle so that the XGLImportCallback can
identify this texture by that handle at a later point.  For example, when you later receive a
mesh you can do the following:

	int i;
	for (i = 0; i < mesh->NumTextures; i++)
	{
		int textureIndex = mesh->GetTextureCallbackHandle(i);
	}

This textureIndex(ices) is the same value you passed back in this callback and indicates that
the above mesh contains this texture.
*/

	virtual int TextureCallback(CXGLTexture *texture) = 0;

/*
TextureBufferCallback(CXGLTextureBuffer* textureBuffer)

This callback is make when a textureRGBBuffer or textureRGBABuffer is encountered in an XGL file.
TextureBuffers are associated with textures.  The ownership of this data is now yours.  You must
delete this object when you are through with i.

The implementation of this function should return a handle so that the XGLImportCallback can
identify the texture by that handle at a later point.  For example, when you later receive a
texture you can do the following:

	int textureBufferIndex = texture->GetTextureBufferCallbackHandle();

This textureBufferIndex is the same value you passed back in this callback and indicates that
the above texture cotains this textureBuffer.
*/

	virtual int TextureBufferCallback(CXGLTextureBuffer *textureBuffer) = 0;

/*
ErrorCallback(const char* errorString)

This callback is made if an error is detected in the XGL file.  The 
errorString parameter includes text that describes the error.  
*/
	virtual void ErrorCallback(const char* errorString) = 0;

/*
BackgroundCallback(CXGLBackground* background)

The ownership of this background is now yours.  You must delete it when you
are through with it.
*/
	virtual void BackgroundCallback(CXGLBackground* background) = 0;

/*
LightingCallback(CXGLLighting* lighting)

The ownership of this lighting is now yours.  You must delete it when you are
through with it.
*/
	virtual void LightingCallback(CXGLLighting* lighting) = 0;

/*
ExtgeomCallback(CXGLExtgeom* extgeom)

The ownership of this extgeom is now yours.  You must delete it when you are
through with it.
*/
	virtual int ExtgeomCallback(CXGLExtgeom* extgeom) = 0;

};
#endif