#include "XGLMeshSmooth.h"
#include "XGLTypes.h"
#include <math.h>
#include <memory.h>

class CSmoothNormal
{
public:
	double X, Y, Z;
};



bool
Normalize(double *nx, double *ny, double *nz)
{
	double mag = sqrt((*nx)*(*nx)+(*ny)*(*ny)+(*nz)*(*nz));
	if (mag == 0.0)
	{
		(*nx)=0; (*ny)=0; (*nz)=0;
		return false;
	}
	else
	{
		double invMag = 1.0/mag;
		(*nx) *= invMag;
		(*ny) *= invMag;
		(*nz) *= invMag;
		return true;
	}
}

bool
GetAreFacesInSmoothGroup(CXGLFace *f1, CXGLFace *f2)
{
	if((f1->SmoothingGroups) & (f2->SmoothingGroups))
		return true;

	return false;
}



bool
GetCanSmoothFaces(CSmoothNormal *n1, CSmoothNormal *n2)
{
	const double squareRootOfOneHalf = 0.70710678118654752440084436210485;

	if ((n1->X*n2->X + n1->Y*n2->Y + n1->Z*n2->Z) < squareRootOfOneHalf)
		return false;
	else
		return true;
	
}

CXGLFaceVertex*
GetXGLFaceVertex(unsigned int fvI, CXGLFace *faces, unsigned int *fI, unsigned int *vI)
{
	*fI = fvI / 3;
	*vI = fvI - ((*fI)*3);
	return faces[(*fI)].FaceVertices + *vI;
}

double
GetAngleForFV(CXGLFace *face, unsigned int fvI, CXGLVertex *vertices)
{
	CXGLVertex *v = vertices + face->FaceVertices[fvI].VertexI;
	fvI++;
	if (fvI > 2) fvI = 0;
	CXGLVertex *v1 = vertices + face->FaceVertices[fvI].VertexI;
	fvI++;
	if (fvI > 2) fvI = 0;
	CXGLVertex *v2 = vertices + face->FaceVertices[fvI].VertexI;

	double dx1 = v1->X - v->X;
	double dy1 = v1->Y - v->Y;
	double dz1 = v1->Z - v->Z;
	if (Normalize(&dx1, &dy1, &dz1) == false)
		return 0.0;

	double dx2 = v2->X - v->X;
	double dy2 = v2->Y - v->Y;
	double dz2 = v2->Z - v->Z;
	if (Normalize(&dx2, &dy2, &dz2) == false)
		return 0.0;

	double dot = dx1*dx2 + dy1*dy2 + dz1*dz2;

	return acos(dot);
}

void
SmoothFaceVertices(unsigned int numFVs, unsigned int *fvs, CXGLFace *faces, CXGLVertex *vertices, CSmoothNormal *faceNormals, CXGLNormal *normals, unsigned int *numNormals)
{
	unsigned int i;
	for (i=0; i<numFVs; i++)
	{
		unsigned int fI, vI;
		CXGLFaceVertex *xglFV = GetXGLFaceVertex(fvs[i], faces, &fI, &vI);
		if (xglFV->NormalI == -1)
		{
			double nx = 0.0;
			double ny = 0.0;
			double nz = 0.0;

			unsigned int j;
			for (j=0; j<numFVs; j++)
			{
				unsigned int testFI, testVI;
				GetXGLFaceVertex(fvs[j], faces, &testFI, &testVI);
				if (GetAreFacesInSmoothGroup(faces + fI, faces + testFI))
				{
					if (GetCanSmoothFaces(faceNormals + fI, faceNormals + testFI))
					{
						double angle = GetAngleForFV(faces + testFI, testVI, vertices);
						nx += faceNormals[testFI].X * angle;
						ny += faceNormals[testFI].Y * angle;
						nz += faceNormals[testFI].Z * angle;
					}
				}
			}

			if (Normalize(&nx, &ny, &nz))
			{
				normals[*numNormals].X = (float) nx;
				normals[*numNormals].Y = (float) ny;
				normals[*numNormals].Z = (float) nz;
			}
			else
			{
				normals[*numNormals].X = (float) faceNormals[fI].X;
				normals[*numNormals].Y = (float) faceNormals[fI].Y;
				normals[*numNormals].Z = (float) faceNormals[fI].Z;
			}
			xglFV->NormalI = (*numNormals);
			(*numNormals)++;

		}
	}
}

CXGLVertex *SortVertices;

int __cdecl 
CompareVertexIndices(const void *arg1, const void *arg2 )
{
	unsigned __int32 v1 = *((unsigned __int32*) arg1);
	unsigned __int32 v2 = *((unsigned __int32*) arg2);

	if (SortVertices[v1].X < SortVertices[v2].X)
		return 1;
	if (SortVertices[v1].X > SortVertices[v2].X)
		return -1;

	if (SortVertices[v1].Y < SortVertices[v2].Y)
		return 1;
	if (SortVertices[v1].Y > SortVertices[v2].Y)
		return -1;

	if (SortVertices[v1].Z < SortVertices[v2].Z)
		return 1;
	if (SortVertices[v1].Z > SortVertices[v2].Z)
		return -1;

	return 0;
}

void
FindVertexDuplicates(CXGLVertex *vertices, unsigned __int32 *newVRefs, unsigned __int32 numVertices)
{
	unsigned __int32 *vertexOrder = new unsigned __int32[numVertices];
	unsigned __int32 i;
	for (i=0; i<numVertices; i++)
		vertexOrder[i] = i;
	
	SortVertices = vertices;
	qsort(vertexOrder, numVertices, sizeof(unsigned __int32), CompareVertexIndices);

	unsigned __int32 referencedVertex = vertexOrder[0];
	for (i=0; i<numVertices; i++)
	{
		if (CompareVertexIndices((void*) &referencedVertex, (void*) &vertexOrder[i]) != 0)
		{
			referencedVertex = vertexOrder[i];
		}
		
		newVRefs[vertexOrder[i]] = referencedVertex;
	}

	delete[] vertexOrder;
	

}

void SmoothXGLMesh(CXGLMesh *mesh)
{
	CXGLVertex *vertices = mesh->Vertices;
	unsigned int numVertices = mesh->NumVertices;

	CXGLFace *faces = mesh->Faces;
	unsigned int numFaces = mesh->NumFaces;


	
	unsigned int numNormals = mesh->NumNormals;
	CXGLNormal *normals = new CXGLNormal[numNormals + 3*numFaces];
	memcpy(normals, mesh->Normals, numNormals * sizeof(CXGLNormal));
	
	//This list will contain linked listst that link fvs that touch a vertex
	unsigned int *fvList = new unsigned int[numFaces * 3];

	//This is a map of vertices to fvs
	unsigned int *fvsForVertex = new unsigned int[numVertices];

	CSmoothNormal *faceNormals = new CSmoothNormal[numFaces];

	unsigned int i;
	for (i=0; i<numVertices; i++)
	{
		fvsForVertex[i] = 0xffffffff;
	}

	unsigned __int32 *newVRefs = new unsigned __int32[numVertices];
	FindVertexDuplicates(vertices, newVRefs, numVertices);

	unsigned int fvI = 0;
	for (i=0; i<numFaces; i++)
	{
		unsigned int v1 = newVRefs[faces[i].FaceVertices[0].VertexI];
		faces[i].FaceVertices[0].VertexI = v1;

		unsigned int v2 = newVRefs[faces[i].FaceVertices[1].VertexI];
		faces[i].FaceVertices[1].VertexI = v2;

		unsigned int v3 = newVRefs[faces[i].FaceVertices[2].VertexI];
		faces[i].FaceVertices[2].VertexI = v3;

		//Compute the face normal
		double dx1 = vertices[v2].X - vertices[v1].X;
		double dy1 = vertices[v2].Y - vertices[v1].Y;
		double dz1 = vertices[v2].Z - vertices[v1].Z;

		double dx2 = vertices[v3].X - vertices[v1].X;
		double dy2 = vertices[v3].Y - vertices[v1].Y;
		double dz2 = vertices[v3].Z - vertices[v1].Z;
		
		faceNormals[i].X = dy1*dz2 - dy2*dz1;
		faceNormals[i].Y = dz1*dx2 - dz2*dx1;
		faceNormals[i].Z = dx1*dy2 - dx2*dy1;

		Normalize(&(faceNormals[i].X), &(faceNormals[i].Y), &(faceNormals[i].Z));

		//Build the lists to link up face vertices
		fvList[fvI] = fvsForVertex[v1];
		fvsForVertex[v1] = fvI;
		fvI++;

		fvList[fvI] = fvsForVertex[v2];
		fvsForVertex[v2] = fvI;
		fvI++;

		fvList[fvI] = fvsForVertex[v3];
		fvsForVertex[v3] = fvI;
		fvI++;
	}

	delete[] newVRefs;



	unsigned int *fvs = new unsigned int[numFaces * 3];
	for (i=0; i<numVertices; i++)
	{
		unsigned int fvI = fvsForVertex[i];
		if (fvI != 0xffffffff)
		{
			unsigned int numFVs = 0;
			while (1)
			{
				fvs[numFVs] = fvI;
				numFVs++;

				unsigned int nextFVI = fvList[fvI];
				if (nextFVI == 0xffffffff)
					break;
				
				fvList[fvI] = 0xffffffff;
				fvI = nextFVI;
			}

			SmoothFaceVertices(numFVs, fvs, faces, vertices, faceNormals, normals, &numNormals);
		}
	}

	

	mesh->NumNormals = numNormals;
	delete[] mesh->Normals;

	mesh->Normals = new CXGLNormal[numNormals];
	memcpy(mesh->Normals, normals, sizeof(CXGLNormal) * numNormals);

	
	delete[] fvs;
	delete[] fvList;
	delete[] fvsForVertex;
	delete[] faceNormals;
	delete[] normals;



}





 	



		




		
	





