#include "XGLResolver.h"
#include "XGLContext.h"
#include "XGLImport.h"
#include "XGLErrors.h"
#include "XGLImportCallback.h"
#include "XGLTypes.h"
#include "XGLHandleTypes.h"
#include "XGLMeshSmooth.h"
#include <memory.h>
#include <assert.h>

void 
CXGLResolver::Init(CXGLImportCallback *callback)
{
	Callback = callback;

	//initialize our handles
	ObjectHandles.SetSize(1);
	MeshHandles.SetSize(1);
	TextureHandles.SetSize(1);
	TextureBufferHandles.SetSize(1);
}
	

void
CXGLResolver::UnInit()
{
	int i; 

	int numToDelete = ObjectHandles.GetNumElements() - 1;

	for (i = 0; i < numToDelete; i++)
	{
		CXGLObjectHandle *toDelete = ObjectHandles.GetElement(i);
		delete toDelete;
	}

	numToDelete = MeshHandles.GetNumElements() - 1;

	for (i = 0; i < numToDelete; i++)
	{
		CXGLMeshHandle *toDelete = MeshHandles.GetElement(i);
		delete toDelete;
	}

	numToDelete = TextureHandles.GetNumElements() - 1;

	for (i = 0; i < numToDelete; i++)
	{
		CXGLTextureHandle *toDelete = TextureHandles.GetElement(i);
		delete toDelete;
	}

	numToDelete = TextureBufferHandles.GetNumElements() - 1;

	for (i = 0; i < numToDelete; i++)
	{
		CXGLTextureBufferHandle *toDelete = TextureBufferHandles.GetElement(i);
		delete toDelete;
	}

	ObjectHandles.SetSize(0);
	MeshHandles.SetSize(0);
	TextureHandles.SetSize(0);
	TextureBufferHandles.SetSize(0);
}

CXGLContext*
CXGLResolver::CreateContext(CXGLContext *parent)
{
	CXGLContext *ret = new CXGLContext;
	
	if (parent != NULL)
		parent->AddRef();

	ret->Parent = parent;
	ret->Init();
	
	return ret;
}

//returns parent
CXGLContext*
CXGLResolver::ReleaseContext(CXGLContext *context)
{
	CXGLContext *ret = context->Parent;
	context->ReleaseRef(); //this line should destroy us if we have no children
	return ret;
}


//definitions
void
CXGLResolver::AddObjectDefinition(CXGLContext *addContext, CXGLContext *resolveContext, int xglID, CXGLObject *object)
{
	CXGLObjectHandle *ret = new CXGLObjectHandle;
	ObjectHandles.SetElement((ObjectHandles.GetNumElements() - 1), ret);
	ObjectHandles.SetSize(ObjectHandles.GetNumElements() + 1);
	
	ret->ObjectPathID = object->PathID;

	if (CanObjectResolve(resolveContext, object))
	{
		ret->CallbackHandle = Callback->ObjectCallback(object);
		ret->Object = NULL;
		ret->Context = NULL;
		//The object is now in the Output State
	}
	else
	{
		resolveContext->AddRef();
		ret->Object = object;
		ret->CallbackHandle = -1;
		ret->Context = resolveContext;

		//The object is now in the Incomplete State
	}

	ret->XGLIDObjectReferenceID = -1;
	addContext->SetObjectHandle(xglID, ret);

	CheckAndResolveAllObjects();
}

CXGLObjectHandle*
CXGLResolver::CreateObjectReference(CXGLContext *context, int xglID)
{
	CXGLObjectHandle *ret;
	ret = context->GetObjectHandleForXGLID(xglID);
	if (ret == NULL)
	{
		ret = new CXGLObjectHandle;
		ObjectHandles.SetElement((ObjectHandles.GetNumElements() - 1), ret);
		ObjectHandles.SetSize(ObjectHandles.GetNumElements() + 1);

		context->AddRef();
		ret->CallbackHandle = -1;
		ret->Context = context;
		ret->Object = NULL;
		ret->ObjectPathID = -1;
		ret->XGLIDObjectReferenceID = xglID;

		//The object is in the Undefined State
	}

	return ret;
}

void
CXGLResolver::AddMeshDefinition(CXGLContext *addContext, CXGLContext *resolveContext, int xglID, CXGLMesh *mesh)
{
	CXGLMeshHandle *ret = new CXGLMeshHandle;
	MeshHandles.SetElement((MeshHandles.GetNumElements() - 1), ret);
	MeshHandles.SetSize(MeshHandles.GetNumElements() + 1);

	if (CanMeshResolve(resolveContext, mesh))
	{
		SmoothXGLMesh(mesh);
		ret->CallbackHandle = Callback->MeshCallback(mesh);
		ret->Mesh = NULL;
		ret->Context = NULL;
	
		//The mesh is now in the Output State
	}
	else
	{
		resolveContext->AddRef();
		ret->CallbackHandle = -1;
		ret->Mesh = mesh;
		ret->Context = resolveContext;
	
		//The mesh is now in the Incomplete State
	}

	//The mesh is not in the Undefined State
	ret->XGLIDMeshReferenceID = -1;
	addContext->SetMeshHandle(xglID, ret);
	
	CheckAndResolveAllObjects();
}

CXGLMeshHandle*
CXGLResolver::CreateMeshReference(CXGLContext *context, int xglID)
{
	CXGLMeshHandle *ret;
	ret = context->GetMeshHandleForXGLID(xglID);
	if (ret == NULL)
	{
		ret = new CXGLMeshHandle;
		MeshHandles.SetElement((MeshHandles.GetNumElements() - 1), ret);
		MeshHandles.SetSize(MeshHandles.GetNumElements() + 1);

		context->AddRef();

		ret->CallbackHandle = -1;
		ret->Context = context;
		ret->Mesh = NULL;
		ret->XGLIDMeshReferenceID = xglID;
		//The object is in the Undefined State
	}
	return ret;
}

void
CXGLResolver::AddTextureDefinition(CXGLContext *context, int xglID, CXGLTexture *texture)
{
	CXGLTextureHandle *ret = new CXGLTextureHandle;
	TextureHandles.SetElement((TextureHandles.GetNumElements() - 1), ret);
	TextureHandles.SetSize(TextureHandles.GetNumElements() + 1);

	if (CanTextureResolve(context, texture))
	{
		ret->CallbackHandle = Callback->TextureCallback(texture);
		ret->Texture = NULL;
		ret->Context = NULL;
	
		//The texture is now in the Output State
	}
	else
	{
		context->AddRef();
		ret->CallbackHandle = -1;
		ret->Texture = texture;
		ret->Context = context;
	
		//The texture is now in the Incomplete State
	}

	//The texture is not in the Undefined State
	ret->XGLIDTextureReferenceID = -1;
	context->SetTextureHandle(xglID, ret);
	
	CheckAndResolveAllObjects();
}

CXGLTextureHandle*
CXGLResolver::CreateTextureReference(CXGLContext *context, int xglID)
{
	CXGLTextureHandle *ret;
	ret = context->GetTextureHandleForXGLID(xglID);
	if (ret == NULL)
	{
		ret = new CXGLTextureHandle;
		TextureHandles.SetElement((TextureHandles.GetNumElements() - 1), ret);
		TextureHandles.SetSize(TextureHandles.GetNumElements() + 1);

		context->AddRef();

		ret->CallbackHandle = -1;
		ret->Context = context;
		ret->Texture = NULL;
		ret->XGLIDTextureReferenceID = xglID;
		//The object is in the Undefined State
	}
	return ret;
}

void
CXGLResolver::AddTextureBufferDefinition(CXGLContext *context, int xglID, CXGLTextureBuffer *textureBuffer)
{
	CXGLTextureBufferHandle *ret = new CXGLTextureBufferHandle;
	TextureBufferHandles.SetElement((TextureBufferHandles.GetNumElements() - 1), ret);
	TextureBufferHandles.SetSize(TextureBufferHandles.GetNumElements() + 1);

	ret->CallbackHandle = Callback->TextureBufferCallback(textureBuffer);
	ret->Context = NULL;
	//The textureBuffer is now in the Output State

	//The texture is not in the Undefined State
	ret->XGLIDTextureBufferReferenceID = -1;
	context->SetTextureBufferHandle(xglID, ret);
	
	CheckAndResolveAllObjects();
}

CXGLTextureBufferHandle*
CXGLResolver::CreateTextureBufferReference(CXGLContext *context, int xglID)
{
	CXGLTextureBufferHandle *ret;
	ret = context->GetTextureBufferHandleForXGLID(xglID);
	if (ret == NULL)
	{
		ret = new CXGLTextureBufferHandle;
		TextureBufferHandles.SetElement((TextureBufferHandles.GetNumElements() - 1), ret);
		TextureBufferHandles.SetSize(TextureBufferHandles.GetNumElements() + 1);

		context->AddRef();

		ret->CallbackHandle = -1;
		ret->Context = context;
		ret->XGLIDTextureBufferReferenceID = xglID;
		//The object is in the Undefined State
	}
	return ret;
}

void
CXGLResolver::AddMaterialDefinition(CXGLContext *context, int xglID, CXGLMaterial *material)
{
	context->SetMaterial(xglID, material);
}

void
CXGLResolver::AddLineStyleDefinition(CXGLContext *context, int xglID, CXGLLineStyle *lineStyle)
{
	context->SetLineStyle(xglID, lineStyle);
}

void
CXGLResolver::AddPointStyleDefinition(CXGLContext *context, int xglID, CXGLPointStyle *pointStyle)
{
	context->SetPointStyle(xglID, pointStyle);
}

void
CXGLResolver::AddVertexDefinition(CXGLContext *context, int xglID, CXGLVertex *vertex)
{
	context->SetVertex(xglID, vertex);	
}

void
CXGLResolver::AddNormalDefinition(CXGLContext *context, int xglID, CXGLNormal *normal)
{
	context->SetNormal(xglID, normal);
}

void
CXGLResolver::AddTextureCoordinateDefinition(CXGLContext *context, int xglID, CXGLTextureCoordinate *textureCoordinate)
{
	context->SetTextureCoordinate(xglID, textureCoordinate);
}


void
CXGLResolver::AddWorld(CXGLContext* context, CXGLWorld *world)
{
	//don't remove the check...
	ResolveWorld(context, world);
	Callback->WorldCallback(world);
}

void
CXGLResolver::AddBackground(CXGLContext *context, CXGLBackground *background)
{
	context->SetBackground();
	Callback->BackgroundCallback(background);
}

void
CXGLResolver::AddLighting(CXGLContext *context, CXGLLighting *lighting)
{
	context->SetLighting();
	Callback->LightingCallback(lighting);
}

CXGLDataHandle*
CXGLResolver::AddData(CXGLContext *context, CXGLData *data)
{
	return (CXGLDataHandle*)(Callback->DataCallback(data));
}

CXGLObjectHandle*
CXGLResolver::AddObject(CXGLContext *context, CXGLObject *object)
{
	CXGLObjectHandle *ret = new CXGLObjectHandle;
	ObjectHandles.SetElement((ObjectHandles.GetNumElements() - 1), ret);
	ObjectHandles.SetSize(ObjectHandles.GetNumElements() + 1);
	
	ret->ObjectPathID = object->PathID;

	if (CanObjectResolve(context, object))
	{
		//make object callback
		ret->CallbackHandle = Callback->ObjectCallback(object);
		ret->Object = NULL;
		ret->Context = NULL;
		ret->XGLIDObjectReferenceID = -1;
	}
	else
	{
		context->AddRef();
		ret->CallbackHandle = -1;
		ret->Context = context;
		ret->XGLIDObjectReferenceID = -1;
		ret->Object = object;
	}

	return ret;
}

CXGLMeshHandle*
CXGLResolver::AddMesh(CXGLContext *context, CXGLMesh *mesh)
{
	CXGLMeshHandle *ret = new CXGLMeshHandle;
	MeshHandles.SetElement((MeshHandles.GetNumElements() - 1), ret);
	MeshHandles.SetSize(MeshHandles.GetNumElements() + 1);
	
	if (CanMeshResolve(context, mesh))
	{
		//make mesh callback
		SmoothXGLMesh(mesh);
		ret->CallbackHandle = Callback->MeshCallback(mesh);
		ret->Mesh = NULL;
		ret->Context = NULL;
		ret->XGLIDMeshReferenceID = -1;
	}
	else
	{
		ret->Mesh = new CXGLMesh;
		memcpy(ret->Mesh, mesh, sizeof(CXGLMesh));

		context->AddRef();
		ret->CallbackHandle = -1;
		ret->Context = context;
		ret->XGLIDMeshReferenceID = -1;
	}

	return ret;
}

CXGLTextureHandle*
CXGLResolver::AddTexture(CXGLContext *context, CXGLTexture *texture)
{
	CXGLTextureHandle *ret = new CXGLTextureHandle;
	TextureHandles.SetElement((TextureHandles.GetNumElements() - 1), ret);
	TextureHandles.SetSize(TextureHandles.GetNumElements() + 1);
	
	if (CanTextureResolve(context, texture))
	{
		//make texture callback
		ret->CallbackHandle = Callback->TextureCallback(texture);
		ret->Texture = NULL;
		ret->Context = NULL;
		ret->XGLIDTextureReferenceID = -1;
	}
	else
	{
		ret->Texture = new CXGLTexture;
		memcpy(ret->Texture, texture, sizeof(CXGLTexture));

		context->AddRef();
		ret->CallbackHandle = -1;
		ret->Context = context;
		ret->XGLIDTextureReferenceID = -1;
	}

	return ret;
}

CXGLTextureBufferHandle* 
CXGLResolver::AddTextureBuffer(CXGLContext *context, CXGLTextureBuffer *textureBuffer)
{
	CXGLTextureBufferHandle *ret = new CXGLTextureBufferHandle;
	TextureBufferHandles.SetElement((TextureBufferHandles.GetNumElements() - 1), ret);
	TextureBufferHandles.SetSize(TextureBufferHandles.GetNumElements() + 1);

	//make the callback
	ret->CallbackHandle = Callback->TextureBufferCallback(textureBuffer);
	ret->Context = NULL;
	ret->XGLIDTextureBufferReferenceID = -1;

	return ret;
}

void 
CXGLResolver::AddError(const char *errorString)
{
	Callback->ErrorCallback(errorString);
}

//resolve functions
int 
CXGLResolver::CanObjectResolve(CXGLContext *context, CXGLObject *object)
{
	if (object == NULL)
		return 1;

	//we cannot resolve if we're going to loop infinitely trying to resolve ourself from within ourself
	if (object->IsInResolveLoop == 1)
		return 0;
	object->IsInResolveLoop = 1;


	unsigned int i;
	for (i = 0; i < object->NumObjects; i++)
	{
		CXGLObjectHandle *child = object->ObjectHandles.GetElement(i);
		if (child->CallbackHandle == -1)
		{
			//If the child object has not yet been output, check if we can output it

			//Try to make the child Incomplete if it is Undefined
			if (child->XGLIDObjectReferenceID != -1)
			{
				CXGLObjectHandle *newChild = child->Context->GetObjectHandleForXGLID(child->XGLIDObjectReferenceID);
				if (newChild == NULL)
				{
					object->IsInResolveLoop = 0;
					return 0;
				}

				//bugbug - Leaks child
				child->Context->ReleaseRef();
				child->Context = NULL;

				child = newChild;
				object->ObjectHandles.SetElement(i, newChild);
			}

			//At this point, child is either Incomplete or Output

			//If child is incomplete, try to make it Output
			if (child->CallbackHandle == -1)
			{
				if (CanObjectResolve(child->Context, child->Object) == 0)
				{
					object->IsInResolveLoop = 0;
					return 0;
				}

				//Output the object
				child->CallbackHandle = Callback->ObjectCallback(child->Object);
				child->Object = NULL;
			}
			
			//At this point, the child is in the Output State
		}
		//Make sure the child has been output
		assert(child->CallbackHandle != -1);
	}

	//we no longer have to worry about trying to resolve ourself from within ourself here
	object->IsInResolveLoop = 0;

	if (object->MeshHandle != NULL)
	{
		if (object->MeshHandle->CallbackHandle == -1)
		{
			//If the mesh has not yet been output, try to output it.

			//Try to make the mesh Incomplete if it is UnDefined
			if (object->MeshHandle->XGLIDMeshReferenceID != -1)
			{
				CXGLMeshHandle *newMeshHandle = 
					object->MeshHandle->Context->GetMeshHandleForXGLID(
					object->MeshHandle->XGLIDMeshReferenceID);

				if (newMeshHandle == NULL)
					return 0;

				//bugbug - Leaks mesh handle

				object->MeshHandle = newMeshHandle;
			}

			//At this point, the Mesh may be Incomplete or Output
			if (object->MeshHandle->CallbackHandle == -1)
			{
				if (CanMeshResolve(object->MeshHandle->Context, object->MeshHandle->Mesh) == 0)
					return 0;

				SmoothXGLMesh(object->MeshHandle->Mesh);
				object->MeshHandle->CallbackHandle = Callback->MeshCallback(object->MeshHandle->Mesh);
				object->MeshHandle->Mesh = NULL;
			}
		}
		assert(object->MeshHandle->CallbackHandle != -1);
	}

	//At this point, all child objects have been output and if this object has a mesh, that mesh has
	//been output, so we can output this object

	return 1;
}

int 
CXGLResolver::CanMeshResolve(CXGLContext *context, CXGLMesh *mesh)
{
	if (mesh == NULL)
		return 1;

	unsigned int i; 
	for (i = 0; i < mesh->NumMaterials; i++)
	{
		if (mesh->Materials[i].XGLID != -1)
		{
			CXGLMaterial *newMat;
			newMat = context->GetMaterial(mesh->Materials[i].XGLID);
			if (newMat == NULL)
				return 0;
		
			memcpy(&(mesh->Materials[i]), newMat, sizeof(CXGLMaterial));
			mesh->Materials[i].XGLID = -1;
		}
	}

	for (i = 0; i < mesh->NumLineStyles; i++)
	{
		if (mesh->LineStyles[i].XGLID != -1)
		{
			CXGLLineStyle *newLineStyle;
			newLineStyle = context->GetLineStyle(mesh->LineStyles[i].XGLID);
			if (newLineStyle == NULL)
				return 0;

			memcpy(&(mesh->LineStyles[i]), newLineStyle, sizeof(CXGLLineStyle));
			mesh->LineStyles[i].XGLID = -1;
		}
	}

	for (i = 0; i < mesh->NumPointStyles; i++)
	{
		if (mesh->PointStyles[i].XGLID != -1)
		{
			CXGLPointStyle *newPointStyle;
			newPointStyle = context->GetPointStyle(mesh->PointStyles[i].XGLID);
			if (newPointStyle == NULL)
				return 0;

			memcpy(&(mesh->PointStyles[i]), newPointStyle, sizeof(CXGLPointStyle));
			mesh->PointStyles[i].XGLID = -1;
		}
	}

	for (i = 0; i < mesh->NumVertices; i++)
	{
		if (mesh->Vertices[i].XGLID != -1)
		{
			CXGLVertex *newV;
			newV = context->GetVertex(mesh->Vertices[i].XGLID);
			if (newV == NULL)
				return 0;
			
			memcpy(&(mesh->Vertices[i]), newV, sizeof(CXGLVertex));
			mesh->Vertices[i].XGLID = -1;
		}
	}

	for (i = 0; i < mesh->NumNormals; i++)
	{
		if (mesh->Normals[i].XGLID != -1)
		{
			CXGLNormal *newN;
			newN = context->GetNormal(mesh->Normals[i].XGLID);
			if (newN == NULL)
				return 0;
			
			memcpy(&(mesh->Normals[i]), newN, sizeof(CXGLNormal));
			mesh->Normals[i].XGLID = -1;
		}
	}

	for (i = 0; i < mesh->NumTextureCoordinates; i++)
	{
		if (mesh->TextureCoordinates[i].XGLID != -1)
		{
			CXGLTextureCoordinate *newTC;
			newTC = context->GetTextureCoordinate(mesh->TextureCoordinates[i].XGLID);
			if (newTC == NULL)
				return 0;

			memcpy(&(mesh->TextureCoordinates[i]), newTC, sizeof(CXGLTextureCoordinate));
			mesh->TextureCoordinates[i].XGLID = -1;
		}
	}

	for (i = 0; i < mesh->NumTextures; i++)
	{
		CXGLTextureHandle *textureHandle = mesh->TextureHandles.GetElement(i);
		if (textureHandle->CallbackHandle == -1)
		{
			//If the texture object has not yet been ouput, check if we can output it

			//Try to make the texture Incomplete if it is Undefined
			if (textureHandle->XGLIDTextureReferenceID != -1)
			{
				CXGLTextureHandle *newTextureHandle =
					textureHandle->Context->GetTextureHandleForXGLID(textureHandle->XGLIDTextureReferenceID);

				if (newTextureHandle == NULL)
					return 0;

				//bugbug - Leaks texture handle
				mesh->TextureHandles.SetElement(i, newTextureHandle);
			}

			//At this point, the Texture may be Incomplete or Output
			textureHandle = mesh->TextureHandles.GetElement(i);
			if (textureHandle->CallbackHandle == -1)
			{
				if (CanTextureResolve(textureHandle->Context, textureHandle->Texture) == 0)
					return 0;

				textureHandle->CallbackHandle = Callback->TextureCallback(textureHandle->Texture);
				textureHandle->Texture = NULL;

			}
		}
		assert(textureHandle->CallbackHandle != -1);
	}
	
	return 1;
}

int
CXGLResolver::CanTextureResolve(CXGLContext *context, CXGLTexture *texture)
{
	if (texture == NULL)
		return 1;

	if (texture->TextureBufferHandle != NULL)
	{
		if (texture->TextureBufferHandle->CallbackHandle == -1)
		{
			//Try to make the textureBuffer Complete if it is UnDefined
			if (texture->TextureBufferHandle->XGLIDTextureBufferReferenceID != -1)
			{
				CXGLTextureBufferHandle *newTextureBufferHandle = 
					texture->TextureBufferHandle->Context->GetTextureBufferHandleForXGLID(
					texture->TextureBufferHandle->XGLIDTextureBufferReferenceID);

				if (newTextureBufferHandle == NULL)
					return 0;

				//bugbug - Leaks textureBuffer handle

				texture->TextureBufferHandle = newTextureBufferHandle;
			}
		}
	
		//At this point, the textureBuffer was found and should be output
		assert(texture->TextureBufferHandle->CallbackHandle != -1);
	}

	return 1;
}

void
CXGLResolver::ResolveWorld(CXGLContext* context, CXGLWorld* world)
{
	CheckAndResolveAllObjects();
}

void
CXGLResolver::CheckAndResolveAllObjects()
{
	int currentObjectHandleIndex = 0;
	CXGLObjectHandle *currentObjectHandle = ObjectHandles.GetElement(currentObjectHandleIndex);
	CXGLObjectHandle *freeObjectHandle = ObjectHandles.GetLastElement();
	while(currentObjectHandle != freeObjectHandle)
	{
		if (currentObjectHandle->CallbackHandle == -1)
		{
			if (currentObjectHandle->XGLIDObjectReferenceID == -1)
			{
				if (CanObjectResolve(currentObjectHandle->Context, currentObjectHandle->Object))
				{
					currentObjectHandle->CallbackHandle = Callback->ObjectCallback(currentObjectHandle->Object);
					currentObjectHandle->Object = NULL;
					currentObjectHandle->Context = NULL;
				}
			}
			else
			{

			}
		}

		currentObjectHandleIndex ++;
		currentObjectHandle = ObjectHandles.GetElement(currentObjectHandleIndex);
	}
}

CXGLExtgeomHandle*
CXGLResolver::AddExtgeom(CXGLContext *context, CXGLExtgeom *extgeom)
{
	return (CXGLExtgeomHandle*)(Callback->ExtgeomCallback(extgeom));
}
