#include "ZLIBCompress.h"
#include <assert.h>
#include <fcntl.h>
#include <io.h>
#include <string.h>

bool
CZLIBBufferCompress::Initialize(const char *zlibModulePath)
{
	UnCompressedBuffer = NULL;
	UnCompressedBufferLength = 0;

	//Initialize the error buffer and the error state
	Error = false;
	strcpy(ErrorBuffer, "");

#ifdef WIN32
	/* run time load the ZLib dll */
	char fullPath[MAX_PATH];
	_snprintf(fullPath, sizeof(fullPath), "%s\\%s", zlibModulePath, "zlib");
	fullPath[sizeof(fullPath) - 1] = '\0';

	DllHandle = LoadLibraryA(fullPath);
	if (DllHandle == NULL)
		DllHandle = LoadLibraryA("zlib");

	if (DllHandle == NULL)
	{
		SetError("Failed to load zlib.dll\r\n");
		return false;
	}
	
	/* find the functions we'll need from the dll */
	*((int*) &deflateInit_) = (int) GetProcAddress(DllHandle, "deflateInit_");
	*((int*) &deflate) = (int) GetProcAddress(DllHandle, "deflate");
	*((int*) &deflateEnd) = (int) GetProcAddress(DllHandle, "deflateEnd");
	*((int*) &deflateReset) = (int) GetProcAddress(DllHandle, "deflateReset");
	if (deflateInit_ == NULL	||
		deflate == NULL			||
		deflateEnd == NULL		||
		deflateReset == NULL)
	{
		BOOL ret = FreeLibrary(DllHandle);
		assert(ret);
		SetError("Invalid zlib.dll\r\n");
		return false;
	}
#endif

	ZStream.opaque = Z_NULL;
	ZStream.zalloc = Z_NULL;
	ZStream.zfree = Z_NULL;

	int ret = deflateInit(&ZStream, Z_BEST_COMPRESSION);
	if (ret < 0)
	{
#ifdef WIN32
		BOOL ret = FreeLibrary(DllHandle);
		assert(ret);
#endif
		SetZLibError("Failed to initialize zlib for compression\r\n", ret);
		return false;
	}

	ZStream.avail_in = 0;

	return true;
}

bool
CZLIBBufferCompress::UnInitialize()
{
	int result = deflateEnd(&ZStream);
	if (result < 0)
	{
#ifdef WIN32
		assert(DllHandle);
		FreeLibrary(DllHandle);
#endif

		SetZLibError("Failed to end compression\r\n", result);
		return false;
	}

#ifdef WIN32
	assert(DllHandle != NULL);
	FreeLibrary(DllHandle);
#endif
	
	return true;
}

bool 
CZLIBBufferCompress::SetData(const char *data)
{
	return SetData(data, strlen(data));
}

bool 
CZLIBBufferCompress::SetData(const char *data, unsigned int length)
{
	assert(data);
	if (data == NULL)
	{
		SetError("Invalid uncompressed data was set\r\n");
		return false;
	}

	assert(UnCompressedBuffer == NULL);
	if (UnCompressedBuffer)
	{
		SetError("Attempted to set uncompressed data twice.\r\n");
		return false;
	}

	UnCompressedBuffer = new char[length+1];

	assert(UnCompressedBuffer != NULL);
	memcpy(UnCompressedBuffer, data, length);

	UnCompressedBufferLength = length;

	return true;
}

bool 
CZLIBBufferCompress::ReleaseData()
{
	assert(UnCompressedBuffer);
	if (UnCompressedBuffer == NULL)
	{
		SetError("No uncompressed data was set\r\n");
		return false;
	}

	delete[] UnCompressedBuffer;
	UnCompressedBuffer = NULL;
	UnCompressedBufferLength = 0;
	return true;
}

bool 
CZLIBBufferCompress::GetCompressedData(char *dataToFill, unsigned int *bytesRead)
{
	assert(dataToFill != NULL);
	if (dataToFill == NULL)
	{
		SetError("Invalid dataToFill pointer\r\n");
		return false;
	}

	assert(UnCompressedBuffer);
	if (UnCompressedBuffer == NULL)
	{
		SetError("Invalid data to compress\r\n");
		return false;
	}

	char *currentPointer = dataToFill;
	*bytesRead = 0;

	if (UnCompressedBufferLength == 0)
		return true; //we just compressed all of the data
	
	ZStream.avail_in = UnCompressedBufferLength;
	ZStream.next_in = (unsigned char*)UnCompressedBuffer;

	while (1)
	{
		ZStream.avail_out = sizeof(CompressedBuffer);
		ZStream.next_out = CompressedBuffer;

		int result = deflate(&ZStream, Z_NO_FLUSH);
		if (result < 0)
		{
			SetZLibError("Compression Failed\r\n", result);
			return false;
		}

		if (ZStream.avail_out == sizeof(CompressedBuffer))
			break;

		unsigned int compressedLength = sizeof(CompressedBuffer) - ZStream.avail_out;
		memcpy(currentPointer, CompressedBuffer, compressedLength);
		currentPointer += compressedLength;
		*bytesRead += compressedLength;

		if (result != Z_OK)
			break;

		if (ZStream.avail_out != 0)
			break;
	}


	unsigned int finishBytesRead;
	bool ret = Finish(currentPointer, &finishBytesRead);

	*bytesRead += finishBytesRead;

	return ret;
}

bool 
CZLIBBufferCompress::GetCompressedDataNoFlush(char *dataToFill, unsigned int *bytesRead)
{
	assert(dataToFill != NULL);
	if (dataToFill == NULL)
	{
		SetError("Invalid dataToFill pointer\r\n");
		return false;
	}

	assert(UnCompressedBuffer);
	if (UnCompressedBuffer == NULL)
	{
		SetError("Invalid data to compress\r\n");
		return false;
	}

	char *currentPointer = dataToFill;
	*bytesRead = 0;

	if (UnCompressedBufferLength == 0)
		return true; //we just compressed all of the data
	
	ZStream.avail_in = UnCompressedBufferLength;
	ZStream.next_in = (unsigned char*)UnCompressedBuffer;

	while (1)
	{
		ZStream.avail_out = sizeof(CompressedBuffer);
		ZStream.next_out = CompressedBuffer;

		int result = deflate(&ZStream, Z_NO_FLUSH);
		if (result < 0)
		{
			SetZLibError("Compression Failed\r\n", result);
			return false;
		}

		if (ZStream.avail_out == sizeof(CompressedBuffer))
			break;

		unsigned int compressedLength = sizeof(CompressedBuffer) - ZStream.avail_out;
		memcpy(currentPointer, CompressedBuffer, compressedLength);
		currentPointer += compressedLength;
		*bytesRead += compressedLength;

		if (result != Z_OK)
			break;

		if (ZStream.avail_out != 0)
			break;
	}

	return true;
}

bool 
CZLIBBufferCompress::Finish(char *dataToFill, unsigned int *bytesRead)
{
	assert(ZStream.avail_in == 0);

	char *currentPointer = dataToFill;
	*bytesRead = 0;
	
	int result = Z_OK;
	while (1)
	{
		ZStream.avail_out = sizeof(CompressedBuffer);
		ZStream.next_out = CompressedBuffer;

		result = deflate(&ZStream, Z_FINISH);
		if (result < 0)
		{
			SetZLibError("Compression Failed\r\n", result);
			return false;
		}

		if (ZStream.avail_out != sizeof(CompressedBuffer))
		{
			unsigned int compressedLength = sizeof(CompressedBuffer) - ZStream.avail_out;
			memcpy(currentPointer, CompressedBuffer, compressedLength);
			currentPointer += compressedLength;
			*bytesRead += compressedLength;
		}
	
		if (result == Z_STREAM_END)
			break;
	}

	return true;
}

bool 
CZLIBBufferCompress::Reset()
{
	if (deflateReset(&ZStream) == Z_OK)
		return true;
	else
		return false;
}
	
char*
CZLIBBufferCompress::GetLastErrorString()
{
	if (Error)
		return ErrorBuffer;
	else
		return NULL;
}

void 
CZLIBBufferCompress::SetError(char *errorString)
{
	assert(strlen(errorString) < 1024);
	strcpy(ErrorBuffer, errorString);
	Error = true;
}

void 
CZLIBBufferCompress::SetZLibError(char *errorString, int zlibError)
{
	SetError(errorString);
	char temp[1024];
	sprintf(temp, "Zlib Error Number: %d\r\n", zlibError);
	strcat(ErrorBuffer, temp);
}

bool
CZLIBFileCompress::Initialize(const char *zlibModulePath)
{
#ifdef WIN32
	/* run time load the ZLib dll */
	char fullPath[MAX_PATH];
	_snprintf(fullPath, sizeof(fullPath), "%s\\%s", zlibModulePath, "zlib");
	fullPath[sizeof(fullPath) - 1] = '\0';

	DllHandle = LoadLibraryA(fullPath);
	if (DllHandle == NULL)
		DllHandle = LoadLibraryA("zlib");

	if (DllHandle == NULL)
	{
		SetError("Failed to load zlib.dll\r\n");
		return false;
	}
	
	/* find the functions we'll need from the dll */
	*((int*) &deflateInit_) = (int) GetProcAddress(DllHandle, "deflateInit_");
	*((int*) &deflate) = (int) GetProcAddress(DllHandle, "deflate");
	*((int*) &deflateEnd) = (int) GetProcAddress(DllHandle, "deflateEnd");
	if (deflateInit_ == NULL	||
		deflate == NULL			||
		deflateEnd == NULL)
	{
		BOOL ret = FreeLibrary(DllHandle);
		assert(ret);
		SetError("Invalid zlib.dll\r\n");
		return false;
	}
#endif

	ZStream.opaque = Z_NULL;
	ZStream.zalloc = Z_NULL;
	ZStream.zfree = Z_NULL;

	int ret = deflateInit(&ZStream, Z_BEST_COMPRESSION);
	if (ret < 0)
	{
#ifdef WIN32
		BOOL ret = FreeLibrary(DllHandle);
		assert(ret);
#endif
		SetZLibError("Failed to initialize zlib for compression\r\n", ret);
		return false;
	}

	ZStream.avail_in = 0;

	return true;
}

bool
CZLIBFileCompress::Initialize(const char *zlibModulePath, const char *fileName)
{
	//Initialize the error buffer and the error state
	Error = false;
	strcpy(ErrorBuffer, "");

	assert(fileName);
	if (fileName == NULL)
	{
		SetError("Invalid Filename\r\n");
		return false;
	}

	/* Create the File Stream */
	Stream = fopen(fileName, "wb");
	if (Stream == NULL)
	{
		char temp[1024];
		sprintf(temp, "Unable to open file %s\r\n", fileName);
		SetError(temp);
		return false;
	}

	IsStreamUserOwned = false;

	return Initialize(zlibModulePath);
}

bool
CZLIBFileCompress::Initialize(const char *zlibModulePath, FILE *file)
{
	//Initialize the error buffer and the error state
	Error = false;
	strcpy(ErrorBuffer, "");

	assert(file);
	if (file == NULL)
	{
		SetError("Invalid File\r\n");
		return false;
	}

	//set the file to binary mode if not already done
	int result = _setmode(_fileno( file ), _O_BINARY );
	if (result == -1 )
		NULL;

	IsStreamUserOwned = true;
	Stream = file;

	return Initialize(zlibModulePath);
}

bool
CZLIBFileCompress::UnInitialize()
{
	if (Error)
	{
#ifdef WIN32
		assert(DllHandle);
		FreeLibrary(DllHandle);
#endif
		return false;
	}

	int result = deflateEnd(&ZStream);
	if (result < 0)
	{
#ifdef WIN32
		assert(DllHandle);
		FreeLibrary(DllHandle);
#endif

		SetZLibError("Failed to end compression\r\n", result);
		return false;
	}

#ifdef WIN32
	assert(DllHandle != NULL);
	FreeLibrary(DllHandle);
#endif

	if (!IsStreamUserOwned)
	{
		result = fclose(Stream);
		if (result != 0)
		{
			SetError("Failed to close file\r\n");
			return false;
		}
	}
	
	return true;
}


bool
CZLIBFileCompress::OutputData(const char *data)
{
	return OutputData(data, strlen(data));
}


bool
CZLIBFileCompress::OutputData(const char *data, unsigned int length)
{
	if (Error)
		return false;

	if (length == 0)
		return true; //we just compressed the data

	assert(data != NULL);
	if (data == NULL)
	{
		SetError("Invalid data to compress\r\n");
		return false;
	}

	ZStream.avail_in = length;
	ZStream.next_in = (unsigned char*)data;

	while (1)
	{
		ZStream.avail_out = sizeof(CompressedBuffer);
		ZStream.next_out = CompressedBuffer;

		int result = deflate(&ZStream, Z_NO_FLUSH);
		if (result < 0)
		{
			SetZLibError("Compression Failed\r\n", result);
			return false;
		}

		if (ZStream.avail_out == sizeof(CompressedBuffer))
			break;

		unsigned int size = fwrite(CompressedBuffer, 1, sizeof(CompressedBuffer) - ZStream.avail_out, Stream);
		if (size != (sizeof(CompressedBuffer) - ZStream.avail_out))
		{
			SetErrorEx("Error writing compressed buffer");
			return false;
		}
		
		if (result != Z_OK)
			break;

		if (ZStream.avail_out != 0)
			break;
	}

	return true;
}

bool 
CZLIBFileCompress::Finish()
{
	if (Error)
		return false;

	assert(ZStream.avail_in == 0);

	int result = Z_OK;
	while (1)
	{
		ZStream.avail_out = sizeof(CompressedBuffer);
		ZStream.next_out = CompressedBuffer;

		result = deflate(&ZStream, Z_FINISH);
		if (result < 0)
		{
			SetZLibError("Compression Failed\r\n", result);
			return false;
		}

		unsigned int size;
		if (ZStream.avail_out != sizeof(CompressedBuffer))
		{
			size = fwrite(CompressedBuffer, 1, sizeof(CompressedBuffer) - ZStream.avail_out, Stream);
			if (size != (sizeof(CompressedBuffer) - ZStream.avail_out))
			{
				SetErrorEx("Error writing compressed buffer");
				return false;
			}
		}
	
		if (result == Z_STREAM_END)
			break;
	}

	return true;
}

char*
CZLIBFileCompress::GetLastErrorString()
{
	if (Error)
		return ErrorBuffer;
	else
		return NULL;
}

void 
CZLIBFileCompress::SetError(char *errorString)
{
	assert(strlen(errorString) < 1024);
	strcpy(ErrorBuffer, errorString);
	Error = true;
}

void 
CZLIBFileCompress::SetErrorEx(char *errorString)
{
	sprintf(ErrorBuffer, "%s: %s\r\n", errorString, strerror(errno));
	Error = true;
}

void 
CZLIBFileCompress::SetZLibError(char *errorString, int zlibError)
{
	SetError(errorString);
	char temp[1024];
	sprintf(temp, "Zlib Error Number: %d\r\n", zlibError);
	strcat(ErrorBuffer, temp);
}

