#include "ZLIBDecompress.h"
#include <assert.h>
#include <fcntl.h>
#include <io.h>
#include <string.h>

bool
CZLIBDecompress::Initialize(const char *zlibModulePath)
{
	//Initialize the error buffer and error state
	Error = false;
	strcpy(ErrorBuffer, "");

#ifdef WIN32
	/* run time load the ZLib dll */
	char fullPath[MAX_PATH];
	_snprintf(fullPath, sizeof(fullPath), "%s\\%s", zlibModulePath, "zlib");
	fullPath[sizeof(fullPath) - 1] = '\0';

	DllHandle = LoadLibraryA(fullPath);
	if (DllHandle == NULL)
		DllHandle = LoadLibraryA("zlib");

	if (DllHandle == NULL)
	{
		SetError("Failed to load zlib.dll\r\n");
		return false;
	}
	
	//find the functions we'll need from the dll
	*((int*) &inflateInit_) = (int) GetProcAddress(DllHandle, "inflateInit_");
	*((int*) &inflate) = (int) GetProcAddress(DllHandle, "inflate");
	*((int*) &inflateEnd) = (int) GetProcAddress(DllHandle, "inflateEnd");
	*((int*) &inflateReset) = (int) GetProcAddress(DllHandle, "inflateReset");
	if (inflateInit_ == NULL	||
		inflate	== NULL			||
		inflateEnd == NULL		||
		inflateReset == NULL)
	{
		BOOL ret = FreeLibrary(DllHandle);
		assert(ret);
		SetError("Invalid zlib.dll\r\n");
		return false;
	}
#endif

	ZStream.opaque = Z_NULL;
	ZStream.zalloc = Z_NULL;
	ZStream.zfree = Z_NULL;
	ZStream.avail_in = 0;
	ZStream.next_in = Z_NULL;

	NumCompressedBytes = 0;
	CompressedBuffer = NULL;

	int ret = inflateInit(&ZStream);
	if (ret < 0)
	{
#ifdef WIN32
		BOOL result = FreeLibrary(DllHandle);
		assert(result);
#endif
		SetZLibError("Failed to initialize zlib for decompression\r\n", ret);
		return false;
	}
	
	return true;
}

bool
CZLIBDecompress::SetCompressedData(unsigned int compressedDataLength, const char *compressedData)
{
	assert(CompressedBuffer == NULL);
	if (CompressedBuffer != NULL)
	{
		SetError("Attempted to set compressed data twice.\r\n");
		return false;
	}

	CompressedBuffer = new char[compressedDataLength];
	memcpy(CompressedBuffer, compressedData, sizeof(char)*compressedDataLength);
	NumCompressedBytes = compressedDataLength;

	ZStream.avail_in = NumCompressedBytes;
	ZStream.next_in = (unsigned char*) CompressedBuffer;

	return true;
}

bool
CZLIBDecompress::ReleaseCompressedData()
{
	assert(CompressedBuffer);
	if (CompressedBuffer == NULL)
	{
		SetError("Attempted to release compressed data before setting any.\r\n");
		return false;
	}

	delete[] CompressedBuffer;
	CompressedBuffer = NULL;
	NumCompressedBytes = 0;
	return true;
}

bool 
CZLIBDecompress::UnInitialize()
{
	int result = inflateEnd(&ZStream);
	if (result < 0)
	{
#ifdef WIN32
		assert(DllHandle);
		FreeLibrary(DllHandle);
#endif
		SetZLibError("Failed to end decompression\r\n", result);
		return false;
	}

#ifdef WIN32
	assert(DllHandle);
	FreeLibrary(DllHandle);
#endif

	return true;
}

unsigned int 
CZLIBDecompress::GetFillBufferLength()
{
	return 65536;
}

bool 
CZLIBDecompress::GetData(char *dataToFill, unsigned int *bytesRead, bool *needsMoreCompressedData, bool *finished)
{
	assert(dataToFill);
	if (dataToFill == NULL)
	{
		SetError("Invalid Fill Buffer\r\n");
		return false;
	}

	assert(NumCompressedBytes != 0);

	ZStream.avail_out = GetFillBufferLength();
	ZStream.next_out = (unsigned char*) dataToFill;

	int result = inflate(&ZStream, Z_NO_FLUSH);
	if (result < 0)
	{
		SetZLibError("Decompression Failed\r\n", result);
		return false;
	}

	*bytesRead = GetFillBufferLength() - ZStream.avail_out;
	*needsMoreCompressedData = (ZStream.avail_in == 0);
	*finished = (result == Z_STREAM_END);
	
	if(result == Z_STREAM_END)
		int doSomething = 1;

	return true;
}

bool 
CZLIBDecompress::Reset()
{
	int result = inflateReset(&ZStream);
	if (result == Z_OK)
		return true;
	else
	{
		SetZLibError("Failed to reset the decompressor\r\n", result);
		return false;
	}
}

char*
CZLIBDecompress::GetLastErrorString()
{
	if (Error)
		return ErrorBuffer;
	else
		return NULL;
}

void 
CZLIBDecompress::SetError(char *errorString)
{
	assert(strlen(errorString) < 1024);
	strcpy(ErrorBuffer, errorString);
	Error = true;
}

void 
CZLIBDecompress::SetZLibError(char *errorString, int zlibError)
{
	SetError(errorString);
	char temp[1024];
	sprintf(temp, "Zlib Error Number: %d\r\n", zlibError);
	strcat(ErrorBuffer, temp);
}
