#include "stdafx.h"
#include "Import.h"
#include <stdio.h>
#include <math.h>
#include "XGLImport\XGLTypes.h"
#include "XGLImport\XGLImport.h"
#include "XGLState.h"
#include "OpenGL.h"
#include "ScaleImage.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

CImport Import;

CImport::CImport()
{
}

CImport::~CImport()
{
}

void
CImport::ImportFile(const char *fileName)
{
	//setup the importer
	CXGLImport xglImport;
	xglImport.InitCallback(this);

	//initialize the resizeable arrays
	ObjectMinX.SetSize(0);
	ObjectMinY.SetSize(0);
	ObjectMinZ.SetSize(0);
	ObjectMaxX.SetSize(0);
	ObjectMaxY.SetSize(0);
	ObjectMaxZ.SetSize(0);
	MeshMinX.SetSize(0);
	MeshMinY.SetSize(0);
	MeshMinZ.SetSize(0);
	MeshMaxX.SetSize(0);
	MeshMaxY.SetSize(0);
	MeshMaxZ.SetSize(0);

	DisplayListNodes.SetSize(0);

	//initialize the parsed data variables
	World = NULL;
	Background = NULL;
	Lighting = NULL;

	//delete the old lists (if any)
	DeleteDisplayLists();

	//Rotate the world to a forward looking direction
	XGLState.SetWorldRotation(0.0, 0.0, 1.0, 0.0, 1.0, 0.0);

	char *lastDot = strrchr(fileName, '.');
	int type = CXGLImport::OtherFileType;
	if (lastDot)
	{
	
		if (stricmp(lastDot, ".xgl") == 0)
			type = CXGLImport::XGLFileType;
		else if (stricmp(lastDot, ".zgl") == 0)
			type = CXGLImport::ZGLFileType;
		else
			type = CXGLImport::OtherFileType;
	}
	

	//start the xglImport importing
	xglImport.Import(fileName, type, NULL);
}

void 
CImport::WorldCallback(CXGLWorld *world)
{
	//the world is finished and therefore the entire file is parsed
	World = world;

	//This function runs through all of the objects checking to see if they
	//are children of other objects.  If they are children we need to calculate
	//their transforms from world space to object space.
	//CalculateWorldTransforms();

	//This function calculates the total world size our objects take up
	//so that we can set our view frustum to a reasonable size for viewing
	CalculateWorldSize();

	//This creates a world display list that will be used to display the parsed file
	CreateWorldDisplayList(World);

	//This deletes the data structures used to store the parsed data.  The
	//data is no longer needed because we have already created the display list.
	DeletePreviousData();

	//Display the newly created world display list 
	CallDisplayList();
}

int
CImport::MeshCallback(CXGLMesh *mesh)
{
	/* 
	This function returns us a parsed mesh.  We're going to create a display 
	list for this mesh and return to the importcallback an id for the list, so
	it can link together objects with mesh lists later.
	*/

	int isLinePoints = 0;
	if(mesh->NumLines > 0 || mesh->NumPoints > 0)
		isLinePoints = 1;

	//return if we have no faces

	if (mesh->NumFaces == 0 && isLinePoints == 0)
	{
		//delete the mesh
		delete mesh;
		return -2;
	}
	
	unsigned int ret = DisplayListNodes.GetNumElements();
	CDisplayListNode *current = DisplayListNodes.AddElement();

	unsigned int numNormalFaces, numTextureFaces, numAlphaTextureFaces, numAlphaFaces;
	CalculateFaceCounts(mesh, &numNormalFaces, &numTextureFaces, &numAlphaTextureFaces, &numAlphaFaces);

	//draw the mesh to the list
	if (numNormalFaces || isLinePoints)
		current->NormalFaces = CreateMeshDisplayList(mesh, NormalFaces, (numAlphaFaces || numAlphaTextureFaces));
	if (numTextureFaces || isLinePoints)
		current->TextureFaces = CreateMeshDisplayList(mesh, TextureFaces, (numAlphaFaces || numAlphaTextureFaces));
	if (numAlphaTextureFaces || isLinePoints)
		current->AlphaTextureFaces = CreateMeshDisplayList(mesh, AlphaTextureFaces, true);
	if (numAlphaFaces || isLinePoints)
		current->AlphaFaces = CreateMeshDisplayList(mesh, AlphaFaces, true);

	//calculate the size of the mesh
	double minX, minY, minZ, maxX, maxY, maxZ;
	CalculateMeshSize(mesh, &minX, &minY, &minZ, &maxX, &maxY, &maxZ);
	
	current->ExtentsIndex = MeshMinX.GetNumElements();

	//store the size of the mesh
	MeshMinX.AddElement(minX);
	MeshMinY.AddElement(minY);
	MeshMinZ.AddElement(minZ);
	MeshMaxX.AddElement(maxX);
	MeshMaxY.AddElement(maxY);
	MeshMaxZ.AddElement(maxZ);

	//delete the mesh
	delete mesh;

	return ret;
}

int
CImport::ObjectCallback(CXGLObject *object)
{
	/*
	This function gives us a fully formed object.  We return an id for this 
	object so that the importcallback can link this object to other objects 
	or the world later
	*/
	
	unsigned int ret = DisplayListNodes.GetNumElements();
	CDisplayListNode *current = DisplayListNodes.AddElement();

	int meshIndex = object->GetMeshCallbackHandle();
	if (meshIndex >= 0)
	{
		CDisplayListNode *mesh = DisplayListNodes.GetElementRef(meshIndex);
		if (mesh->NormalFaces)
			current->NormalFaces = CreateObjectDisplayList(object, NormalFaces);
		if (mesh->TextureFaces)
			current->TextureFaces = CreateObjectDisplayList(object, TextureFaces);
		if (mesh->AlphaTextureFaces)
			current->AlphaTextureFaces = CreateObjectDisplayList(object, AlphaTextureFaces);
		if (mesh->AlphaFaces)
			current->AlphaFaces = CreateObjectDisplayList(object, AlphaFaces);
	}
	else
	{
		current->NormalFaces = CreateObjectDisplayList(object, NormalFaces);
		current->TextureFaces = CreateObjectDisplayList(object, TextureFaces);
		current->AlphaTextureFaces = CreateObjectDisplayList(object, AlphaTextureFaces);
		current->AlphaFaces = CreateObjectDisplayList(object, AlphaFaces);
	}

	//calculate the size of the object
	double minX, minY, minZ, maxX, maxY, maxZ;
	if (CalculateObjectSize(object, &minX, &minY, &minZ, &maxX, &maxY, &maxZ))
	{
		current->ExtentsIndex = ObjectMinX.GetNumElements();

		//store the size of the object
		ObjectMinX.AddElement(minX);
		ObjectMinY.AddElement(minY);
		ObjectMinZ.AddElement(minZ);
		ObjectMaxX.AddElement(maxX);
		ObjectMaxY.AddElement(maxY);
		ObjectMaxZ.AddElement(maxZ);
	}
	else
	{
		//delete the object
		delete object;
		return -2;
	}

	//delete the object
	delete object;
	return ret;
}

int
CImport::DataCallback(CXGLData* data)
{
	/*
	We ignore all data associated with objects at this time
	*/
	delete data;
	return -1;
}

int 
CImport::TextureCallback(CXGLTexture* texture)
{
	int ret = CreateTexture(texture);
	delete texture;
	return ret;
}

int
GetPowerOf2(int num)
{
	int curr = 1;

	while(curr < num)
	{
		curr = curr<<1;
	}

	return curr;
}

int 
CImport::TextureBufferCallback(CXGLTextureBuffer *textureBuffer)
{
	//resize the texture
	int width = textureBuffer->TextureWidth;
	int height = textureBuffer->TextureHeight;

	int newWidth = GetPowerOf2(width);
	int newHeight = GetPowerOf2(height);
	
	newWidth = __max(newWidth, MIN_TEXTURE_DIM_PIXELS);
	newHeight = __max(newHeight, MIN_TEXTURE_DIM_PIXELS);

	newWidth = __min(newWidth, MAX_TEXTURE_DIM_PIXELS);
	newHeight = __min(newHeight, MAX_TEXTURE_DIM_PIXELS);

	unsigned char *newRGBAImage = new unsigned char[newWidth*newHeight*4];

	ScaleImage(width, height, textureBuffer->RGBABuffer, newWidth, newHeight, newRGBAImage);

	delete[] textureBuffer->RGBABuffer;
	textureBuffer->RGBABuffer = newRGBAImage;
	textureBuffer->TextureWidth = newWidth;
	textureBuffer->TextureHeight = newHeight;

	int ret = CreateTextureBuffer(textureBuffer);
	delete textureBuffer;
	return ret;
}

void 
CImport::ErrorCallback(const char* errorString)
{
	/*
	This function indicates that the xml being parsed was not well formed or 
	that the xgl file does not follow the xgl specifications
	*/

	//Delete any our data structures that have been allocated thus far
	DeletePreviousData();

	//give the user the error....
	AfxMessageBox(errorString);
}

void 
CImport::BackgroundCallback(CXGLBackground* background)
{
	//set the background color
	Background = background;
}

void 
CImport::LightingCallback(CXGLLighting* lighting)
{
	//set the lighting
	Lighting = lighting;
}

void 
CImport::DeletePreviousData()
{
	/*
	This function deletes any data structures allocated 
	when parsing an xgl file.  This function can only be called after
	we have started parsing an xgl file, and should be called after we have 
	created the display list from this data
	*/	

	//uninitialize the resizeable arrays
	ObjectMinX.SetSize(0);
	ObjectMinY.SetSize(0);
	ObjectMinZ.SetSize(0);
	ObjectMaxX.SetSize(0);
	ObjectMaxY.SetSize(0);
	ObjectMaxZ.SetSize(0);
	MeshMinX.SetSize(0);
	MeshMinY.SetSize(0);
	MeshMinZ.SetSize(0);
	MeshMaxX.SetSize(0);
	MeshMaxY.SetSize(0);
	MeshMaxZ.SetSize(0);

	DisplayListNodes.SetSize(0);

	if (Background != NULL)
	{
		delete Background;
		Background = NULL;
	}
	
	if (Lighting != NULL)
	{
		delete Lighting;
		Lighting = NULL;
	}
	
	if (World != NULL)
	{
		delete World;
		World = NULL;
	}
}

void
CImport::CalculateMeshSize(CXGLMesh *mesh, double *meshMinX, double *meshMinY, double *meshMinZ, double *meshMaxX, double *meshMaxY, double *meshMaxZ)
{
	/*
	This function calculates a bounding box for the points of the passed in mesh.  The bounding
	box size is in this mesh's object space
	*/

	double minX = 1E30;
	double minY = 1E30;
	double minZ = 1E30;
	double maxX = -1E30;
	double maxY = -1E30;
	double maxZ = -1E30;

	assert(mesh->NumFaces > 0 || mesh->NumLines > 0 || mesh->NumPoints > 0);
	
	//find the extents of our mesh
	unsigned int f;
	for (f = 0; f < mesh->NumFaces; f++)
	{
		//loop through all of our points finding our maximum and minimum values
		int t;
		for (t = 0; t < 3; t++)
		{
			//the x, y, and z of this vertex
			double x = mesh->Vertices[mesh->Faces[f].FaceVertices[t].VertexI].X;
			double y = mesh->Vertices[mesh->Faces[f].FaceVertices[t].VertexI].Y;
			double z = mesh->Vertices[mesh->Faces[f].FaceVertices[t].VertexI].Z;

			assert(x < 1E30);
			assert(y < 1E30);
			assert(z < 1E30);
			assert(x > -1E30);
			assert(y > -1E30);
			assert(z > -1E30);

			//if it's the biggest or smallest in any dimesion, then store it.
			minX = __min(minX, x);
			minY = __min(minY, y);
			minZ = __min(minZ, z);
			maxX = __max(maxX, x);
			maxY = __max(maxY, y);
			maxZ = __max(maxZ, z);
		}
	}

	if(mesh->NumPoints > NULL)
	{
		for(int i = 0; i < (int) mesh->NumPoints; i++)
		{
			CXGLPoint *point = mesh->Points + i;
			CXGLVertex *vertex = mesh->Vertices + point->PointVertex.VertexI;

			minX = __min(minX, vertex->X);
			minY = __min(minY, vertex->Y);
			minZ = __min(minZ, vertex->Z);
			maxX = __max(maxX, vertex->X);
			maxY = __max(maxY, vertex->Y);
			maxZ = __max(maxZ, vertex->Z);
		}
	}

	if(mesh->NumLines > NULL)
	{
		for(int i = 0; i < (int) mesh->NumLines; i++)
		{
			CXGLLine *line = mesh->Lines + i;

			CXGLVertex *vertex = mesh->Vertices + line->LineVertices[0].VertexI;

			minX = __min(minX, vertex->X);
			minY = __min(minY, vertex->Y);
			minZ = __min(minZ, vertex->Z);
			maxX = __max(maxX, vertex->X);
			maxY = __max(maxY, vertex->Y);
			maxZ = __max(maxZ, vertex->Z);

			vertex = mesh->Vertices + line->LineVertices[1].VertexI;

			minX = __min(minX, vertex->X);
			minY = __min(minY, vertex->Y);
			minZ = __min(minZ, vertex->Z);
			maxX = __max(maxX, vertex->X);
			maxY = __max(maxY, vertex->Y);
			maxZ = __max(maxZ, vertex->Z);
		}
	}

	*meshMinX = minX;
	*meshMinY = minY;
	*meshMinZ = minZ;
	*meshMaxX = maxX;
	*meshMaxY = maxY;
	*meshMaxZ = maxZ;
}

bool
CImport::CalculateObjectSize(CXGLObject *object, double *objectMinX, double *objectMinY, double *objectMinZ, double *objectMaxX, double *objectMaxY, double *objectMaxZ)
{
	/* 
	This function calculates a bounding box for the passed in object.  The bounding
	box's size is in the object's parent space.
	*/

	double minX = 1E30;
	double minY = 1E30;
	double minZ = 1E30;
	double maxX = -1E30;
	double maxY = -1E30;
	double maxZ = -1E30;

	//go through our object children (if any) and find the maximum box of them
	bool areExtentsValid = false;
	unsigned int i;
	for (i = 0; i < object->NumObjects; i++)
	{
		int objectIndex = object->GetObjectCallbackHandle(i);
		if (objectIndex >= 0)
		{
			CDisplayListNode *child = DisplayListNodes.GetElementRef(objectIndex);
			areExtentsValid = true;
			minX = __min(minX, ObjectMinX.GetElement(child->ExtentsIndex));
			minY = __min(minY, ObjectMinY.GetElement(child->ExtentsIndex));
			minZ = __min(minZ, ObjectMinZ.GetElement(child->ExtentsIndex));
			maxX = __max(maxX, ObjectMaxX.GetElement(child->ExtentsIndex));
			maxY = __max(maxY, ObjectMaxY.GetElement(child->ExtentsIndex));
			maxZ = __max(maxZ, ObjectMaxZ.GetElement(child->ExtentsIndex));
		}
	}

	int meshIndex = object->GetMeshCallbackHandle();
	if (meshIndex >= 0)
	{
		CDisplayListNode *mesh = DisplayListNodes.GetElementRef(meshIndex);
		areExtentsValid = true;
		minX = MeshMinX.GetElement(mesh->ExtentsIndex);
		minY = MeshMinY.GetElement(mesh->ExtentsIndex);
		minZ = MeshMinZ.GetElement(mesh->ExtentsIndex);
		maxX = MeshMaxX.GetElement(mesh->ExtentsIndex);
		maxY = MeshMaxY.GetElement(mesh->ExtentsIndex);
		maxZ = MeshMaxZ.GetElement(mesh->ExtentsIndex);
	}

	if (areExtentsValid)
	{
		//transform the mesh extents
		//our object to world transform
		double fx = object->Transform.FX;
		double fy = object->Transform.FY;
		double fz = object->Transform.FZ;
		double ux = object->Transform.UX;
		double uy = object->Transform.UY;
		double uz = object->Transform.UZ;
		double xp = object->Transform.XP;
		double yp = object->Transform.YP;
		double zp = object->Transform.ZP;
		double s = object->Transform.Scale;
		double rx = uy*fz - uz*fy;
		double ry = uz*fx - ux*fz;
		double rz = ux*fy - uy*fx;

		//the object bounding box (untransformed)
		double boxX[8];
		double boxY[8];
		double boxZ[8];
		boxX[0] = minX; boxY[0] = minY; boxZ[0] = minZ;
		boxX[1] = minX; boxY[1] = minY; boxZ[1] = maxZ;
		boxX[2] = minX; boxY[2] = maxY; boxZ[2] = minZ;
		boxX[3] = minX; boxY[3] = maxY; boxZ[3] = maxZ;
		boxX[4] = maxX; boxY[4] = minY; boxZ[4] = maxZ;
		boxX[5] = maxX; boxY[5] = minY; boxZ[5] = minZ;
		boxX[6] = maxX; boxY[6] = maxY; boxZ[6] = maxZ;
		boxX[7] = maxX; boxY[7] = maxY; boxZ[7] = minZ;

		//the object bounding box (transformed to worldspace)
		double tBoxX[8];
		double tBoxY[8];
		double tBoxZ[8];

		//the results
		double objMinX = 1E30;
		double objMinY = 1E30;
		double objMinZ = 1E30;
		double objMaxX = -1E30;
		double objMaxY = -1E30;
		double objMaxZ = -1E30;

		int t;
		for (t = 0; t < 8; t++)
		{
			//transform the box's points one at a time
			tBoxX[t] = (rx*boxX[t] + ux*boxY[t] + fx*boxZ[t])*s + xp;
			tBoxY[t] = (ry*boxX[t] + uy*boxY[t] + fy*boxZ[t])*s + yp;
			tBoxZ[t] = (rz*boxX[t] + uz*boxY[t] + fz*boxZ[t])*s + zp;

			//if the transformed box corner is the smallest or biggest so far store it.
			objMinX = __min(tBoxX[t], objMinX);
			objMinY = __min(tBoxY[t], objMinY);
			objMinZ = __min(tBoxZ[t], objMinZ);
			objMaxX = __max(tBoxX[t], objMaxX);
			objMaxY = __max(tBoxY[t], objMaxY);
			objMaxZ = __max(tBoxZ[t], objMaxZ);
		}
		
		*objectMinX = objMinX;
		*objectMinY = objMinY;
		*objectMinZ = objMinZ;
		*objectMaxX = objMaxX;
		*objectMaxY = objMaxY;
		*objectMaxZ = objMaxZ;

		assert(objMinX > -1E30);
		assert(objMinY > -1E30);
		assert(objMinZ > -1E30);
		assert(objMaxX < 1E30);
		assert(objMaxY < 1E30);
		assert(objMaxZ < 1E30);

		return true;
	}
	else
	{
		return false;
	}	
}

void 
CImport::CalculateFaceCounts(CXGLMesh *mesh, unsigned int *numNormalFaces, unsigned int *numTextureFaces, unsigned int *numAlphaTextureFaces, unsigned int *numAlphaFaces)
{
	*numNormalFaces = 0;
	*numTextureFaces = 0;
	*numAlphaTextureFaces = 0;
	*numAlphaFaces = 0;

	unsigned int i;
	for (i = 0; i < mesh->NumFaces; i++)
	{
		CXGLFace *currentFace = mesh->Faces + i;
		CXGLMaterial *currentMaterial = mesh->Materials + currentFace->MaterialI;

		if (currentMaterial->Alpha != 1.0)
		{
			if (currentFace->TextureI != -1)
			{
				(*numAlphaTextureFaces) ++;
			}
			else
			{
				(*numAlphaFaces) ++;
			}
		}
		else
		{
			if (currentFace->TextureI != -1)
			{
				(*numTextureFaces) ++;
			}
			else
			{
				(*numNormalFaces) ++;
			}
		}
	}
}

void 
CImport::CalculateWorldSize()
{
	/*
	This function calculates a bounding box for all of the objects in the current world.
	This box is used to determine how to set our view frustum. It should be set in a way 
	that enables the whole world to be seen in our window but ensures that it's also not 
	too small in that window as well.
	*/

	//the results
	double worldMinX = 1E30;
	double worldMinY = 1E30;
	double worldMinZ = 1E30;
	double worldMaxX = -1E30;
	double worldMaxY = -1E30;
	double worldMaxZ = -1E30;

	int meshIndex = World->GetMeshCallbackHandle();
	if (meshIndex >= 0)
	{
		CDisplayListNode *mesh = DisplayListNodes.GetElementRef(meshIndex);
		worldMinX = MeshMinX.GetElement(mesh->ExtentsIndex);
		worldMinY = MeshMinY.GetElement(mesh->ExtentsIndex);
		worldMinZ = MeshMinZ.GetElement(mesh->ExtentsIndex);
		worldMaxX = MeshMaxX.GetElement(mesh->ExtentsIndex);
		worldMaxY = MeshMaxY.GetElement(mesh->ExtentsIndex);
		worldMaxZ = MeshMaxZ.GetElement(mesh->ExtentsIndex);
	}
	else
	{
		//find the maximums (or bounding boxes) in object coordinates for all objects in our world
		unsigned int i;
		for (i = 0; i < World->NumObjects; i++)
		{
			int objectIndex = World->GetObjectCallbackHandle(i);
			if (objectIndex >= 0)
			{
				CDisplayListNode *object = DisplayListNodes.GetElementRef(objectIndex);
				worldMinX = __min(worldMinX, ObjectMinX.GetElement(object->ExtentsIndex));
				worldMinY = __min(worldMinY, ObjectMinY.GetElement(object->ExtentsIndex));
				worldMinZ = __min(worldMinZ, ObjectMinZ.GetElement(object->ExtentsIndex));
				worldMaxX = __max(worldMaxX, ObjectMaxX.GetElement(object->ExtentsIndex));
				worldMaxY = __max(worldMaxY, ObjectMaxY.GetElement(object->ExtentsIndex));
				worldMaxZ = __max(worldMaxZ, ObjectMaxZ.GetElement(object->ExtentsIndex));
			}
		}
	}

	//a few size properties of the world bounding box
	double xdiff = worldMaxX - worldMinX;
	double ydiff = worldMaxY - worldMinY;
	double zdiff = worldMaxZ - worldMinZ;
	double xtot = worldMaxX + worldMinX;
	double ytot = worldMaxY + worldMinY;
	double ztot = worldMaxZ + worldMinZ;
	double max = (__max(xdiff, __max(ydiff, zdiff)))*0.5*sqrt(3);

	double xcenter = xtot * 0.5;
	double ycenter = ytot * 0.5;
	double zcenter = ztot * 0.5;

	//the approach is center the world to 0, 0, 0 and then scale our front view plane
	//to a size appropriate for the world size and store the results in our state class

	//translate the world to the center (0, 0, 0);
	XGLState.SetWorldPositionAndExtents(xcenter, ycenter, zcenter, max);

	//look at the world
	XGLState.SetWindowCoords(0, 0, 0, max*2, max*2, max*2);
}

int
CImport::ExtgeomCallback(CXGLExtgeom* extgeom)
{
	/*
	We ignore all data associated with objects at this time
	*/
	delete extgeom;
	return -1;
}
