#include "XGLImport.h"
#include "XGLErrors.h"
#include "XGLStatusCallback.h"
#include "xmlparse.h"
#include "ZLIBUtility/ZLIBDecompress.h"
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <float.h>


#include "XGLImportCallback.h"
#include "XGLTags.h"

class CXGLDataHandle;

//parser state defines
#define STATE_IMPORTING 0
#define STATE_STOPPED 1


//#define TRACE_PARSE

void 
CElementRequires::AddElement(int tagNum)
{
	TagArray.AddElement(tagNum);
}

int
CElementRequires::HowMany(int tagNum)
{

	int count = 0;
	int i;
	int numElements = TagArray.GetNumElements();

	for(i = 0; i < numElements; i++)
	{
		if(TagArray.GetElement(i) == tagNum)
		{
			count++;
		}
	}

	return count;
}


void
CElementRequires::Init()
{
	TagArray.SetExactSize(0);
}


CStateStack::CStateStack()
{
	Top = NULL;
}

CStateStack::~CStateStack()
{
	while(Top != NULL)
	{
		Pop();
	}
}

void
CStateStack::Push(int data)
{
	CStateStackElem* newElem= new CStateStackElem();

	assert(newElem != NULL);

	newElem->Data = data;
	newElem->Next = Top;
	Top = newElem;

}

int
CStateStack::Pop()
{
	assert(Top != NULL);

	//Hitting an access violation here means you didn't init the stack.
	int retVal = Top->Data;
	CStateStackElem* oldTop = Top;
	Top = Top->Next;
	delete(oldTop);

	return(retVal);
}

int 
CStateStack::GetTopValue()
{
	assert(Top != NULL);
	
	//Hitting an access violation here means you didn't init the stack.
	return Top->Data;
}

CObjectStack::CObjectStack()
{
	Top = NULL;
}

void 
CObjectStack::Push(CXGLObject *object)
{
	CObjectStackElem* newElem = new CObjectStackElem();
	newElem->Object = object;
	newElem->Next = Top;
	Top = newElem;
}

CObjectStackElem*
CObjectStack::GetTop()
{
	return Top;
}

CXGLObject*
CObjectStack::Pop()
{
	assert(Top != NULL);

	CXGLObject *ret = Top->Object;
	
	CObjectStackElem* oldTop = Top;
	Top = Top->Next;
	delete oldTop;

	return ret;
}

CObjectStack::~CObjectStack()
{
	while(Top != NULL)
	{
		CXGLObject* obj = Pop();
		delete obj;
	}
}

void
CXGLImport::Error(const char* errorString)
{
	if(State != STATE_STOPPED)
	{
		Resolver.AddError(errorString);
		StopImporting();
	}
}


int
CXGLImport::SetBit(unsigned int bitNum, unsigned int* num)
{
	if(bitNum > 31)
	{
		return 0;
	}

	unsigned int newBit = 1;

	newBit = newBit << (bitNum - 1);

	*num = *num | newBit;

	return 1;
}

CXGLImport::CXGLImport()
{
	Init();
}

CXGLImport::~CXGLImport()
{
	UnInit();
}

void 
CXGLImport::Init()
{
	State = STATE_STOPPED;

	CurrentWorld = NULL;
	CurrentLighting = NULL;
	CurrentBackground = NULL;
	CurrentData = NULL;
	CurrentMesh = NULL;
	CurrentTexture = NULL;
	CurrentContext = NULL;

	NeedsData = 0;

	Parser = NULL;
}

void 
CXGLImport::UnInit()
{
	if(CurrentWorld != NULL)
	{
		delete CurrentWorld;
		CurrentWorld = NULL;
	}

	if(CurrentLighting != NULL)
	{
		delete CurrentLighting;
		CurrentLighting = NULL;
	}

	if(CurrentBackground != NULL)
	{
		delete CurrentBackground;
		CurrentBackground = NULL;
	}

	if(CurrentData != NULL)
	{
		delete CurrentData;
		CurrentData = NULL;
	}

	if(CurrentMesh != NULL)
	{
		delete CurrentMesh;
		CurrentMesh = NULL;
	}

	if(CurrentTexture != NULL)
	{
		delete CurrentTexture;
		CurrentTexture = NULL;
	}

	while(CurrentContext != NULL)
	{
		CurrentContext = Resolver.ReleaseContext(CurrentContext);
	}

	int numPatches = Patches.GetNumElements();

	int i;

	for(i = 0; i < numPatches; i++)
	{
		CXGLPatch* patch = Patches.GetElement(i);		
		delete patch;
	}

	Patches.SetSize(0);

//	delete[] CopyString;
	CopyString.SetExactSize(0);

	Resolver.UnInit();
}

void
CXGLImport::InitCallback(CXGLImportCallback* callback)
{
	Resolver.Init(callback);
}

int
CXGLImport::GetTag(const char* inputTagName)
{
	int tag;

	if (inputTagName == NULL)
	{
		assert(0);
		return -1;
	}

	if (strlen(inputTagName) >= 256)
		return -1;

	//make the inputTagName lowercase
	char tagName[256];
	strcpy(tagName, inputTagName);
	strlwr(tagName);
	

	if(strcmp(tagName, VERTEX_REF) == 0)
	{
		tag = VertexRefTag;
	}
	else if(strcmp(tagName, NORMAL_REF) == 0)
	{
		tag = NormalRefTag;
	}
	else if(strcmp(tagName, VERTEX_POSITION) == 0)
	{
		tag = VertexPositionTag;
	}
	else if(strcmp(tagName, NORMAL) == 0)
	{
		tag = NormalTag;
	}
	else if(strcmp(tagName, FACE_VERTEX1) == 0)
	{
		tag = FaceVertexTag1;
	}
	else if(strcmp(tagName, FACE_VERTEX2) == 0)
	{
		tag = FaceVertexTag2;
	}
	else if(strcmp(tagName, FACE_VERTEX3) == 0)
	{
		tag = FaceVertexTag3;
	}
	else if(strcmp(tagName, LINE_VERTEX1) == 0)
	{
		tag = LineVertexTag1;
	}
	else if(strcmp(tagName, LINE_VERTEX2) == 0)
	{
		tag = LineVertexTag2;
	}
	else if(strcmp(tagName, POINT_VERTEX1) == 0)
	{
		tag = PointVertexTag1;
	}
	else if(strcmp(tagName, SMOOTHING_GROUP) == 0)
	{
		tag = SmoothingGroupTag;
	}
	else if(strcmp(tagName, TEXTURECOORDINATE_REF) == 0)
	{
		tag = TextureCoordinateRefTag;
	}
	else if(strcmp(tagName, TEXTURECOORDINATE) == 0)
	{
		tag = TextureCoordinateTag;
	}
	else if(strcmp(tagName, MATERIAL_REF) == 0)
	{
		tag = MaterialRefTag;
	}
	else if(strcmp(tagName, FACE) == 0)
	{
		tag = FaceTag;
	}
	else if(strcmp(tagName, LINE) == 0)
	{
		tag = LineTag;
	}
	else if(strcmp(tagName, POINT) == 0)
	{
		tag = PointTag;
	}
	else if(strcmp(tagName, MATERIAL_EMISSIVE) == 0)
	{
		tag = MaterialEmissiveTag;
	}	
	else if(strcmp(tagName, MATERIAL_DIFFUSE) == 0)
	{
		tag = MaterialDiffuseTag;
	}
	else if(strcmp(tagName, MATERIAL_SPECULAR) == 0)
	{
		tag = MaterialSpecularTag;
	}
	else if(strcmp(tagName, MATERIAL_AMBIENT) == 0)
	{
		tag = MaterialAmbientTag;
	}
	else if(strcmp(tagName, MATERIAL_SHININESS) == 0)
	{
		tag = MaterialShininessTag;
	}
	else if(strcmp(tagName, MATERIAL_ALPHA) == 0)
	{
		tag = MaterialAlphaTag;
	}
	else if(strcmp(tagName, MATERIAL) == 0)
	{
		tag = MaterialTag;
	}
	else if(strcmp(tagName, PATCH) == 0)
	{
		tag = PatchTag;
	}
	else if(strcmp(tagName, TEXTURERGBA_REF) == 0)
	{
		tag = TextureRGBARefTag;
	}
	else if(strcmp(tagName, TEXTURERGB_REF) == 0)
	{
		tag = TextureRGBRefTag;
	}
	else if(strcmp(tagName, TEXTURERGBA) == 0)
	{
		tag = TextureRGBATag;
	}
	else if(strcmp(tagName, TEXTURERGB) == 0)
	{
		tag = TextureRGBTag;
	}
	else if(strcmp(tagName, REPLACE) == 0)
	{
		tag = ReplaceTag;
	}
	else if(strcmp(tagName, MODULATE) == 0)
	{
		tag = ModulateTag;
	}
	else if(strcmp(tagName, DECAL) == 0)
	{
		tag = DecalTag;
	}
	else if(strcmp(tagName, REPEAT) == 0)
	{
		tag = RepeatTag;
	}
	else if(strcmp(tagName, CLAMP) == 0)
	{
		tag = ClampTag;
	}
	else if(strcmp(tagName, TEXTUREBORDERCOLOR) == 0)
	{
		tag = TextureBorderColorTag;
	}
	else if(strcmp(tagName, TEXTURE_REF) == 0)
	{
		tag = TextureRefTag;
	}
	else if(strcmp(tagName, TEXTURE) == 0)
	{
		tag = TextureTag;
	}
	else if(strcmp(tagName, LINESTYLE) == 0)
	{
		tag = LineStyleTag;
	}
	else if(strcmp(tagName, LINESTYLE_REF) == 0)
	{
		tag = LineStyleRefTag;
	}
	else if(strcmp(tagName, LINESTYLE_WIDTH) == 0)
	{
		tag = LineStyleWidthTag;
	}
	else if(strcmp(tagName, LINESTYLE_PATTERN) == 0)
	{
		tag = LineStylePatternTag;
	}
	else if(strcmp(tagName, LINESTYLE_PATTERNFACTOR) == 0)
	{
		tag = LineStylePatternFactorTag;
	}
	else if(strcmp(tagName, POINTSTYLE) == 0)
	{
		tag = PointStyleTag;
	}
	else if(strcmp(tagName, POINTSTYLE_REF) == 0)
	{
		tag = PointStyleRefTag;
	}
	else if(strcmp(tagName, POINTSTYLE_SIZE) == 0)
	{
		tag = PointStyleSizeTag;
	}
	else if(strcmp(tagName, MESH_REF) == 0)
	{
		tag = MeshRefTag;
	}
	else if(strcmp(tagName, OBJECT_REF) == 0)
	{
		tag = ObjectRefTag;
	}
	else if(strcmp(tagName, MESH) == 0)
	{
		tag = MeshTag;
	}
	else if(strcmp(tagName, SURFACE) == 0)
	{
		tag = SurfaceTag;
	}
	else if(strcmp(tagName, TRANSFORM) == 0)
	{
		tag = TransformTag;
	}
	else if(strcmp(tagName, FORWARD) == 0)
	{
		tag = TransformForwardTag;
	}
	else if(strcmp(tagName, UP) == 0)
	{
		tag = TransformUpTag;
	}
	else if(strcmp(tagName, TRANS_POSITION) == 0)
	{
		tag = TransformPositionTag;
	}
	else if(strcmp(tagName, SCALE) == 0)
	{
		tag = TransformScaleTag;
	}
	else if(strcmp(tagName, DATA) == 0)
	{
		tag = DataTag;
	}
	else if(strcmp(tagName, STRING) == 0)
	{
		tag = StringTag;
	}
	else if(strcmp(tagName, BIN) == 0)
	{
		tag = BinTag;
	}
	else if(strcmp(tagName, NAME) == 0)
	{
		tag = NameTag;
	}
	else if(strcmp(tagName, OBJECT) == 0)
	{
		tag = ObjectTag;
	}
	else if(strcmp(tagName, LIGHTING) == 0)
	{
		tag = LightingTag;
	}
	else if(strcmp(tagName, LIGHT_DIFFUSE) == 0)
	{
		tag = LightDiffuseTag;
	}
	else if(strcmp(tagName, LIGHT_SPECULAR) == 0)
	{
		tag = LightSpecularTag;
	}
	else if(strcmp(tagName, LIGHT_AMBIENT) == 0)
	{
		tag = LightAmbientTag;
	}
	else if(strcmp(tagName, LIGHT_DIRECTION) == 0)
	{
		tag = LightDirectionTag;
	}
	else if(strcmp(tagName, DIRECTIONAL_LIGHT) == 0)
	{
		tag = DirectionalLightTag;
	}
	else if(strcmp(tagName, BACKGROUND) == 0)
	{
		tag = BackgroundTag;
	}
	else if(strcmp(tagName, BACKGROUND_COLOR) == 0)
	{
		tag = BackgroundColorTag;
	}
	else if(strcmp(tagName, EXTERNAL_OBJECT_REF) == 0)
	{
		tag = ExternalObjectRefTag;
	}
	else if(strcmp(tagName, EXTERNAL_OBJECT_STATIC_REF) == 0)
	{
		tag = ExternalObjectStaticRefTag;
	}
	else if(strcmp(tagName, WORLD) == 0)
	{
		tag = WorldTag;
	}
	else if(strcmp(tagName, REF) == 0)
	{
		tag = RefTag;
	}
	else if(strcmp(tagName, EXTENTS) == 0)
	{
		tag = ExtentsTag;
	}
	else if(strcmp(tagName, REF_TYPE) == 0)
	{
		tag = RefTypeTag;
	}
	else if(strcmp(tagName, EXTGEOM) == 0)
	{
		tag = ExtGeomTag;
	}
	else if(strcmp(tagName, EXTGEOMELEM) == 0)
	{
		tag = ExtGeomElemTag;
	}
	else if(strcmp(tagName, EXTGEOMELEMDATA) == 0)
	{
		tag = ExtGeomElemDataTag;
	}
	else if(strcmp(tagName, EXTGEOMELEMEXTENTS) == 0)
	{
		tag = ExtGeomElemExtentsTag;
	}
	else if(strcmp(tagName, EXTGEOMELEMPRIORITY) == 0)
	{
		tag = ExtGeomElemPriorityTag;
	}
	else 
	{
		tag = OtherTag;
	}

	return tag;
}

extern "C" 
void 
StartElement(void *userData, const char *name, const char **atts)
{
	CXGLImport* importer = (CXGLImport*) userData;

	importer->ImportStartElement(name, atts);

}

void
CXGLImport::ImportStartElement(const char* name, const char** atts)
{

	int tag = GetTag(name);

#ifdef TRACE_PARSE
	CString s;
	s.Format("%d: %d\r\n", CurrentState, tag);
	OutputDebugString(s);
#endif	
	
	if(tag == -1)
	{
		Error(TAG_NAME_TOO_LONG);
		return;
	}

	CurrentParserState = OtherState;

	if(IsStopped())
	{
		return;
	}
	
	UpdateStatus();

	switch(CurrentState)
	{
	case FaceVertexState:
		{
			FaceVertexStateBeginTagHandler(tag, atts);
			break;
		}
	case FaceState:
		{
			FaceStateBeginTagHandler(tag, atts);
			break;
		}
	case LineState:
		{
			LineStateBeginTagHandler(tag, atts);
			break;
		}
	case PointState:
		{
			PointStateBeginTagHandler(tag, atts);
			break;
		}
	case MaterialState:
		{
			MaterialStateBeginTagHandler(tag, atts);
			break;
		}
	case LineStyleState:
		{
			LineStyleStateBeginTagHandler(tag, atts);
			break;
		}
	case PointStyleState:
		{
			PointStyleStateBeginTagHandler(tag, atts);
			break;
		}
	case MeshState:
		{	
			MeshStateBeginTagHandler(tag, atts);
			break;
		}
	case TransformState:
		{
			TransformStateBeginTagHandler(tag, atts);
			break;
		}
	case DataState:
		{
			DataStateBeginTagHandler(tag, atts);
			break;
		}
	case ObjectState:
		{
			ObjectStateBeginTagHandler(tag, atts);
			break;
		}
	case LightingState:
		{
			LightingStateBeginTagHandler(tag, atts);
			break;
		}
	case BackgroundState:
		{
			BackgroundStateBeginTagHandler(tag, atts);	
			break;
		}
	case WorldState:
		{
			WorldStateBeginTagHandler(tag, atts);
			break;
		}
	case SpaceState:
		{
			SpaceStateBeginTagHandler(tag, atts);
			break;
		}
	case TextureState:
		{
			TextureStateBeginTagHandler(tag, atts);
			break;
		}
	case ExternalObjectReferenceState:
		{
			ExternalObjectStateBeginTagHandler(tag, atts);
			break;
		}
	case ExtgeomState:
		{
			ExtgeomStateBeginTagHandler(tag, atts);
			break;
		}
	case OtherState:
		{
			OtherStateBeginTagHandler(tag, atts);
			break;
		}
	}
}


extern "C" 
void 
EndElement(void *userData, const char *name)
{
	CXGLImport* importer = (CXGLImport*) userData;

	importer->ImportEndElement(name);

	return;
}

void
CXGLImport::ImportEndElement(const char* name)
{	
	int tag = GetTag(name);

	if(CurrentParserState == CharDataState)
	{
		AccumCharElement();
	}

#ifdef TRACE_PARSE
	CString s;
	s.Format("%d: %d\r\n", CurrentState, tag);
	OutputDebugString(s);
#endif
	
	CurrentParserState = OtherState;

	if(IsStopped())
	{
		return;
	}

	switch(CurrentState)
	{
	case FaceVertexState:
		{
			FaceVertexStateEndTagHandler(tag);
			break;
		}
	case FaceState:
		{
			FaceStateEndTagHandler(tag);
			break;
		}
	case LineState:
		{
			LineStateEndTagHandler(tag);
			break;
		}
	case PointState:
		{
			PointStateEndTagHandler(tag);
			break;
		}
	case MaterialState:
		{
			MaterialStateEndTagHandler(tag);
			break;
		}
	case LineStyleState:
		{
			LineStyleStateEndTagHandler(tag);
			break;
		}
	case PointStyleState:
		{
			PointStyleStateEndTagHandler(tag);
			break;
		}
	case MeshState:
		{	
			MeshStateEndTagHandler(tag);
			break;
		}
	case TransformState:
		{
			TransformStateEndTagHandler(tag);
			break;
		}
	case DataState:
		{
			DataStateEndTagHandler(tag);
			break;
		}
	case ObjectState:
		{
			ObjectStateEndTagHandler(tag);
			break;
		}
	case LightingState:
		{
			LightingStateEndTagHandler(tag);
			break;
		}
	case BackgroundState:
		{
			BackgroundStateEndTagHandler(tag);
			break;
		}
	case WorldState:
		{
			WorldStateEndTagHandler(tag);
			break;
		}
	case SpaceState:
		{
			SpaceStateEndTagHandler(tag);
			break;
		}
	case TextureState:
		{
			TextureStateEndTagHandler(tag);
			break;
		}
	case ExternalObjectReferenceState:
		{
			ExternalObjectStateEndTagHandler(tag);
			break;
		}
	case ExtgeomState:
		{
			ExtgeomStateEndTagHandler(tag);
			break;
		}
	case OtherState:
		{
			OtherStateEndTagHandler(tag);
			break;
		}
	}
}

void
CXGLImport::UpdateStatus()
{
	if(Status != NULL)
	{
		if (Parser != NULL)
		{
			CurrentByte = XML_GetCurrentByteIndex(Parser);
			Status->SetProgress(CurrentByte, NumBytesInFile);
		}
	}

}


void
CXGLImport::SpaceStateBeginTagHandler(int tag, const char**atts)
{
	switch(tag)
	{
	case WorldTag:
		{
			//Our CurrentState is now within the world
			StateStack.Push(CurrentState);
			CurrentState = WorldState;

			//create the world context
			CurrentContext = Resolver.CreateContext(NULL);

			WorldRequires.Init();

			//initialize the currentworld
			InitializeCurrentWorld();
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void
CXGLImport::SpaceStateEndTagHandler(int tag)
{
	//this shouldn't ever be called
	assert(0);
}

void
CXGLImport::WorldStateBeginTagHandler(int tag, const char** atts)
{
	WorldRequires.AddElement(tag);

	switch(tag)
	{
	case ObjectTag:
		{	
			//Our CurrentState is now within the Object
			StateStack.Push(CurrentState);
			CurrentState = ObjectState;

			//create a child context and make it our current context
			CurrentContext = Resolver.CreateContext(CurrentContext);

			//reset what type of object data we're expecting
			CurrentObjectState = OtherObjectState;

			//find the path id (if any for this object)
			int pathID = -1;
			int xglID = -1;
			GetDefinitionAndPathID(&xglID, &pathID, atts);
			InitializeCurrentObject(pathID, xglID);

			CurrentWorldState = OtherWorldState;

			break;
		}
	case ExternalObjectRefTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = ExternalObjectReferenceState;

			CurrentContext = Resolver.CreateContext(CurrentContext);

			int pathID = GetPathID(atts);

			InitializeCurrentObjectExternal(pathID, 0);
			
			CurrentWorldState = OtherWorldState;

			break;
		}
	case ExternalObjectStaticRefTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = ExternalObjectReferenceState;

			CurrentContext = Resolver.CreateContext(CurrentContext);

			int pathID = GetPathID(atts);

			InitializeCurrentObjectExternal(pathID, 1);

			CurrentWorldState = OtherWorldState;
			break;
		}
	case ObjectRefTag:
		{
			NeedsData = 1;
			
			//indicate that we're prepared for an object reference
			CurrentWorldState = WorldObjectRefState;			
			break;
		}
	case DataTag:
		{
			DataRequires.Init();
			
			char **currentAtt = (char **)atts;
			if (*currentAtt == NULL)
			{
				Error(DATA_ERROR2);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, ORGANIZATION) != 0)
				{
					Error(DATA_ERROR1);
				}
				else
				{
					//Our CurrentState is now within the Data
					StateStack.Push(CurrentState);
					CurrentState = DataState;

					//Initialize the Data that we're about to write into
					InitializeCurrentData(*(currentAtt + 1));
				}

				currentAtt += 2;
			}

			CurrentWorldState = OtherWorldState;
			break;
		}
	case NameTag:
		{
			//indicate that we're prepared for a world name
			CurrentWorldState = WorldNameState;
			break;
		}
	case LightingTag:
		{
			LightingRequires.Init();
			
			//Our CurrentState is within the Lighting
			StateStack.Push(CurrentState);
			CurrentState = LightingState;
			
			//initialize the lighting we're about to write into
			InitializeCurrentLighting();
	
			CurrentWorldState = OtherWorldState;

			break;
		}
	case BackgroundTag:
		{
			BackgroundRequires.Init();
			
			//Our CurrentState is within the Background
			StateStack.Push(CurrentState);
			CurrentState = BackgroundState;

			//initialize the background we're about to write into
			InitializeCurrentBackground();

			CurrentWorldState = OtherWorldState;

			break;
		}
	case MeshTag:
		{
			//Our CurrentState is now within the mesh
			StateStack.Push(CurrentState);
			CurrentState = MeshState;

			//create a new context
			CurrentContext = Resolver.CreateContext(CurrentContext);

			//find the id for this mesh
			CurrentMeshID = GetDefinitionID(atts);

			//initialize this beast
			InitializeCurrentMesh();

			CurrentWorldState = OtherWorldState;
			break;
		}
	case MeshRefTag:
		{
			NeedsData = 1;
			CurrentWorldState = WorldMeshRefState;

			break;
		}
	case MaterialTag:
		{
			MaterialRequires.Init();
			
			//Our CurrentState is now within the Material
			StateStack.Push(CurrentState);
			CurrentState = MaterialState;
			
			//find the id for this material
			CurrentMaterialID = GetDefinitionID(atts);
			if (CurrentMaterialID == -1)
			{
				Error(MAT_ID_ERROR1);
			}
			else
			{
				InitializeCurrentMaterial();
			}

			CurrentWorldState = OtherWorldState;
			break;
		}
	case LineStyleTag:
		{
			LineStyleRequires.Init();

			//Our CurrentState is now within the LineStyle
			StateStack.Push(CurrentState);
			CurrentState = LineStyleState;

			//find the id for this linestyle
			CurrentLineStyleID = GetDefinitionID(atts);
			if (CurrentLineStyleID == -1)
			{
				Error(LINESTYLE_ID_ERROR1);
			}
			else
			{
				InitializeCurrentLineStyle();
			}

			CurrentWorldState = OtherWorldState;
			break;
		}
	case PointStyleTag:
		{
			PointStyleRequires.Init();

			//Our CurrentState is now within the LineStyle
			StateStack.Push(CurrentState);
			CurrentState = PointStyleState;

			//find the id for this linestyle
			CurrentPointStyleID = GetDefinitionID(atts);
			if (CurrentPointStyleID == -1)
			{
				Error(POINTSTYLE_ID_ERROR1);
			}
			else
			{
				InitializeCurrentPointStyle();
			}

			CurrentWorldState = OtherWorldState;
			break;
		}
	case TextureRGBTag:
		{
			NeedsData = 1;
			
			CurrentWorldState = WorldTextureRGBState;

			//find the width, height and id for this textureBuffer (if any)
			int height = -1;
			int width = -1;
			int xglID = -1;
			GetWidthHeightAndDefinitionID(&width, &height, &xglID, atts);

			//texture buffer didn't have a width
			if (width == -1)
			{
				Error(TEXTUREBUFFER_ERROR2);
			}

			//texture buffer didn't have a height
			if (height == -1)
			{
				Error(TEXTUREBUFFER_ERROR3);
			}

			//texture buffer was defined inside a texture
			if (xglID == -1)
			{
				Error(TEXTUREBUFFER_ERROR4);
			}
			else
			{
				CurrentTextureBufferID = xglID;

				//initialize the texture buffer we're going to write into
				InitializeCurrentTextureBuffer();
				CurrentTextureBuffer->TextureWidth = width;
				CurrentTextureBuffer->TextureHeight = height;
				CurrentTextureBuffer->TextureBufferFormat = TextureRGB;
			}

			break;
		}
	case TextureRGBATag:
		{
			NeedsData = 1;

			CurrentWorldState = WorldTextureRGBAState;

			//find the width, height and id for this textureBuffer (if any)
			int height = -1;
			int width = -1;
			int xglID = -1;
			GetWidthHeightAndDefinitionID(&width, &height, &xglID, atts);

			//texture buffer didn't have a width
			if (width == -1)
			{
				Error(TEXTUREBUFFER_ERROR2);
			}

			//texture buffer didn't have a height
			if (height == -1)
			{
				Error(TEXTUREBUFFER_ERROR3);
			}

			//texture buffer was defined inside a texture
			if (xglID == -1)
			{
				Error(TEXTUREBUFFER_ERROR4);
			}
			else
			{
				CurrentTextureBufferID = xglID;

				//initialize the texture buffer we're going to write into
				InitializeCurrentTextureBuffer();
				CurrentTextureBuffer->TextureWidth = width;
				CurrentTextureBuffer->TextureHeight = height;
				CurrentTextureBuffer->TextureBufferFormat = TextureRGBA;
			}

			break;
		}
	case TextureTag:
		{
			TextureRequires.Init();
			
			//Our CurrentState is within the Texture
			StateStack.Push(CurrentState);
			CurrentState = TextureState;

			//find the id for this texture
			CurrentTextureID = GetDefinitionID(atts);
			if (CurrentTextureID == -1)
			{
				Error(TEXTURE_ID_ERROR3);
			}
			else
			{
				InitializeCurrentTexture();
			}
	
			CurrentWorldState = OtherWorldState;
			break;
		}
	case TextureCoordinateTag:
		{
			//the next piece of data should be a texture coordinate
			CurrentWorldState = WorldTextureCoordinateState;

			//find the id from the attribute
			CurrentTextureCoordinateID = GetDefinitionID(atts);
			if (CurrentTextureCoordinateID == -1)
			{
				Error(TEXTURECOORDINATE_ID_ERROR3);
			}

			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentWorldState = OtherWorldState;
			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void
CXGLImport::WorldStateEndTagHandler(int tag)
{

	if(NeedsData)
	{
		Error(WORLD_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case WorldTag:
		{			

			if(WorldRequires.HowMany(LightingTag) != 1)
			{
				Error(WORLD_LIGHTING_ERROR);
				return;
			}
			
			if(WorldRequires.HowMany(BackgroundTag) != 1)
			{
				Error(WORLD_BACKGROUND_ERROR);
				return;
			}

			//Our current state should be the space state
			CurrentState = StateStack.Pop();			
			
			assert(CurrentState == SpaceState);

			if (VerifyCurrentWorldResolved() == 0)
				return;

			Resolver.AddWorld(CurrentContext, CurrentWorld);

			//release the top context
			CurrentWorld = NULL;
			Resolver.ReleaseContext(CurrentContext);
			CurrentContext = NULL;
			break;
		}
	case TextureRGBTag:
		{
			assert(CurrentTextureBufferID != -1);
			if (CurrentTextureBufferID != -1)
			{
				Resolver.AddTextureBufferDefinition(CurrentContext, CurrentTextureBufferID, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
				
			CurrentTextureBufferID = -1;
			
			break;
		}
	case TextureRGBATag:
		{
			assert(CurrentTextureBufferID != -1);
			if (CurrentTextureBufferID != -1)
			{
				Resolver.AddTextureBufferDefinition(CurrentContext, CurrentTextureBufferID, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
			
			CurrentTextureBufferID = -1;
			break;
		}
	case NameTag:
		{
			//if we have an empty name tag put in an empty string
			if (CurrentWorld->Name == NULL)
			{
				CurrentWorld->Name = new char[strlen("") + 1];
				strcpy(CurrentWorld->Name, "");
			}

			break;
		}
	default:
		{
			break;
		}
	}

	CurrentWorldState = OtherWorldState;
}

void
CXGLImport::ObjectStateBeginTagHandler(int tag, const char** atts)
{
	ObjectStack.GetTop()->ObjectRequires.AddElement(tag);

	switch(tag)
	{
	case TransformTag:
		{
			//Our CurrentState is now within the Transform
			StateStack.Push(CurrentState);
			CurrentState = TransformState;

			TransformRequires.Init();

			CurrentObjectState = OtherObjectState;
			
			break;
		}
	case MeshTag:
		{
			//Our CurrentState is now within the Mesh
			StateStack.Push(CurrentState);
			CurrentState = MeshState;
			
			//create a new context
			CurrentContext = Resolver.CreateContext(CurrentContext);

			//find the id for this mesh (if any)
			CurrentMeshID = GetDefinitionID(atts);

			//initialize the mesh we're going to write into
			InitializeCurrentMesh();

			CurrentObjectState = OtherObjectState;
			break;
		}
	case MeshRefTag:
		{
			NeedsData = 1;
			//indicate that we're prepared for a mesh reference
			CurrentObjectState = ObjectMeshRefState;

			break;
		}
	case ObjectTag:
		{
			//Our CurrentState is now within the Object
			StateStack.Push(CurrentState);
			CurrentState = ObjectState;

			//create a child context and make it our current context
			CurrentContext = Resolver.CreateContext(CurrentContext);

			//reset what type of object data we're expecting
			CurrentObjectState = OtherObjectState;

			//find the path id (if any for this object)
			int pathID = -1;
			int xglID = -1;
			GetDefinitionAndPathID(&xglID, &pathID, atts);
			InitializeCurrentObject(pathID, xglID);

			break;
		}
	case ExternalObjectRefTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = ExternalObjectReferenceState;

			CurrentContext = Resolver.CreateContext(CurrentContext);

			int pathID = GetPathID(atts);

			InitializeCurrentObjectExternal(pathID, 0);
			
			CurrentWorldState = OtherWorldState;
			break;
		}
	case ExternalObjectStaticRefTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = ExternalObjectReferenceState;

			CurrentContext = Resolver.CreateContext(CurrentContext);

			int pathID = GetPathID(atts);

			InitializeCurrentObjectExternal(pathID, 1);
			
			CurrentWorldState = OtherWorldState;
			break;
		}
	case ObjectRefTag:
		{
			//indicate that we're prepared for an object reference
			NeedsData = 1;

			CurrentObjectState = ObjectObjectRefState;			
			break;
		}
	case NameTag:
		{
			CurrentObjectState = ObjectNameState;			
			break;
		}
	case DataTag:
		{
			DataRequires.Init();
			
			char **currentAtt = (char **)atts;
			if (*currentAtt == NULL)
			{
				Error(DATA_ERROR2);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, ORGANIZATION) != 0)
				{
					Error(DATA_ERROR1);
				}
				else
				{
					//Our CurrentState is now within the Data
					StateStack.Push(CurrentState);
					CurrentState = DataState;

					//Initialize the Data that we're about to write into
					InitializeCurrentData(*(currentAtt + 1));
				}

				currentAtt += 2;
			}

			CurrentObjectState = OtherObjectState;
			break;
		}
	case MaterialTag:
		{
			MaterialRequires.Init();
			
			//Our CurrentState is now within the Material
			StateStack.Push(CurrentState);
			CurrentState = MaterialState;
			
			//find the id for this material
			CurrentMaterialID = GetDefinitionID(atts);
			if (CurrentMaterialID == -1)
			{				
				Error(MAT_ID_ERROR2);
			}
			else
			{
				InitializeCurrentMaterial();
			}

			CurrentObjectState = OtherObjectState;
			break;
		}
	case LineStyleTag:
		{
			LineStyleRequires.Init();

			//Our CurrentState is now within the LineStyle
			StateStack.Push(CurrentState);
			CurrentState = LineStyleState;

			//find the id for this linestyle
			CurrentLineStyleID = GetDefinitionID(atts);
			if (CurrentLineStyleID == -1)
			{
				Error(LINESTYLE_ID_ERROR2);
			}
			else
			{
				InitializeCurrentLineStyle();
			}

			CurrentObjectState = OtherObjectState;
			break;
		}
	case PointStyleTag:
		{
			PointStyleRequires.Init();

			//Our CurrentState is now within the LineStyle
			StateStack.Push(CurrentState);
			CurrentState = PointStyleState;

			//find the id for this linestyle
			CurrentPointStyleID = GetDefinitionID(atts);
			if (CurrentPointStyleID == -1)
			{
				Error(POINTSTYLE_ID_ERROR2);
			}
			else
			{
				InitializeCurrentPointStyle();
			}

			CurrentObjectState = OtherObjectState;
			break;
		}
	case TextureTag:
		{
			TextureRequires.Init();

			//got a texture at the object level
			//Our CurrentState is within the Texture
			
			StateStack.Push(CurrentState);
			CurrentState = TextureState;

			//find the id for this texture
			CurrentTextureID = GetDefinitionID(atts);
			if (CurrentTextureID == -1)
			{
				Error(TEXTURE_ID_ERROR2);
			}
			else
			{
				InitializeCurrentTexture();
			}
	
			CurrentObjectState = OtherObjectState;
			
			break;
		}
	case TextureRGBTag:
		{
			NeedsData = 1;
			
			CurrentObjectState = ObjectTextureRGBState;

			//find the width, height and id for this textureBuffer (if any)
			int height = -1;
			int width = -1;
			int xglID = -1;
			GetWidthHeightAndDefinitionID(&width, &height, &xglID, atts);

			//texture buffer didn't have a width
			if (width == -1)
			{
				Error(TEXTUREBUFFER_ERROR2);
			}

			//texture buffer didn't have a height
			if (height == -1)
			{
				Error(TEXTUREBUFFER_ERROR3);
			}

			//texture buffer was defined inside a texture
			if (xglID == -1)
			{
				Error(TEXTUREBUFFER_ERROR6);
			}
			else
			{
				CurrentTextureBufferID = xglID;

				//initialize the texture buffer we're going to write into
				InitializeCurrentTextureBuffer();
				CurrentTextureBuffer->TextureWidth = width;
				CurrentTextureBuffer->TextureHeight = height;
				CurrentTextureBuffer->TextureBufferFormat = TextureRGB;
			}

			break;
		}
	case TextureRGBATag:
		{
			NeedsData = 1;
			
			CurrentObjectState = ObjectTextureRGBAState;

			//find the width, height and id for this textureBuffer (if any)
			int height = -1;
			int width = -1;
			int xglID = -1;
			GetWidthHeightAndDefinitionID(&width, &height, &xglID, atts);

			//texture buffer didn't have a width
			if (width == -1)
			{
				Error(TEXTUREBUFFER_ERROR2);
			}

			//texture buffer didn't have a height
			if (height == -1)
			{
				Error(TEXTUREBUFFER_ERROR3);
			}

			//texture buffer was defined inside a texture
			if (xglID == -1)
			{
				Error(TEXTUREBUFFER_ERROR6);
			}
			else
			{
				CurrentTextureBufferID = xglID;

				//initialize the texture buffer we're going to write into
				InitializeCurrentTextureBuffer();
				CurrentTextureBuffer->TextureWidth = width;
				CurrentTextureBuffer->TextureHeight = height;
				CurrentTextureBuffer->TextureBufferFormat = TextureRGBA;
			}

			break;
		}
	case TextureCoordinateTag:
		{
			NeedsData = 1;
			
			//the next piece of data should be a texture coordinate
			CurrentObjectState = ObjectTextureCoordinateState;

			//find the id from the attribute
			CurrentTextureCoordinateID = GetDefinitionID(atts);
			if (CurrentTextureCoordinateID == -1)
			{
				Error(TEXTURECOORDINATE_ID_ERROR2);
			}

			break;
		}
	case ExtGeomTag:
		{
			char **currentAtt = (char **)atts;
			if (*currentAtt == NULL)
			{
				Error(EXTGEOM_NO_ORG_ATTRIBUTE);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, ORGANIZATION) != 0)
				{
					Error(EXTGEOM_ATTR_NOT_CALLED_ORG);
				}
				else
				{
					//Our CurrentState is now within the Extgeom
					StateStack.Push(CurrentState);
					CurrentState = ExtgeomState;
					//Initialize the Extgeom that we're about to write into
					InitializeCurrentExtgeom(*(currentAtt + 1));
				}

				currentAtt += 2;
			}

			CurrentObjectState = OtherObjectState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentObjectState = OtherObjectState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}


void
CXGLImport::ExternalObjectStateBeginTagHandler(int tag, const char** atts)
{
	ObjectStack.GetTop()->ObjectRequires.AddElement(tag);

	switch(tag)
	{
	case RefTag:
		{
			NeedsData = 1;
			CurrentExternalObjectState = ExternalReferenceState;
			break;
		}
	case ExtentsTag:
		{
			NeedsData = 1;
			CurrentExternalObjectState = ExternalExtentsState;
			break;
		}
	case RefTypeTag:
		{
			NeedsData = 1;
			CurrentExternalObjectState = ExternalRefTypeState;
			break;
		}
	case NameTag:
		{
			CurrentExternalObjectState = ExternalNameState;
			break;
		}
	case TransformTag:
		{
			//Our CurrentState is now within the Transform
			StateStack.Push(CurrentState);
			CurrentState = TransformState;

			TransformRequires.Init();

			CurrentExternalObjectState = OtherObjectState;
			
			break;
		}
	case DataTag:
		{
			DataRequires.Init();
			
			char **currentAtt = (char **)atts;
			if (*currentAtt == NULL)
			{
				Error(DATA_ERROR2);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, ORGANIZATION) != 0)
				{
					Error(DATA_ERROR1);
				}
				else
				{
					//Our CurrentState is now within the Data
					StateStack.Push(CurrentState);
					CurrentState = DataState;

					//Initialize the Data that we're about to write into
					InitializeCurrentData(*(currentAtt + 1));
				}

				currentAtt += 2;
			}

			CurrentExternalObjectState = OtherObjectState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentExternalObjectState = OtherExternalState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}



void
CXGLImport::ObjectStateEndTagHandler(int tag)
{
	if (NeedsData)
	{
		Error(OBJECT_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case ObjectTag:
		{

			if((ObjectStack.GetTop()->Object->MeshHandle != NULL) && (ObjectStack.GetTop()->Object->NumObjects > 0))
			{
				Error(OBJECT_MESH_SUB_OBJ_ERROR);
				return;
			}
			
			CElementRequires* objReq = &(ObjectStack.GetTop()->ObjectRequires);

			int numTransforms = objReq->HowMany(TransformTag);

			if(numTransforms > 1)
			{
				Error(OBJECT_NUM_TRANSFORMS_ERROR);
				return;
			}
			
			//pop our object from the top of the stack
			CXGLObject *object;
			object = ObjectStack.Pop();


			CXGLContext *parentContext = CurrentContext->GetParentContext();
			
			//give the object to the resolver		
			if (object->XGLID != -1)
			{
				Resolver.AddObjectDefinition(parentContext, CurrentContext, object->XGLID, object);
			}
			else
			{
				CXGLObjectHandle *objectHandle;
				objectHandle = Resolver.AddObject(CurrentContext, object);
			
				// store the object handle
				CObjectStackElem *parent = ObjectStack.GetTop();
				
				//if there's no object on the stack, store the handle in the world
				if (parent == NULL)
				{
					if (CurrentWorld->MeshHandle != NULL)
					{
						Error(WORLD_MESH_SUB_OBJ_ERROR);
						return;
					}

					//if there is a user specified pathID
					if (objectHandle->ObjectPathID != -1)
					{
						int *value = TopObjectPathIDHash.GetMatchingValue(&objectHandle->ObjectPathID);
						if (value)
						{
							Error(UNIQUE_PATHID_ERROR);
							return;
						}

						//add the pathid to the hash
						TopObjectPathIDHash.Add(&objectHandle->ObjectPathID);
					}

					//add the new object to the world
					CurrentWorld->ObjectHandles.AddElement(objectHandle);
					CurrentWorld->NumObjects++;
				}
				else
				{
					if (parent->Object->MeshHandle != NULL)
					{
						Error(OBJECT_MESH_SUB_OBJ_ERROR);
						return;
					}

					//if there is a user specified pathID
					if (objectHandle->ObjectPathID != -1)
					{
						int *value = parent->PathIDHash.GetMatchingValue(&objectHandle->ObjectPathID);
						if (value)
						{
							Error(UNIQUE_PATHID_ERROR);
							return;
						}

						//add the pathid to the hash
						parent->PathIDHash.Add(&objectHandle->ObjectPathID);
					}

					//add the new object to the parent
					parent->Object->ObjectHandles.AddElement(objectHandle);
					parent->Object->NumObjects++;
				}
			}

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			//pop up in the contexts
			CurrentContext = Resolver.ReleaseContext(CurrentContext);

			break;
		}
	case TextureRGBTag:
		{
			assert(CurrentTextureBufferID != -1);
			if (CurrentTextureBufferID != -1)
			{
				Resolver.AddTextureBufferDefinition(CurrentContext, CurrentTextureBufferID, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
				
			CurrentTextureBufferID = -1;
			
			break;
		}
	case TextureRGBATag:
		{
			assert(CurrentTextureBufferID != -1);
			if (CurrentTextureBufferID != -1)
			{
				Resolver.AddTextureBufferDefinition(CurrentContext, CurrentTextureBufferID, CurrentTextureBuffer);
			}
			
			CurrentTextureBufferID = -1;
			break;
		}
	case NameTag:
		{
			CObjectStackElem *parent = ObjectStack.GetTop();	
			assert(parent != NULL);

			//if we have an empty name tag put in an empty string
			if (parent->Object->Name == NULL)
			{
				parent->Object->Name = new char[strlen("") + 1];
				strcpy(parent->Object->Name, "");
			}
		
			break;
		}
	default:
		{
			break;
		}

	}

	CurrentObjectState = OtherObjectState;
}

void
CXGLImport::OtherStateBeginTagHandler(int tag, const char **atts)
{
	OtherStateLevel++;

}

void
CXGLImport::OtherStateCharDataHandler(const char *s)
{
	//ignore stuff here

}

void
CXGLImport::OtherStateEndTagHandler(int tag)
{
	assert(!NeedsData);
	
	OtherStateLevel--;

	if(OtherStateLevel == 0)
	{
		CurrentState = StateStack.Pop();
	}

}


void
CXGLImport::MeshStateBeginTagHandler(int tag, const char** atts)
{
	switch(tag)
	{
	case FaceTag:
		{
			//Our CurrentState is now in the Face
			StateStack.Push(CurrentState);
			CurrentState = FaceState;

			//reset the data flag
			CurrentFaceState = OtherFaceState;
			
			FaceRequires.Init();

			CurrentMeshState = OtherMeshState;

			break;
		}
	case LineTag:
		{
			//Our CurrentState is now in the Line
			StateStack.Push(CurrentState);
			CurrentState = LineState;

			//reset the data flag
			CurrentLineState = OtherLineState;

			LineRequires.Init();

			CurrentMeshState = OtherMeshState;

			break;
		}
	case PointTag:
		{
			//Our CurrentState is now in the Point
			StateStack.Push(CurrentState);
			CurrentState = PointState;

			//reset the data flag
			CurrentPointState = OtherPointState;

			PointRequires.Init();

			CurrentMeshState = OtherMeshState;

			break;
		}
	case VertexPositionTag:
		{
			NeedsData = 1;
			
			//the next piece of data should be a position
			CurrentMeshState = MeshPositionState;

			//find the id from the attribute
			CurrentVertexID = GetDefinitionID(atts);
			if (CurrentVertexID == -1)
			{
				Error(POS_ID_ERROR1);
			}

			break;
		}
	case NormalTag:
		{	
			NeedsData = 1;
			
			//the next piece of data should be a normal
			CurrentMeshState = MeshNormalState;

			//find the id from the attribute
			CurrentNormalID = GetDefinitionID(atts);
			if (CurrentNormalID == -1)
			{
				Error(NORMAL_ID_ERROR1);
			}

			break;
		}
	case TextureCoordinateTag:
		{
			NeedsData = 1;
			
			//the next piece of data should be a texture coordinate
			CurrentMeshState = MeshTextureCoordinateState;

			//find the id from the attribute
			CurrentTextureCoordinateID = GetDefinitionID(atts);
			if (CurrentTextureCoordinateID == -1)
			{
				Error(TEXTURECOORDINATE_ID_ERROR1);
			}

			break;
		}
	case MaterialTag:
		{
			MaterialRequires.Init();
			
			//reset the data flag
			CurrentMeshState = OtherMeshState;

			//change to the material state
			StateStack.Push(CurrentState);
			CurrentState = MaterialState;

			//find the id for this material
			CurrentMaterialID = GetDefinitionID(atts);
			if (CurrentMaterialID == -1)
			{
				Error(MAT_ID_ERROR3);
			}
			else
			{
				InitializeCurrentMaterial();
			}

			break;
		}
	case LineStyleTag:
		{
			LineStyleRequires.Init();

			//Our CurrentState is now within the LineStyle
			StateStack.Push(CurrentState);
			CurrentState = LineStyleState;

			//find the id for this linestyle
			CurrentLineStyleID = GetDefinitionID(atts);
			if (CurrentLineStyleID == -1)
			{
				Error(LINESTYLE_ID_ERROR3);
			}
			else
			{
				InitializeCurrentLineStyle();
			}

			CurrentMeshState = OtherMeshState;
			break;
		}
	case PointStyleTag:
		{
			PointStyleRequires.Init();

			//Our CurrentState is now within the LineStyle
			StateStack.Push(CurrentState);
			CurrentState = PointStyleState;

			//find the id for this linestyle
			CurrentPointStyleID = GetDefinitionID(atts);
			if (CurrentPointStyleID == -1)
			{
				Error(POINTSTYLE_ID_ERROR3);
			}
			else
			{
				InitializeCurrentPointStyle();
			}

			CurrentMeshState = OtherMeshState;
			break;
		}
	case SurfaceTag:
		{
			CurrentMesh->DoubleSided = 1;
			CurrentMeshState = OtherMeshState;
			break;
		}
	case PatchTag:
		{

			if(CurrentPatchIndex != -1)
			{
				Error(PATCH_IN_PATCH);
				return;
			}
			
			int patchID = GetPatchID(atts);

			if(patchID == -1)
			{
				patchID = CurrentPatchHigh + 1;
			}

			//bugbug
			if (patchID == 0)
				patchID = 9999;
			
			if(patchID < 1)
			{
				Error(PATCH_ID_RANGE_ERROR);
				return;
			}

			CurrentPatchHigh = __max(patchID, CurrentPatchHigh);

			//Initialize this patch
			InitializeCurrentPatch(patchID);

			CurrentMeshState = OtherMeshState;

			break;
		}
	case DataTag:
		{
			CurrentMeshState = OtherMeshState;
			
			DataRequires.Init();
			
			char **currentAtt = (char **)atts;
			if (*currentAtt == NULL)
			{
				Error(DATA_ERROR2);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, ORGANIZATION) != 0)
				{
					Error(DATA_ERROR1);
				}
				else
				{
					//Our CurrentState is now within the Data
					StateStack.Push(CurrentState);
					CurrentState = DataState;

					//Initialize the Data that we're about to write into
					InitializeCurrentData(*(currentAtt + 1));
				}

				currentAtt += 2;
			}

			break;
		}
	case TextureTag:
		{
			TextureRequires.Init();
			
			//got a texture at the mesh level
			//Our CurrentState is within the Texture
			StateStack.Push(CurrentState);
			CurrentState = TextureState;

			//find the id for this texture
			CurrentTextureID = GetDefinitionID(atts);
			if (CurrentTextureID == -1)
			{
				Error(TEXTURE_ID_ERROR1);
			}
			else
			{
				InitializeCurrentTexture();
			}

			CurrentMeshState = OtherMeshState;

			break;
		}
	case TextureRGBTag:
		{
			NeedsData = 1;
			
			CurrentMeshState = MeshTextureRGBState;

			//find the width, height and id for this textureBuffer (if any)
			int height = -1;
			int width = -1;
			int xglID = -1;
			GetWidthHeightAndDefinitionID(&width, &height, &xglID, atts);

			//texture buffer didn't have a width
			if (width == -1)
			{
				Error(TEXTUREBUFFER_ERROR2);
			}

			//texture buffer didn't have a height
			if (height == -1)
			{
				Error(TEXTUREBUFFER_ERROR3);
			}

			//texture buffer was defined inside a texture
			if (xglID == -1)
			{
				Error(TEXTUREBUFFER_ERROR7);
			}
			else
			{
				CurrentTextureBufferID = xglID;

				//initialize the texture buffer we're going to write into
				InitializeCurrentTextureBuffer();
				CurrentTextureBuffer->TextureWidth = width;
				CurrentTextureBuffer->TextureHeight = height;
				CurrentTextureBuffer->TextureBufferFormat = TextureRGB;
			}

			break;
		}
	case TextureRGBATag:
		{
			NeedsData = 1;
			
			CurrentMeshState = MeshTextureRGBAState;

			//find the width, height and id for this textureBuffer (if any)
			int height = -1;
			int width = -1;
			int xglID = -1;
			GetWidthHeightAndDefinitionID(&width, &height, &xglID, atts);

			//texture buffer didn't have a width
			if (width == -1)
			{
				Error(TEXTUREBUFFER_ERROR2);
			}

			//texture buffer didn't have a height
			if (height == -1)
			{
				Error(TEXTUREBUFFER_ERROR3);
			}

			//texture buffer was defined inside a texture
			if (xglID == -1)
			{
				Error(TEXTUREBUFFER_ERROR7);
			}
			else
			{
				CurrentTextureBufferID = xglID;

				//initialize the texture buffer we're going to write into
				InitializeCurrentTextureBuffer();
				CurrentTextureBuffer->TextureWidth = width;
				CurrentTextureBuffer->TextureHeight = height;
				CurrentTextureBuffer->TextureBufferFormat = TextureRGBA;
			}

			break;
		}
	case ExtGeomTag:    // it's inside mesh after begining
		{
			char **currentAtt = (char **)atts;
			if (*currentAtt == NULL)
			{
				Error(EXTGEOM_NO_ORG_ATTRIBUTE);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, ORGANIZATION) != 0)
				{
					Error(EXTGEOM_ATTR_NOT_CALLED_ORG);
				}
				else
				{
					//Our CurrentState is now within the Extgeom
					StateStack.Push(CurrentState);
					CurrentState = ExtgeomState;
					//Initialize the Extgeom that we're about to write into
					InitializeCurrentExtgeom(*(currentAtt + 1));
				}

				currentAtt += 2;
			}

			CurrentMeshState = MeshExtGeomState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentMeshState = OtherMeshState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}


void
CXGLImport::MeshStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(MESH_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case MeshTag:
		{
			CXGLContext *meshContext = CurrentContext;
			CXGLContext *parentContext = CurrentContext->GetParentContext();
		
			//we need to copy the data we have accumulated in resizeable arrays
			//and make a new CXGLMesh with the copied data
			CreateCurrentMeshFromResizeableArrays();

			if (CurrentMeshID != -1)
			{
				Resolver.AddMeshDefinition(parentContext, meshContext, CurrentMeshID, CurrentMesh);
				CurrentMesh = NULL;
			}
			else
			{
				CXGLMeshHandle *meshHandle;
				meshHandle = Resolver.AddMesh(meshContext, CurrentMesh);
		
				CurrentMesh = NULL;

				// store the mesh handle
				CObjectStackElem *parent = ObjectStack.GetTop();
				
				//we should never have parent of NULL, b/c that indicates
				//that we have a mesh in the world that is not defined
				if(parent == NULL)
				{
					//parent is the world
					if(CurrentWorld->NumObjects > 0)
					{
						Error(WORLD_MESH_SUB_OBJ_ERROR);
						return;
					}

					if(CurrentWorld->MeshHandle != NULL)
					{
						Error(WORLD_NUM_MESHES_ERROR);
						return;
					}

					CurrentWorld->MeshHandle = meshHandle;
				}
				else
				{
					//parent is an object
					if(parent->Object->MeshHandle != NULL)
					{
						Error(OBJECT_NUM_MESHES_ERROR);
						return;
					}

					if(parent->Object->NumObjects > 0)
					{
						Error(OBJECT_MESH_SUB_OBJ_ERROR);
						return;
					}

					parent->Object->MeshHandle = meshHandle;
				}
			}

			//pop up to our parent state
			CurrentState = StateStack.Pop();
		
			//pop up in the contexts
			CurrentContext = Resolver.ReleaseContext(CurrentContext);		
			
			//uninitialize the mesh
			UnInitializeCurrentMesh();

			break;
		}
	case PatchTag:
		{
			//there is no current patch anymore
			CurrentPatchIndex = -1;

			break;
		}
	case TextureRGBTag:
		{
			assert(CurrentTextureBufferID != -1);
			if (CurrentTextureBufferID != -1)
			{
				Resolver.AddTextureBufferDefinition(CurrentContext, CurrentTextureBufferID, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
				
			CurrentTextureBufferID = -1;
			
			break;
		}
	case TextureRGBATag:
		{
			assert(CurrentTextureBufferID != -1);
			if (CurrentTextureBufferID != -1)
			{
				Resolver.AddTextureBufferDefinition(CurrentContext, CurrentTextureBufferID, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
			
			CurrentTextureBufferID = -1;
			break;
		}
	}

	CurrentMeshState = OtherMeshState;
}


void
CXGLImport::FaceVertexStateBeginTagHandler(int tag, const char** atts)
{

	FaceVertexRequires.AddElement(tag);

	switch(tag)
	{
	case VertexPositionTag:
		{
			NeedsData = 1;

			//indicate the next piece of data is a position
			CurrentFVState = FVPositionState;

			//find the id of the position
			CurrentVertexID = GetDefinitionID(atts);
	
			break;
		}
	case NormalTag:
		{
			NeedsData = 1;
			
			//indicate the next piece of data is a position
			CurrentFVState = FVNormalState;

			//find the id of the normal
			CurrentNormalID = GetDefinitionID(atts);

			break;
		}
	case TextureCoordinateTag:
		{
			NeedsData = 1;
			
			//indicate the next piece of data is a texture coordinate
			CurrentFVState = FVTextureCoordinateState;

			//find the id of the texture coordinate
			CurrentTextureCoordinateID = GetDefinitionID(atts);

			break;
		}
			
	case NormalRefTag:
		{
			NeedsData = 1;
			
			CurrentFVState = FVNormalRefState;
			break;
		}
	case VertexRefTag:
		{
			NeedsData = 1;
			
			CurrentFVState = FVVertexRefState;
			break;
		}
	case TextureCoordinateRefTag:
		{

			NeedsData = 1;
			
			CurrentFVState = FVTextureCoordinateRefState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentFVState = OtherFVState;
			
			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}


void 
CXGLImport::FaceStateBeginTagHandler(int tag, const char** atts)
{

	FaceRequires.AddElement(tag);

	switch(tag)
	{
	case FaceVertexTag1:
		{
			//set the current facevertex
			CurrentFaceVertex = 0;
		
			StateStack.Push(CurrentState);

			FaceVertexRequires.Init();

			CurrentState = FaceVertexState;

			CurrentFaceState = OtherFaceState;

			break;
		}
	case FaceVertexTag2:
		{
			//set the current facevertex
			CurrentFaceVertex = 1;

			StateStack.Push(CurrentState);

			FaceVertexRequires.Init();
			
			CurrentState = FaceVertexState;

			CurrentFaceState = OtherFaceState;

			break;
		}
	case FaceVertexTag3:
		{

			//set the current facevertex
			CurrentFaceVertex = 2;

			StateStack.Push(CurrentState);

			FaceVertexRequires.Init();
			
			CurrentState = FaceVertexState;

			CurrentFaceState = OtherFaceState;

			break;
		}
	case MaterialTag:
		{
			MaterialRequires.Init();
			
			//we are now in a material state
			StateStack.Push(CurrentState);
			CurrentState = MaterialState;

			//find the id of the material
			CurrentMaterialID = GetDefinitionID(atts);
		
			InitializeCurrentMaterial();

			CurrentFaceState = OtherFaceState;
			break;
		}
	case TextureTag:
		{
			TextureRequires.Init();
			
			//Our CurrentState is within the Texture
			StateStack.Push(CurrentState);
			CurrentState = TextureState;

			//find the id for this texture
			CurrentTextureID = GetDefinitionID(atts);
			
			InitializeCurrentTexture();
	
			CurrentFaceState = OtherFaceState;

			break;
		}
	case TextureRefTag:
		{
			NeedsData = 1;
			CurrentFaceState = FaceTextureRefState;
			break;
		}
	case MaterialRefTag:
		{
			NeedsData = 1;
			
			CurrentFaceState = FaceMaterialRefState;
			break;
		}
	case SmoothingGroupTag:
		{
			NeedsData = 1;
			
			CurrentFaceState = FaceSmoothState;
			break;
		}
	case SurfaceTag:
		{
			CurrentMesh->DoubleSided = 1;
			CurrentFaceState = OtherFaceState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentFaceState = OtherFaceState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void
CXGLImport::LineStateBeginTagHandler(int tag, const char** atts)
{
	LineRequires.AddElement(tag);

	switch(tag)
	{
	case LineVertexTag1:
		{
			//set the current facevertex
			CurrentFaceVertex = 0;
		
			StateStack.Push(CurrentState);

			FaceVertexRequires.Init();

			CurrentState = FaceVertexState;

			CurrentLineState = OtherLineState;

			break;
		}
	case LineVertexTag2:
		{
			//set the current facevertex
			CurrentFaceVertex = 1;

			StateStack.Push(CurrentState);

			FaceVertexRequires.Init();
			
			CurrentState = FaceVertexState;

			CurrentLineState = OtherLineState;

			break;
		}
	case MaterialTag:
		{
			MaterialRequires.Init();
			
			//we are now in a material state
			StateStack.Push(CurrentState);
			CurrentState = MaterialState;

			//find the id of the material
			CurrentMaterialID = GetDefinitionID(atts);
		
			InitializeCurrentMaterial();

			CurrentLineState = OtherLineState;
			break;
		}
	case TextureTag:
		{
			TextureRequires.Init();
			
			//Our CurrentState is within the Texture
			StateStack.Push(CurrentState);
			CurrentState = TextureState;

			//find the id for this texture
			CurrentTextureID = GetDefinitionID(atts);
			
			InitializeCurrentTexture();
	
			CurrentLineState = OtherLineState;

			break;
		}
	case LineStyleTag:
		{
			LineStyleRequires.Init();

			//Our CurrentState is within the LineStyle
			StateStack.Push(CurrentState);
			CurrentState = LineStyleState;

			//find the id for this linestyle
			CurrentLineStyleID = GetDefinitionID(atts);

			InitializeCurrentLineStyle();

			CurrentLineState = OtherLineState;

			break;
		}
	case TextureRefTag:
		{
			NeedsData = 1;
			CurrentLineState = LineTextureRefState;
			break;
		}
	case MaterialRefTag:
		{
			NeedsData = 1;
			CurrentLineState = LineMaterialRefState;
			break;
		}
	case LineStyleRefTag:
		{
			NeedsData = 1;
			CurrentLineState = LineStyleRefState;
			break;
		}
	case SurfaceTag:
		{
			CurrentMesh->DoubleSided = 1;
			CurrentLineState = OtherLineState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentLineState = OtherLineState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void
CXGLImport::PointStateBeginTagHandler(int tag, const char** atts)
{
	PointRequires.AddElement(tag);

	switch(tag)
	{
	case PointVertexTag1:
		{
			//set the current facevertex
			CurrentFaceVertex = 0;
		
			StateStack.Push(CurrentState);

			FaceVertexRequires.Init();

			CurrentState = FaceVertexState;

			CurrentPointState = OtherPointState;

			break;
		}
	case MaterialTag:
		{
			MaterialRequires.Init();
			
			//we are now in a material state
			StateStack.Push(CurrentState);
			CurrentState = MaterialState;

			//find the id of the material
			CurrentMaterialID = GetDefinitionID(atts);
		
			InitializeCurrentMaterial();

			CurrentPointState = OtherPointState;
			break;
		}
	case TextureTag:
		{
			TextureRequires.Init();
			
			//Our CurrentState is within the Texture
			StateStack.Push(CurrentState);
			CurrentState = TextureState;

			//find the id for this texture
			CurrentTextureID = GetDefinitionID(atts);
			
			InitializeCurrentTexture();
	
			CurrentPointState = OtherPointState;

			break;
		}
	case PointStyleTag:
		{
			PointStyleRequires.Init();

			//Our CurrentState is within the LineStyle
			StateStack.Push(CurrentState);
			CurrentState = PointStyleState;

			//find the id for this linestyle
			CurrentPointStyleID = GetDefinitionID(atts);

			InitializeCurrentPointStyle();

			CurrentPointState = OtherPointState;

			break;
		}
	case TextureRefTag:
		{
			NeedsData = 1;
			CurrentPointState = PointTextureRefState;
			break;
		}
	case MaterialRefTag:
		{
			NeedsData = 1;
			CurrentPointState = PointMaterialRefState;
			break;
		}
	case PointStyleRefTag:
		{
			NeedsData = 1;
			CurrentPointState = PointStyleRefState;
			break;
		}
	case SurfaceTag:
		{
			CurrentMesh->DoubleSided = 1;
			CurrentPointState = OtherPointState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentPointState = OtherPointState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void
CXGLImport::FaceVertexStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(FACE_VERTEX_NEEDS_DATA);
		return;
	}

	switch(tag)
	{
	case FaceVertexTag1:
	case FaceVertexTag2:
	case FaceVertexTag3:
	case LineVertexTag1:
	case LineVertexTag2:
	case PointVertexTag1:
		{
			
			int numPos = FaceVertexRequires.HowMany(VertexRefTag) + FaceVertexRequires.HowMany(VertexPositionTag);

			if(numPos != 1)
			{
				Error(FACE_VERTEX_ONE_POS);
				return;
			}
			
			int numNorm = FaceVertexRequires.HowMany(NormalRefTag) + FaceVertexRequires.HowMany(NormalTag);

			if(numNorm > 1)
			{
				Error(FACE_VERTEX_NORM);
				return;
			}

			int numTexCos = FaceVertexRequires.HowMany(TextureCoordinateTag) + FaceVertexRequires.HowMany(TextureCoordinateRefTag);
			
			if(numTexCos > 1)
			{
				Error(FACE_VERTEX_TC);
				return;
			}

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			break;
		}
	}

	CurrentFVState = OtherFVState;
}

void 
CXGLImport::FaceStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(FACE_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case FaceTag:
		{
			if(FaceRequires.HowMany(FaceVertexTag1) != 1)
			{
				Error(FACE_FV1_REQUIRES);
				return;
			}
			
			if(FaceRequires.HowMany(FaceVertexTag2) != 1)
			{
				Error(FACE_FV2_REQUIRES);
				return;
			}

			if(FaceRequires.HowMany(FaceVertexTag3) != 1)
			{
				Error(FACE_FV3_REQUIRES);
				return;
			}

			if((FaceRequires.HowMany(MaterialTag) + FaceRequires.HowMany(MaterialRefTag)) != 1)
			{
				Error(FACE_MATERIAL_REQUIRES);
				return;
			}
			
			int numTextures = FaceRequires.HowMany(TextureTag) + FaceRequires.HowMany(TextureRefTag);

			if(numTextures > 1)
			{
				Error(FACE_NUM_TEXTURES_ERROR);
				return;
			}

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			//set the patchI (if any) for the face
			CXGLFace *f = Faces.GetLastElementRef();
			f->PatchI = CurrentPatchIndex;

			//check for partial texture completion
			if (f->TextureI == -1)
			{
				//make sure we have no tcs
				f->FaceVertices[0].TextureCoordinateI = -1;
				f->FaceVertices[1].TextureCoordinateI = -1;
				f->FaceVertices[2].TextureCoordinateI = -1;
			}
			else
			{
				if ((f->FaceVertices[0].TextureCoordinateI == -1) ||
					(f->FaceVertices[1].TextureCoordinateI == -1) ||
					(f->FaceVertices[2].TextureCoordinateI == -1))
				{
					f->TextureI = -1;
					f->FaceVertices[0].TextureCoordinateI = -1;
					f->FaceVertices[1].TextureCoordinateI = -1;
					f->FaceVertices[2].TextureCoordinateI = -1;
				}
			}

			//we have one more face
			Faces.SetSize(Faces.GetNumElements() + 1);

			break;
		}
	
	default:
		CurrentFaceState = OtherFaceState;
		break;
	}
}

void 
CXGLImport::LineStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(LINE_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case LineTag:
		{
			CXGLLine *l = Lines.GetLastElementRef();

			if(LineRequires.HowMany(LineVertexTag1) != 1)
			{
				Error(LINE_LV1_REQUIRES);
				return;
			}
			
			if(LineRequires.HowMany(LineVertexTag2) != 1)
			{
				Error(LINE_LV2_REQUIRES);
				return;
			}

			if((LineRequires.HowMany(MaterialTag) + LineRequires.HowMany(MaterialRefTag)) != 1)
			{
				Error(LINE_MATERIAL_REQUIRES);
				return;
			}
			
			int numOfLineStyleTags = LineRequires.HowMany(LineStyleTag) + LineRequires.HowMany(LineStyleRefTag);
			if(numOfLineStyleTags != 1)
			{
				if(numOfLineStyleTags == 0)
				{
					// set default values for linestyle
					CXGLLineStyle *ls = LineStyles.GetLastElementRef();

					ls->Width = 1.0;
					ls->Pattern = 0xffff;
					ls->PatternFactor = 1;
					ls->XGLID = -1;

					l->LineStyleI = LineStyles.GetNumElements() - 1;
					
					LineStyles.SetSize(LineStyles.GetNumElements() + 1);
				}
				else
				{
					Error(LINE_LINESTYLE_REQUIRES);
					return;
				}
			}

			int numTextures = LineRequires.HowMany(TextureTag) + LineRequires.HowMany(TextureRefTag);

			if(numTextures > 1)
			{
				Error(LINE_NUM_TEXTURES_ERROR);
				return;
			}

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			//set the patchI (if any) for the line
			l->PatchI = CurrentPatchIndex;

			//check for partial texture completion
			if (l->TextureI == -1)
			{
				//make sure we have no tcs
				l->LineVertices[0].TextureCoordinateI = -1;
				l->LineVertices[1].TextureCoordinateI = -1;
			}
			else
			{
				if ((l->LineVertices[0].TextureCoordinateI == -1) ||
					(l->LineVertices[1].TextureCoordinateI == -1))
				{
					l->TextureI = -1;
					l->LineVertices[0].TextureCoordinateI = -1;
					l->LineVertices[1].TextureCoordinateI = -1;
				}
			}

			//we have one more line
			Lines.SetSize(Lines.GetNumElements() + 1);

			break;
		}
	
	default:
		CurrentLineState = OtherLineState;
		break;
	}
}

void 
CXGLImport::PointStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(POINT_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case PointTag:
		{
			CXGLPoint *p = Points.GetLastElementRef();

			if(PointRequires.HowMany(PointVertexTag1) != 1)
			{
				Error(POINT_PV1_REQUIRES);
				return;
			}
			
			if((PointRequires.HowMany(MaterialTag) + PointRequires.HowMany(MaterialRefTag)) != 1)
			{
				Error(POINT_MATERIAL_REQUIRES);
				return;
			}

			int numOfPointStyles = PointRequires.HowMany(PointStyleTag) + PointRequires.HowMany(PointStyleRefTag);
			if(numOfPointStyles != 1)
			{
				if(numOfPointStyles == 0)
				{
					// set default values of point style
					CXGLPointStyle *ps = PointStyles.GetLastElementRef();

					ps->Size = 1.0;
					ps->XGLID = -1;

					p->PointStyleI = PointStyles.GetNumElements() - 1;
					PointStyles.SetSize(PointStyles.GetNumElements() + 1);
				}
				else
				{
					Error(POINT_POINTSTYLE_REQUIRES);
					return;
				}
			}

			int numTextures = PointRequires.HowMany(TextureTag) + PointRequires.HowMany(TextureRefTag);

			if(numTextures > 1)
			{
				Error(POINT_NUM_TEXTURES_ERROR);
				return;
			}

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			//set the patchI (if any) for the point
			p->PatchI = CurrentPatchIndex;

			//check for partial texture completion
			if (p->TextureI == -1)
			{
				//make sure we have no tcs
				p->PointVertex.TextureCoordinateI = -1;
			}
			else
			{
				if (p->PointVertex.TextureCoordinateI == -1)
				{
					p->TextureI = -1;
					p->PointVertex.TextureCoordinateI = -1;
				}
			}

			//we have one more point
			Points.SetSize(Points.GetNumElements() + 1);

			break;
		}
	
	default:
		CurrentPointState = OtherPointState;
		break;
	}
}

void
CXGLImport::MaterialStateBeginTagHandler(int tag, const char** atts)
{

	MaterialRequires.AddElement(tag);

	switch(tag)
	{

	case MaterialEmissiveTag:
		{
			NeedsData = 1;
			
			CurrentMaterialState = EmissiveState;
			break;
		}
	case MaterialDiffuseTag:
		{
			NeedsData = 1;
			
			CurrentMaterialState = DiffuseState;
			break;
		}
	case MaterialSpecularTag:
		{
			NeedsData = 1;
			
			CurrentMaterialState = SpecularState;
			break;
		}
	case MaterialAmbientTag:
		{
			NeedsData = 1;
			
			CurrentMaterialState = AmbientState;
			break;
		}

	case MaterialShininessTag:
		{
			NeedsData = 1;
			
			CurrentMaterialState = ShininessState;
			break;
		}

	case MaterialAlphaTag:
		{
			NeedsData = 1;
			
			CurrentMaterialState = AlphaState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentMaterialState = OtherMaterialState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}

}

void
CXGLImport::MaterialStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(MATERIAL_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case MaterialTag:
		{

			if(MaterialRequires.HowMany(MaterialAmbientTag) != 1)
			{
				Error(MATERIAL_AMBIENT_REQUIRES);
				return;
			}

			if(MaterialRequires.HowMany(MaterialDiffuseTag) != 1)
			{
				Error(MATERIAL_DIFFUSE_REQUIRES);
				return;
			}

			if(MaterialRequires.HowMany(MaterialEmissiveTag) > 1)
			{
				Error(MATERIAL_EMISSIVE_REQUIRES);
				return;
			}

			if(MaterialRequires.HowMany(MaterialSpecularTag) > 1)
			{
				Error(MATERIAL_SPECULAR_REQUIRES);
				return;
			}

			if(MaterialRequires.HowMany(MaterialAlphaTag) > 1)
			{
				Error(MATERIAL_ALPHA_REQUIRES);
				return;
			}
			

			if(MaterialRequires.HowMany(MaterialShininessTag) > 1)
			{
				Error(MATERIAL_SHININESS_REQUIRES);
				return;
			}

			//tell the resolver about any material definitions
			if (CurrentMaterialID != -1)
			{
				CXGLMaterial *toHash = new CXGLMaterial;
				memcpy(toHash, &CurrentMaterial, sizeof(*toHash));
				Resolver.AddMaterialDefinition(CurrentContext, CurrentMaterialID, toHash);
			}
			else
			{
				CurrentMaterial.XGLID = -1;
				CXGLMaterial *m = Materials.GetLastElementRef();
				memcpy(m, &CurrentMaterial, sizeof(*m));
				
				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->MaterialI = Materials.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->MaterialI = Materials.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					p->MaterialI = Materials.GetNumElements() - 1;
				}
				else
				{
					assert(0);
				}

				
				Materials.SetSize(Materials.GetNumElements() + 1);
			}

			//reset the CurrentMaterialID
			CurrentMaterialID = -1;

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			break;
		}	
	default:
		{
			break;
		}
	}

	CurrentMaterialState = OtherMaterialState;

}

void
CXGLImport::LineStyleStateBeginTagHandler(int tag, const char** atts)
{
	LineStyleRequires.AddElement(tag);

	switch(tag)
	{

	case LineStyleWidthTag:
		{
			NeedsData = 1;
			CurrentLineStyleState = LineStyleWidthState;
			break;
		}
	case LineStylePatternTag:
		{
			NeedsData = 1;
			CurrentLineStyleState = LineStylePatternState;
			break;
		}
	case LineStylePatternFactorTag:
		{
			NeedsData = 1;
			CurrentLineStyleState = LineStylePatternFactorState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentLineStyleState = OtherLineStyleState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void
CXGLImport::LineStyleStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(LINESTYLE_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case LineStyleTag:
		{

			if(LineStyleRequires.HowMany(LineStyleWidthTag) != 1)
			{
				Error(LINESTYLE_WIDTH_REQUIRES);
				return;
			}

			if(LineStyleRequires.HowMany(LineStylePatternTag) > 1)
			{
				Error(LINESTYLE_PATTERN_REQUIRES);
				return;
			}

			if(LineStyleRequires.HowMany(LineStylePatternFactorTag) > 1)
			{
				Error(LINESTYLE_PATTERNFACTOR_REQUIRES);
				return;
			}

			//tell the resolver about any linestyle definitions
			if (CurrentLineStyleID != -1)
			{
				CXGLLineStyle *toHash = new CXGLLineStyle;
				memcpy(toHash, &CurrentLineStyle, sizeof(*toHash));
				Resolver.AddLineStyleDefinition(CurrentContext, CurrentLineStyleID, toHash);
			}
			else
			{
				CurrentLineStyle.XGLID = -1;
				CXGLLineStyle *ls = LineStyles.GetLastElementRef();
				memcpy(ls, &CurrentLineStyle, sizeof(*ls));
				
				CXGLLine *l = Lines.GetLastElementRef();
				l->LineStyleI = LineStyles.GetNumElements() - 1;
				
				LineStyles.SetSize(LineStyles.GetNumElements() + 1);
			}

			//reset the CurrentLineStyleID
			CurrentLineStyleID = -1;

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			break;
		}	
	default:
		{
			break;
		}
	}

	CurrentLineStyleState = OtherLineStyleState;

}

void
CXGLImport::PointStyleStateBeginTagHandler(int tag, const char** atts)
{
	PointStyleRequires.AddElement(tag);

	switch(tag)
	{

	case PointStyleSizeTag:
		{
			NeedsData = 1;
			CurrentPointStyleState = PointStyleSizeState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentPointStyleState = OtherPointStyleState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void
CXGLImport::PointStyleStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(POINTSTYLE_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case PointStyleTag:
		{

			if(PointStyleRequires.HowMany(PointStyleSizeTag) != 1)
			{
				Error(POINTSTYLE_SIZE_REQUIRES);
				return;
			}

			//tell the resolver about any pointstyle definitions
			if (CurrentPointStyleID != -1)
			{
				CXGLPointStyle *toHash = new CXGLPointStyle;
				memcpy(toHash, &CurrentPointStyle, sizeof(*toHash));
				Resolver.AddPointStyleDefinition(CurrentContext, CurrentPointStyleID, toHash);
			}
			else
			{
				CurrentPointStyle.XGLID = -1;
				CXGLPointStyle *ps = PointStyles.GetLastElementRef();
				memcpy(ps, &CurrentPointStyle, sizeof(*ps));
				
				CXGLPoint *p = Points.GetLastElementRef();
				p->PointStyleI = Points.GetNumElements() - 1;
				
				PointStyles.SetSize(PointStyles.GetNumElements() + 1);
			}

			//reset the CurrentPointStyleID
			CurrentPointStyleID = -1;

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			break;
		}	
	default:
		{
			break;
		}
	}

	CurrentPointStyleState = OtherPointStyleState;

}

void 
CXGLImport::DataStateBeginTagHandler(int tag, const char** atts)
{	
	DataRequires.AddElement(tag);

	switch (tag)
	{
	case StringTag:
		{
			NeedsData = 1;
			
			char **currentAtt = (char **)atts;
			if (*currentAtt == NULL)
			{
				Error(STRING_ERROR2);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, DATANAME) != 0)
				{
					Error(STRING_ERROR1);
				}
				else
				{
					CurrentDataState = StringState;
					
					//get the length of the name and set it in the resizeable array
					unsigned int length = strlen(*(currentAtt + 1)) + 1;
					StringNameLengths.AddElement(length);

					//get the string name and set it in the resizeable array
					char *sn = new char[length];
					strcpy(sn, *(currentAtt + 1));
					StringNames.AddElement(sn);
				}

				currentAtt += 2;
			}
			
			break;
		}
	case BinTag:
		{
			NeedsData = 1;
			
			char **currentAtt = (char **)atts;
			if (*currentAtt == NULL)
			{
				Error(BIN_ERROR2);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, DATANAME) != 0)
				{
					Error(BIN_ERROR1);
				}
				else
				{
					CurrentDataState = BinState;
					
					//get the length of the name and set it in the resizeable array
					unsigned int length = strlen(*(currentAtt + 1)) + 1;
					BinNameLengths.AddElement(length);

					//get the string name and set it in the resizeable array
					char *bn = new char[length];
					strcpy(bn, *(currentAtt + 1));
					BinNames.AddElement(bn);
				}

				currentAtt += 2;
			}
	
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentDataState = OtherDataState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
	
}

void 
CXGLImport::TextureStateBeginTagHandler(int tag, const char** atts)
{
	TextureRequires.AddElement(tag);

	switch (tag)
	{
	case TextureRGBTag:
		{
			NeedsData = 1;
			
			CurrentTextureState = TextureRGBState;

			//find the width, height and id for this textureBuffer (if any)
			int height = -1;
			int width = -1;
			int xglID = -1;
			GetWidthHeightAndDefinitionID(&width, &height, &xglID, atts);

			//texture buffer didn't have a width
			if (width == -1)
			{
				Error(TEXTUREBUFFER_ERROR2);
			}

			//texture buffer didn't have a height
			if (height == -1)
			{
				Error(TEXTUREBUFFER_ERROR3);
			}

			//texture buffer was defined inside a texture
			if (xglID != -1)
			{
				Error(TEXTUREBUFFER_ERROR9);
			}
		
			CurrentTextureBufferID = xglID;

			//initialize the texture buffer we're going to write into
			InitializeCurrentTextureBuffer();
			CurrentTextureBuffer->TextureWidth = width;
			CurrentTextureBuffer->TextureHeight = height;
			CurrentTextureBuffer->TextureBufferFormat = TextureRGB;

			//note the data format in the currenttexture
			CurrentTexture->TextureBufferFormat = TextureRGB;

			break;
		}
	case TextureRGBRefTag:
		{
			NeedsData = 1;
			CurrentTextureState = TextureRGBRefState;

			//note the data format in the currenttexture
			CurrentTexture->TextureBufferFormat = TextureRGB;
			break;
		}
	case TextureRGBATag:
		{	
			NeedsData = 1;
			
			CurrentTextureState = TextureRGBAState;

			//find the width, height and id for this textureBuffer (if any)
			int height = -1;
			int width = -1;
			int xglID = -1;
			GetWidthHeightAndDefinitionID(&width, &height, &xglID, atts);

			//texture buffer didn't have a width
			if (width == -1)
			{
				sprintf(ErrorString, "%d: TextureRGB or TextureRGBA Tag found without a width", TEXTUREBUFFER_ERROR2);
				Error(ErrorString);
			}

			//texture buffer didn't have a width
			if (height == -1)
			{
				sprintf(ErrorString, "%d: TextureRGB or TextureRGBA Tag found without a height", TEXTUREBUFFER_ERROR2);
				Error(ErrorString);
			}

			//texture buffer was defined inside a texture
			if (xglID != -1)
			{
				sprintf(ErrorString, "%d: TextureRGB or TextureRGBA Tag found defined within a Texture Tag", TEXTUREBUFFER_ERROR3);
				Error(ErrorString);
			}

			CurrentTextureBufferID = xglID;

			InitializeCurrentTextureBuffer();
			CurrentTextureBuffer->TextureWidth = width;
			CurrentTextureBuffer->TextureHeight = height;
			CurrentTextureBuffer->TextureBufferFormat = TextureRGBA;

			//note the data format in the currenttexture
			CurrentTexture->TextureBufferFormat = TextureRGBA;

			break;
		}
	case TextureRGBARefTag:
		{				
			NeedsData = 1;
			CurrentTextureState = TextureRGBARefState;

			//note the data format in the currenttexture
			CurrentTexture->TextureBufferFormat = TextureRGBA;
			break;
		}
	case ReplaceTag:
		{
			CurrentTexture->TextureFunction = TextureReplace;
			CurrentTextureState = OtherTextureState;
			break;
		}
	case ModulateTag:
		{
			CurrentTexture->TextureFunction = TextureModulate;
			CurrentTextureState = OtherTextureState;
			break;
		}
	case DecalTag:
		{
			CurrentTexture->TextureFunction = TextureDecal;
			CurrentTextureState = OtherTextureState;
			break;
		}
	case RepeatTag:
		{
			CurrentTexture->TextureWrap = TextureRepeat;
			CurrentTextureState = OtherTextureState;
			break;
		}
	case ClampTag:
		{
			CurrentTexture->TextureWrap = TextureClamp;
			CurrentTextureState = OtherTextureState;
			break;
		}
	case TextureBorderColorTag:
		{
			NeedsData = 1;
			CurrentTextureState = TextureBorderColorState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;
	
			CurrentTextureState = OtherTextureState;
				
			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void
CXGLImport::TransformStateBeginTagHandler(int tag, const char** atts)
{
	TransformRequires.AddElement(tag);

	switch(tag)
	{
	case TransformForwardTag:
		{
			NeedsData = 1;
			CurrentTransformState = ForwardState;
			break;
		}
	case TransformUpTag:
		{	
			NeedsData = 1;
			CurrentTransformState = UpState;
			break;
		}
	case TransformPositionTag:
		{
			NeedsData = 1;
			CurrentTransformState = PositionState;
			break;
		}
	case TransformScaleTag:
		{
			NeedsData = 1;
			CurrentTransformState = ScaleState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentTransformState = OtherTransformState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}

}



void
CXGLImport::TransformStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(TRANSFORM_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case TransformTag:
		{
			if(TransformRequires.HowMany(TransformForwardTag) != 1)
			{
				Error(TRANSFORM_FORWARD_REQUIRES);
				return;
			}

			if(TransformRequires.HowMany(TransformUpTag) != 1)
			{	
				Error(TRANSFORM_UP_REQUIRES);
				return;
			}

			if(TransformRequires.HowMany(TransformPositionTag) != 1)
			{
				Error(TRANSFORM_POSITION_REQUIRES);
				return;
			}

			if(TransformRequires.HowMany(TransformScaleTag) > 1)
			{
				Error(TRANSFORM_NUM_SCALES_ERROR);
				return;
			}

			//pop up to the previous state
			CheckAndFixTransform();
			CurrentState = StateStack.Pop();
			break;
		}
	default:
		{
			break;
		}
	}

	CurrentTransformState = OtherTransformState;
}

//DEBUG
bool
CheckTransform(CXGLTransform *t)
{
	double dot = t->FX*t->UX + t->FY*t->UY + t->FZ*t->UZ;
	return(fabs(dot) < 0.999);
}

void
CXGLImport::CheckAndFixTransform()
{
	//get our object
	CObjectStackElem *top = ObjectStack.GetTop();
	CXGLObject *object = top->Object;

	//normalize the forward vector

	if(object->Transform.Scale <= 0)
	{
		Error(TRANSFORM_ERROR4);
		return;
	}


	double size = sqrt(object->Transform.FX * object->Transform.FX + object->Transform.FY * object->Transform.FY + object->Transform.FZ * object->Transform.FZ);

	if(size == 0.0)
	{
		Error(TRANSFORM_ERROR1);
		return;
	}

	double invMag = 1.0/size;

	object->Transform.FX *= invMag;
	object->Transform.FY *= invMag;
	object->Transform.FZ *= invMag;

	//normalize the up vector

	size = sqrt(object->Transform.UX * object->Transform.UX + object->Transform.UY * object->Transform.UY + object->Transform.UZ * object->Transform.UZ);

	if(size == 0.0)
	{
		Error(TRANSFORM_ERROR2);
		return;
	}

	invMag = 1.0/size;

	object->Transform.UX *= invMag;
	object->Transform.UY *= invMag;
	object->Transform.UZ *= invMag;
	
	//find the projection of the forward in not in the up direction

	double FDotU = (object->Transform.FX * object->Transform.UX + object->Transform.FY * object->Transform.UY + object->Transform.FZ * object->Transform.UZ);

	if ((FDotU == -1) || (FDotU == 1))
	{
		Error(TRANSFORM_ERROR5);
		return;
	}

	double compUFX = object->Transform.FX * FDotU;
	double compUFY = object->Transform.FY * FDotU;
	double compUFZ = object->Transform.FZ * FDotU;

	object->Transform.UX = (float) (object->Transform.UX - compUFX);
	object->Transform.UY = (float) (object->Transform.UY - compUFY);
	object->Transform.UZ = (float) (object->Transform.UZ - compUFZ);

	size = sqrt(object->Transform.UX*object->Transform.UX + object->Transform.UY*object->Transform.UY + object->Transform.UZ*object->Transform.UZ);

	if(size == 0.0)
	{
		Error(TRANSFORM_ERROR3);
		return;
	}

	float iMag = (float) (1.0/size);

	object->Transform.UX *= iMag;
	object->Transform.UY *= iMag;
	object->Transform.UZ *= iMag;

}


void
CXGLImport::BackgroundStateBeginTagHandler(int tag, const char** atts)
{
	BackgroundRequires.AddElement(tag);

	switch(tag)
	{
	case BackgroundColorTag:
		{
			NeedsData = 1;
			
			CurrentBackgroundState = BackgroundColorState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentBackgroundState = OtherBackgroundState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}

	}

}


void
CXGLImport::LightingStateBeginTagHandler(int tag, const char** atts)
{

	switch(tag)
	{
	case LightAmbientTag:
		{
			NeedsData = 1;
			
			LightingRequires.AddElement(tag);
			CurrentLightingState = LightingAmbientState;
			break;
		}
	case DirectionalLightTag:
		{
			//added by djwolf 06/23/00
			if (CurrentLighting->NumLights >= 8)
			{
				Error(LIGHTS_ERROR1);
			}
			else
			{
				LightingRequires.AddElement(tag);
				DirectionalLightRequires.Init();
				InitializeCurrentLight();

				CurrentLightingState = OtherLightingState;
			}
			break;
		}
	case LightDirectionTag:
		{
			NeedsData = 1;
			
			DirectionalLightRequires.AddElement(tag);
			CurrentLightingState = LightingDirectionState;
			break;
		}
	case LightDiffuseTag:
		{
			NeedsData = 1;
			
			DirectionalLightRequires.AddElement(tag);
			CurrentLightingState = LightingDiffuseState;
			break;
		}
	case LightSpecularTag:
		{
			NeedsData = 1;
			
			DirectionalLightRequires.AddElement(tag);
			CurrentLightingState = LightingSpecularState;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;

			CurrentLightingState = OtherLightingState;

			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}

}

void
CXGLImport::BackgroundStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(BACKGROUND_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case BackgroundTag:
		{	
			if(BackgroundRequires.HowMany(BackgroundColorTag) != 1)
			{
				Error(BACKGROUND_REQUIRES);
				return;
			}

			Resolver.AddBackground(CurrentContext, CurrentBackground);

			CurrentBackground = NULL;

			//pop up to our parent state
			CurrentState = StateStack.Pop();
			assert(CurrentState == WorldState);
			break;
		}
	case BackgroundColorTag:
		break;
	default:
		{
			break;
		}
	}

}


void
CXGLImport::LightingStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(LIGHTING_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case DirectionalLightTag:
		{
			if(DirectionalLightRequires.HowMany(LightDirectionTag) != 1)
			{				
				Error(DIRECTIONAL_LIGHT_DIRECTION_REQUIRES);
				return;
			}
			

			int numDiffuse = DirectionalLightRequires.HowMany(LightDiffuseTag);
			if(numDiffuse > 1)	
			{
				Error(DIRECTIONAL_LIGHT_DIFFUSE_REQUIRES);
				return;
			}

			int numSpecular = DirectionalLightRequires.HowMany(LightSpecularTag);
			if(numSpecular > 1)
			{
				Error(DIRECTIONAL_LIGHT_SPECULAR_REQUIRES);
				return;
			}
			
			if((numSpecular + numDiffuse) == 0)
			{
				Error(DIRECTIONAL_LIGHT_DIFSPEC_REQUIRES);
			}

			if (CurrentLighting->NumLights >= 8)
			{
				Error(LIGHTS_ERROR1);
				return;
			}
			else
			{
				CurrentLighting->NumLights++;
			}
			break;
		}
	case LightingTag:
		{
			int numAmbient = LightingRequires.HowMany(LightAmbientTag);
			int numDirectionalLights = LightingRequires.HowMany(DirectionalLightTag);

			if(numAmbient > 1)
			{
				Error(LIGHTING_REQUIRES);
				return;
			}
			
			if((numAmbient + numDirectionalLights) == 0)
			{
				Error(LIGHTING_REQUIRES);
				return;
			}
				
			Resolver.AddLighting(CurrentContext, CurrentLighting);
			CurrentLighting = NULL;
			
			//pop up to our parent state
			CurrentState = StateStack.Pop();
			assert(CurrentState == WorldState);
			break;
		}
	default:
		{
			CurrentLightingState = OtherLightingState;
			break;
		}
	}
}


void
CXGLImport::WorldStateCharDataHandler(const char*s)
{
	switch(CurrentWorldState)
	{
		case WorldTextureRGBState:
		{
			unsigned int length = strlen(s);
			
			//make sure that the width and height are correct
			if (length != (CurrentTextureBuffer->TextureWidth*CurrentTextureBuffer->TextureHeight*6))
			{
				Error(TEXTUREBUFFER_ERROR5);
			}
			else
			{
				int numPixels = (int)((length*0.5)/3);
				CurrentTextureBuffer->RGBABuffer = new unsigned char[numPixels*4]; //for the alpha

				char string[3];

				int i;
				for (i = 0; i < numPixels; i ++)
				{
					memcpy(string, s+6*i+0, 2);
					string[2] = 0;
					int result = strtoul(string, NULL ,16); 
					CurrentTextureBuffer->RGBABuffer[4*i] = (unsigned char)result;

					memcpy(string, s+6*i+2, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+1] = (unsigned char)result;

					memcpy(string, s+6*i+4, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+2] = (unsigned char)result;

					//the alpha bit
					CurrentTextureBuffer->RGBABuffer[4*i+3] = (unsigned char)255;
				}
			}

			break;
		}
		case WorldTextureRGBAState:
		{
			unsigned int length = strlen(s);
			
			//make sure that the width and height are correct
			if (length != (CurrentTextureBuffer->TextureWidth*CurrentTextureBuffer->TextureHeight*8))
			{
				Error(TEXTUREBUFFER_ERROR8);
			}
			else
			{
				int numPixels = (int)((length*0.5)/4);
				CurrentTextureBuffer->RGBABuffer = new unsigned char[numPixels*4]; //for the alpha

				char string[3];

				int i;
				for (i = 0; i < numPixels; i ++)
				{
					memcpy(string, s+8*i+0, 2);
					string[2] = 0;
					int result = strtoul(string, NULL ,16); 
					CurrentTextureBuffer->RGBABuffer[4*i] = (unsigned char)result;

					memcpy(string, s+8*i+2, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+1] = (unsigned char)result;

					memcpy(string, s+8*i+4, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+2] = (unsigned char)result;

					memcpy(string, s+8*i+6, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+3] = (unsigned char)result;
				}
			}

			break;
		}
		case WorldTextureCoordinateState:
		{
			float u, v;
			if(!Get2FloatVector(s, &u, &v))
			{
				return;
			}

			assert(CurrentTextureCoordinateID != -1);
			if (CurrentTextureCoordinateID != -1)
			{
				CXGLTextureCoordinate *tc = new CXGLTextureCoordinate;
				tc->U = u;
				tc->V = v;
				//this is a defined texture coordinate

				Resolver.AddTextureCoordinateDefinition(CurrentContext, CurrentTextureCoordinateID, tc);
			}
		
			break;
		}
	case WorldObjectRefState:
		{
			int id;
			if(!GetInt(s, &id))
			{
				return;
			}

			CXGLObjectHandle* objectHandle = Resolver.CreateObjectReference(CurrentContext, id);

			CurrentWorld->ObjectHandles.AddElement(objectHandle);
			CurrentWorld->NumObjects++;

			break;
		}
	case WorldNameState:
		{
			if (CurrentWorld->Name)
			{
				Error(NAME_ERROR);
			}
			else
			{
				CurrentWorld->Name = new char[strlen(s) + 1];
				strcpy(CurrentWorld->Name, s);
			}

			break;
		}
	case WorldMeshRefState:
		{
			int id;
			if(!GetInt(s, &id))
			{
				return;
			}

			CXGLMeshHandle* meshHandle = Resolver.CreateMeshReference(CurrentContext, id);

			if(CurrentWorld->MeshHandle != NULL)
			{
				Error(WORLD_NUM_MESHES_ERROR);
				return;
			}

			if(CurrentWorld->NumObjects > 0)
			{
				Error(WORLD_MESH_SUB_OBJ_ERROR);
				return;
			}

			CurrentWorld->MeshHandle = meshHandle;

			break;
		}
	}
}

void
CXGLImport::ObjectStateCharDataHandler(const char*s)
{
	
	switch(CurrentObjectState)
	{
	case ObjectMeshRefState:
		{
			int id;

			if(!GetInt(s, &id))
			{
				return;
			}
			
			CXGLMeshHandle* meshHandle = Resolver.CreateMeshReference(CurrentContext, id);

			CObjectStackElem *parent = ObjectStack.GetTop();
			
			assert(parent != NULL);

			if(parent->Object->MeshHandle != NULL)
			{
				Error(OBJECT_NUM_MESHES_ERROR);
				return;
			}

			if(parent->Object->NumObjects > 0)
			{
				Error(OBJECT_MESH_SUB_OBJ_ERROR);
				return;
			}

			parent->Object->MeshHandle = meshHandle;
			break;
		}
	case ObjectObjectRefState:
		{
			int id;
			if(!GetInt(s, &id))
			{
				return;
			}

			CXGLObjectHandle* objectHandle = Resolver.CreateObjectReference(CurrentContext, id);

			CObjectStackElem *parent = ObjectStack.GetTop();

			assert(parent != NULL);

			if(parent->Object->MeshHandle != NULL)
			{
				Error(OBJECT_MESH_SUB_OBJ_ERROR);
				return;
			}

			parent->Object->ObjectHandles.AddElement(objectHandle);

			parent->Object->NumObjects++;

			break;
		}
	case ObjectNameState:
		{
			CObjectStackElem *parent = ObjectStack.GetTop();	
			assert(parent != NULL);

			if (parent->Object->Name)
			{
				Error(NAME_ERROR);
			}
			else
			{
				parent->Object->Name = new char[strlen(s) + 1];
				strcpy(parent->Object->Name, s);
			}
		
			break;
		}
	case ObjectTextureRGBState:
		{
			unsigned int length = strlen(s);
			
			//make sure that the width and height are correct
			if (length != (CurrentTextureBuffer->TextureWidth*CurrentTextureBuffer->TextureHeight*6))
			{
				Error(TEXTUREBUFFER_ERROR5);
			}
			else
			{
				int numPixels = (int)((length*0.5)/3);
				CurrentTextureBuffer->RGBABuffer = new unsigned char[numPixels*4]; //for the alpha

				char string[3];

				int i;
				for (i = 0; i < numPixels; i ++)
				{
					memcpy(string, s+6*i+0, 2);
					string[2] = 0;
					int result = strtoul(string, NULL ,16); 
					CurrentTextureBuffer->RGBABuffer[4*i] = (unsigned char)result;

					memcpy(string, s+6*i+2, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+1] = (unsigned char)result;

					memcpy(string, s+6*i+4, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+2] = (unsigned char)result;

					//the alpha bit
					CurrentTextureBuffer->RGBABuffer[4*i+3] = (unsigned char)255;
				}
			}

			break;
		}
	case ObjectTextureRGBAState:
		{
			unsigned int length = strlen(s);
			
			//make sure that the width and height are correct
			if (length != (CurrentTextureBuffer->TextureWidth*CurrentTextureBuffer->TextureHeight*8))
			{
				Error(TEXTUREBUFFER_ERROR8);
			}
			else
			{
				int numPixels = (int)((length*0.5)/4);
				CurrentTextureBuffer->RGBABuffer = new unsigned char[numPixels*4]; //for the alpha

				char string[3];

				int i;
				for (i = 0; i < numPixels; i ++)
				{
					memcpy(string, s+8*i+0, 2);
					string[2] = 0;
					int result = strtoul(string, NULL ,16); 
					CurrentTextureBuffer->RGBABuffer[4*i] = (unsigned char)result;

					memcpy(string, s+8*i+2, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+1] = (unsigned char)result;

					memcpy(string, s+8*i+4, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+2] = (unsigned char)result;

					memcpy(string, s+8*i+6, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+3] = (unsigned char)result;
				}
			}

			break;
		}
	case ObjectTextureCoordinateState:
		{
			float u, v;
			if(!Get2FloatVector(s, &u, &v))
			{
				return;
			}

			assert(CurrentTextureCoordinateID != -1);
			if (CurrentTextureCoordinateID != -1)
			{
				CXGLTextureCoordinate *tc = new CXGLTextureCoordinate;
				tc->U = u;
				tc->V = v;
				//this is a defined texture coordinate

				Resolver.AddTextureCoordinateDefinition(CurrentContext, CurrentTextureCoordinateID, tc);
			}
		
			break;
		}
	}
}

void
CXGLImport::ExternalObjectStateCharDataHandler(const char *s)
{
	//get our current object
	CXGLObject *object = ObjectStack.GetTop()->Object;
	assert(object);

	switch(CurrentExternalObjectState)
	{
	case ExternalReferenceState:
		{
			object->Ref = new char[strlen(s) + 1];
			strcpy(object->Ref, s);
			break;
		}
	case ExternalExtentsState:
		{
			if (!Get6DoubleVector(	s, 
									&(object->Extents.MinX), &(object->Extents.MinY), &(object->Extents.MinZ), 
									&(object->Extents.MaxX), &(object->Extents.MaxY), &(object->Extents.MaxZ)))
				return;
			
			if ((object->Extents.MinX > object->Extents.MaxX) || 
				(object->Extents.MinY > object->Extents.MaxY) || 
				(object->Extents.MinZ > object->Extents.MaxZ))
			{
				Error(EXTENTS_RELATION_ERROR);
			}

			break;
		}
	case ExternalRefTypeState:
		{
			object->RefType = new char[strlen(s) + 1];
			strcpy(object->RefType, s);
			break;
		}
	case ExternalNameState:
		{
			object->Name = new char[strlen(s) + 1];
			strcpy(object->Name, s);
			break;
		}
	}
}


void
CXGLImport::TransformStateCharDataHandler(const char*s)
{
	//get our object
	CObjectStackElem *top = ObjectStack.GetTop();
	CXGLObject *object = top->Object;

	switch(CurrentTransformState)
	{
	case ForwardState:
		{
			Get3DoubleVector(s, &(object->Transform.FX), &(object->Transform.FY), &(object->Transform.FZ));
			break;
		}
	case UpState:
		{
			Get3DoubleVector(s, &(object->Transform.UX), &(object->Transform.UY), &(object->Transform.UZ));
			break;
		}
	case PositionState:
		{
			Get3DoubleVector(s, &(object->Transform.XP), &(object->Transform.YP), &(object->Transform.ZP));
			break;
		}
	case ScaleState:
		{
			GetDouble(s, &(object->Transform.Scale));
			break;
		}
	default:
		{
			break;
		}
	}	
}

int
CXGLImport::SkipWhiteSpaceAndCommas(char* start, char** next)
{
	*next = start;
	
	int done = 0;

	while(!done)
	{
		switch(**next)
		{
			case ' ':
			case '\t':
			case '\n':
			case ',':
			{
				(*next)++;
				break;
			}
			default:
			{
				done = 1;
			}
		}
	}

	return (start != (*next));
}

int
CXGLImport::GetDouble(const char* string, double*num)
{
	char* rest;
	double myDouble = strtod(string, &rest);

	if(rest == string)
	{
		Error(READ_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(myDouble) || _isnan(myDouble))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}	

	*num = myDouble;

	return 1;
}

int
CXGLImport::GetInt(const char* string, int* num)
{
	char* rest;

	int myInt = (int) strtol(string, &rest, 10);

	if(rest == string)
	{
		Error(READ_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(myInt) || _isnan(myInt))
	{
		Error(INT_NUMBER_ERROR);
		return 0;
	}

	*num = myInt;

	return 1;
}

int
CXGLImport::GetUnsignedInt(const char *string, unsigned int* num)
{
	int val;
	GetInt(string, &val);

	if(!_finite(val) || _isnan(val))
	{
		Error(INT_NUMBER_ERROR);
		return 0;
	}

	*num = (unsigned int) val;

	return 1;
}


int
CXGLImport::GetNormalized3FloatVector(float x, float y, float z, float* newX, float* newY, float* newZ)
{
	float iMag = 1.0f/((float)sqrt(x*x + y*y + z*z));

	if(iMag == 0.0f)
	{
		Error(ZERO_DIRECTIONVECTOR_ERROR);
		return 0;
	}

	*newX = x * iMag;
	*newY = y * iMag;
	*newZ = z * iMag;

	return 1;
}

int
CXGLImport::GetFloat(const char* string, float* num)
{
	double myDouble;
	float myFloat;

	int success = GetDouble(string, &myDouble);

	if(!success)
	{
		return 0;
	}

	myFloat = (float) myDouble;

	if(!_finite(myFloat) || _isnan(myFloat))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}	

	*num = myFloat;

	return 1;
}

int
CXGLImport::Get2DoubleVector(const char* string, double* val1, double* val2)
{
	char* rest;
	char* nextStart;
	double d1, d2;

	d1 = strtod(string, &rest);
	
	if(rest == string)
	{
		Error(TWO_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d2 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{
		Error(TWO_VECTOR_ERROR);
		return 0;
	}	

	if(!_finite(d1) || _isnan(d1))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d2) || _isnan(d2))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	*val1 = d1;
	*val2 = d2;

	return 1;
}


int
CXGLImport::Get3DoubleVector(const char* string, double* val1, double* val2, double* val3)
{
	char* rest;
	char* nextStart;
	double d1, d2, d3;

	d1 = strtod(string, &rest);

	if(rest == string)
	{
		Error(THREE_VECTOR_ERROR);
		return 0;		
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d2 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{
		Error(THREE_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d3 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{
		Error(THREE_VECTOR_ERROR);
		return 0;
	}

	if(!_finite(d1) || _isnan(d1))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}
	
	if(!_finite(d2) || _isnan(d2))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d3) || _isnan(d3))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	*val1 = d1;
	*val2 = d2;
	*val3 = d3;

	return 1;
}

int
CXGLImport::Get4DoubleVector(const char* string, double* val1, double* val2, double* val3, double* val4)
{
	char* rest;
	char* nextStart;
	double d1, d2, d3, d4;

	d1 = strtod(string, &rest);
	
	if(rest == string)
	{
		Error(FOUR_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d2 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{
		Error(FOUR_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d3 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{
		Error(FOUR_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d4 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{	
		Error(FOUR_VECTOR_ERROR);
		return 0;
	}


	if(!_finite(d1) || _isnan(d1))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d2) || _isnan(d2))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d3) || _isnan(d3))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d4) || _isnan(d4))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	*val1 = d1;
	*val2 = d2;
	*val3 = d3;
	*val4 = d4;

	return 1;
}


int
CXGLImport::Get6DoubleVector(const char* string, double* val1, double* val2, double* val3, double* val4, double* val5, double* val6)
{
	char* rest;
	char* nextStart;
	double d1, d2, d3, d4, d5, d6;

	d1 = strtod(string, &rest);
	
	if(rest == string)
	{
		Error(SIX_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d2 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{
		Error(SIX_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d3 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{
		Error(SIX_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d4 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{	
		Error(SIX_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d5 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{	
		Error(SIX_VECTOR_ERROR);
		return 0;
	}

	SkipWhiteSpaceAndCommas(rest, &nextStart);
	d6 = strtod(nextStart, &rest);

	if(rest == nextStart)
	{	
		Error(SIX_VECTOR_ERROR);
		return 0;
	}

	if(!_finite(d1) || _isnan(d1))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d2) || _isnan(d2))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d3) || _isnan(d3))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d4) || _isnan(d4))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d5) || _isnan(d5))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(d6) || _isnan(d6))
	{
		Error(DOUBLE_NUMBER_ERROR);
		return 0;
	}

	*val1 = d1;
	*val2 = d2;
	*val3 = d3;
	*val4 = d4;
	*val5 = d5;
	*val6 = d6;

	return 1;
}


int 
CXGLImport::Get2FloatVector(const char* string, float* val1, float *val2)
{
	double d1, d2;
	float f1, f2;

	int success = Get2DoubleVector(string, &d1, &d2);

	if(!success)
	{
		return 0;
	}

	f1 = (float) d1;
	f2 = (float) d2;

	if(!_finite(f1) || _isnan(f1))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f2) || _isnan(f2))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	*val1 = f1;
	*val2 = f2;

	return 1;
}

int
CXGLImport::Get3FloatVector(const char* string, float *val1, float *val2, float *val3)
{
	double d1, d2, d3;
	float f1, f2, f3;

	int success = Get3DoubleVector(string, &d1, &d2, &d3);

	if(!success)
	{
		return 0;
	}

	f1 = (float) d1;
	f2 = (float) d2;
	f3 = (float) d3;

	if(!_finite(f1) || _isnan(f1))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f2) || _isnan(f2))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}


	if(!_finite(f3) || _isnan(f3))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	*val1 = f1;
	*val2 = f2;
	*val3 = f3;

	return 1;
}

int 
CXGLImport::Get4FloatVector(const char* string, float* val1, float *val2, float *val3, float *val4)
{
	double d1, d2, d3, d4;
	float f1, f2, f3, f4;

	int success = Get4DoubleVector(string, &d1, &d2, &d3, &d4);

	if(!success)
	{
		return 0;
	}

	f1 = (float) d1;
	f2 = (float) d2;
	f3 = (float) d3;
	f4 = (float) d4;

	if(!_finite(f1) || _isnan(f1))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f2) || _isnan(f2))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f3) || _isnan(f3))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f4) || _isnan(f4))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	*val1 = f1;
	*val2 = f2;
	*val3 = f3;
	*val4 = f4;

	return 1;
}

int 
CXGLImport::Get6FloatVector(const char* string, float* val1, float *val2, float *val3, float *val4, float *val5, float *val6)
{
	double d1, d2, d3, d4, d5, d6;
	float f1, f2, f3, f4, f5, f6;

	int success = Get6DoubleVector(string, &d1, &d2, &d3, &d4, &d5, &d6);

	if(!success)
	{
		return 0;
	}

	f1 = (float) d1;
	f2 = (float) d2;
	f3 = (float) d3;
	f4 = (float) d4;
	f5 = (float) d5;
	f6 = (float) d6;

	if(!_finite(f1) || _isnan(f1))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f2) || _isnan(f2))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f3) || _isnan(f3))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f4) || _isnan(f4))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f5) || _isnan(f5))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}

	if(!_finite(f6) || _isnan(f6))
	{
		Error(FLOAT_NUMBER_ERROR);
		return 0;
	}
	
	*val1 = f1;
	*val2 = f2;
	*val3 = f3;
	*val4 = f4;
	*val5 = f5;
	*val6 = f6;
	return 1;
}


int
CXGLImport::IsValidHexChar(char c)
{
	switch (c)
	{
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case 'a':
	case 'b':
	case 'c':
	case 'd':
	case 'e':
	case 'f':
	case 'A':
	case 'B':
	case 'C':
	case 'D':
	case 'E':
	case 'F':
		return 1;

	default:
		return 0;
	}
}

int
CXGLImport::GetBinaryDataFromHex(int hexLen, const char* hexString, unsigned char* data)
{
	//data is preallocated 
	
	char tempStr[3];
	tempStr[2] = NULL;

	int i;
	for(i = 0; i < hexLen; i++)
	{
		char c = hexString[i];

		//check c here!

		if(!IsValidHexChar(c))
		{
			Error(BINARY_DATA_INVALID_CHAR_ERROR);
			return 0;
		}

		tempStr[i%2] = c;
		
		if((i%2) == 1)
		{
			char* endChar = tempStr + 2;
			long val = strtoul(tempStr, &endChar, 16);
			data[i/2] = (unsigned char) val;
		}
	}

	return 1;
}


int 
CXGLImport::GetDefinitionID(const char** atts)
{
	int ret = -1;

	char **currentAtt = (char **)atts;
	while (*currentAtt != NULL)
	{
		if (stricmp(*currentAtt, ID) == 0)
		{
			if(!GetInt((char*) *(currentAtt + 1), &ret))
			{
				return 0;
			}
		}

		currentAtt += 2;
	}

	return ret;
}

int
CXGLImport::GetPathID(const char** atts)
{
	int ret = -1;

	char **currentAtt = (char **)atts;
	while (*currentAtt != NULL)
	{
		if (stricmp(*currentAtt, PATHID) == 0)
		{
			if(!GetInt((char*) *(currentAtt + 1), &ret))
			{
				return 0;
			}
		}

		currentAtt += 2;
	}

	return ret;
}

int
CXGLImport::GetPatchID(const char** atts)
{
	int ret = -1;

	char **currentAtt = (char **)atts;
	while(*currentAtt != NULL)
	{
		if(stricmp(*currentAtt, PATCHID) == 0)
		{
			if(!GetInt((char*) *(currentAtt +1), &ret))
			{	
				return -1;
			}
		}


		//bugbug
		if(stricmp(*currentAtt, ID) == 0)
			GetInt((char*) *(currentAtt +1), &ret);

		currentAtt += 2;
	}

	return ret;
}

void 
CXGLImport::GetDefinitionAndPathID(int *defID, int *pathID, const char** atts)
{
	*defID = -1;
	*pathID = -1;

	char **currentAtt = (char **)atts;
	while (*currentAtt != NULL)
	{
		if (stricmp(*currentAtt, PATHID) != 0)
		{
			if (stricmp(*currentAtt, ID) != 0)
			{
				Error(PATH_ID_ERROR1);
			}
			else
			{
				//we have an id
				GetInt((char*) *(currentAtt + 1), defID);
			}
		}
		else
		{
			//we have a pathid
			GetInt((char*) *(currentAtt + 1), pathID);

			if((*pathID < 0) || (*pathID > 16000000))
			{
				Error(PATH_ID_RANGE_ERROR);
			}
		}

		currentAtt += 2;
	}
}

void 
CXGLImport::GetWidthHeightAndDefinitionID(int *width, int *height, int *defID, const char** atts)
{
	*width = 0;
	*height = 0;
	*defID = -1;

	char **currentAtt = (char **)atts;
	while (*currentAtt != NULL)
	{
		if (stricmp(*currentAtt, WIDTH) != 0)
		{
			if (stricmp(*currentAtt, HEIGHT) != 0)
			{
				if (stricmp(*currentAtt, ID) != 0)
				{
					Error(TEXTUREBUFFER_ERROR1);
				}
				else
				{
					//we have a defID
					GetInt((char*) *(currentAtt + 1), defID);
				}
			}
			else
			{
				//we have a height
				GetInt((char*) *(currentAtt + 1), height);
			}
		}
		else
		{
			//we have a width
			GetInt((char*) *(currentAtt + 1), width);
		}

		currentAtt += 2;
	}
}

void
CXGLImport::MaterialStateCharDataHandler(const char*s)
{
	switch(CurrentMaterialState)
	{
	case EmissiveState:
		{
			if(!Get3FloatVector(s, &(CurrentMaterial.EmissiveR), &(CurrentMaterial.EmissiveG), &(CurrentMaterial.EmissiveB)))
			{
				return;
			}

			if((fabs(CurrentMaterial.EmissiveR -0.5) > 0.50001) || (fabs(CurrentMaterial.EmissiveG - 0.5) > 0.50001) || (fabs(CurrentMaterial.EmissiveB - 0.5) > 0.50001))
			{
				Error(MAT_EMISSIVE_RANGE_ERROR);
			}

			break;
		}
	case DiffuseState:
		{
			if(!Get3FloatVector(s, &(CurrentMaterial.DiffuseR), &(CurrentMaterial.DiffuseG), &(CurrentMaterial.DiffuseB)))
			{
				return;
			}

			if((fabs(CurrentMaterial.DiffuseR -0.5) > 0.50001) || (fabs(CurrentMaterial.DiffuseG -0.5) > 0.50001) || (fabs(CurrentMaterial.DiffuseB -0.5) > 0.50001))
			{
				Error(MAT_DIFFUSE_RANGE_ERROR);
			}

			break;
		}
	case SpecularState:
		{
			if(!Get3FloatVector(s, &(CurrentMaterial.SpecularR), &(CurrentMaterial.SpecularG), &(CurrentMaterial.SpecularB)))
			{
				return;
			}

			if((fabs(CurrentMaterial.SpecularR -0.5) > 0.50001) || (fabs(CurrentMaterial.SpecularG -0.5) > 0.50001) || (fabs(CurrentMaterial.SpecularB -0.5) > 0.50001))
			{
				Error(MAT_SPECULAR_RANGE_ERROR);
			}

			break;
		}
	case AmbientState:
		{
			if(!Get3FloatVector(s, &(CurrentMaterial.AmbientR), &(CurrentMaterial.AmbientG), &(CurrentMaterial.AmbientB)))
			{
				return;
			}

			if((fabs(CurrentMaterial.AmbientR -0.5) > 0.50001) || (fabs(CurrentMaterial.AmbientG -0.5) > 0.50001) || (fabs(CurrentMaterial.AmbientB -0.5) > 0.50001))
			{
				Error(MAT_AMBIENT_RANGE_ERROR);
			}
			break;
		}
	case ShininessState:
		{
			if(!GetFloat(s, &(CurrentMaterial.Shininess)))
			{
				return;
			}
			
			if(fabs(CurrentMaterial.Shininess - 64.0) > 64.0001)
			{
				Error(MAT_SHININESS_RANGE_ERROR);
			}

			break;
		}
	case AlphaState:
		{
			if(!GetFloat(s, &(CurrentMaterial.Alpha)))
			{
				return;
			}

			if(fabs(CurrentMaterial.Alpha -0.5) > 0.50001)
			{
				Error(MAT_ALPHA_RANGE_ERROR);
			}
				
			break;
		}
	default:
		{
			break;
		}
	}
}

void
CXGLImport::LineStyleStateCharDataHandler(const char*s)
{
	switch(CurrentLineStyleState)
	{
	case LineStyleWidthState:
		{
			if (!GetFloat(s, &(CurrentLineStyle.Width)))
			{
				return;
			}

			if (CurrentLineStyle.Width <= 0.0)
			{
				Error(LINESTYLE_WIDTH_RANGE_ERROR);
			}

			break;
		}
	case LineStylePatternState:
		{
			if (strlen(s) != 4)
			{
				Error(LINESTYLE_PATTERN_RANGE_ERROR);
			}
			else
			{
				unsigned char data[2];
				if (!GetBinaryDataFromHex(4, s, data))
				{
					return;
				}

				CurrentLineStyle.Pattern = (data[0] << 8) + data[1];
			}

			break;
		}
	case LineStylePatternFactorState:
		{
			if (!GetUnsignedInt(s, &(CurrentLineStyle.PatternFactor)))
			{
				return;
			}

			if (CurrentLineStyle.PatternFactor < 1 || CurrentLineStyle.PatternFactor > 256)
			{
				Error(LINESTYLE_PATTERNFACTOR_RANGE_ERROR);
			}

			break;
		}
	default:
		{
			break;
		}
	}
}

void
CXGLImport::PointStyleStateCharDataHandler(const char *s)
{
	switch(CurrentPointStyleState)
	{
	case PointStyleSizeState:
		{
			if (!GetFloat(s, &(CurrentPointStyle.Size)))
			{
				return;
			}

			if (CurrentPointStyle.Size <= 0.0)
			{
				Error(POINTSTYLE_SIZE_RANGE_ERROR);
			}

			break;
		}
	default:
		{
			break;
		}
	}
}

void
CXGLImport::BackgroundStateCharDataHandler(const char* s)
{
	switch(CurrentBackgroundState)
	{
	case BackgroundColorState:
		{
			if(!Get3FloatVector(s, &(CurrentBackground->Red), &(CurrentBackground->Green), &(CurrentBackground->Blue)))
			{	
				return;
			}	

			if((fabs(CurrentBackground->Red -0.5) > 0.50001) || (fabs(CurrentBackground->Green -0.5) > 0.50001))
			{
				Error(BACKGROUND_COLOR_RANGE);
			}

			break;
		}
	}

	CurrentBackgroundState = OtherBackgroundState;
}

void
CXGLImport::LightingStateCharDataHandler(const char *s)
{
	switch(CurrentLightingState)
	{
	case LightingAmbientState:
		{
			if(!Get3FloatVector(s, &(CurrentLighting->AmbientR), &(CurrentLighting->AmbientG), &(CurrentLighting->AmbientB)))
			{
				return;
			}

			if((fabs(CurrentLighting->AmbientR -0.5) > 0.50001) || (fabs(CurrentLighting->AmbientG - 0.5) > 0.50001) || (fabs(CurrentLighting->AmbientB - 0.5) > 0.50001))
			{
				Error(LIGHT_AMBIENT_RANGE_ERROR);
			}

			break;
		}
	case LightingDirectionState:
		{
			if(!Get3FloatVector(s, &(CurrentLight->DX), &(CurrentLight->DY), &(CurrentLight->DZ)))
			{
				return;
			}

			if(!GetNormalized3FloatVector(CurrentLight->DX, CurrentLight->DY, CurrentLight->DZ, &(CurrentLight->DX), &(CurrentLight->DY), &(CurrentLight->DZ)))
			{
				return;
			}

			break;
		}
	case LightingDiffuseState:
		{
			if(!Get3FloatVector(s, &(CurrentLight->DiffuseR), &(CurrentLight->DiffuseG), &(CurrentLight->DiffuseB)))
			{
				return;
			}

			if((fabs(CurrentLight->DiffuseR -0.5) > 0.50001) || (fabs(CurrentLight->DiffuseG - 0.5) > 0.50001) || (fabs(CurrentLight->DiffuseB - 0.5) > 0.50001))
			{
				Error(LIGHT_DIFFUSE_RANGE_ERROR);
			}

			break;
		}
	case LightingSpecularState:
		{
			if(!Get3FloatVector(s, &(CurrentLight->SpecularR), &(CurrentLight->SpecularG), &(CurrentLight->SpecularB)))
			{
				return;
			}

			if((fabs(CurrentLight->SpecularR -0.5) > 0.50001) || (fabs(CurrentLight->SpecularG - 0.5) > 0.50001) || (fabs(CurrentLight->SpecularB - 0.5) > 0.50001))
			{
				Error(LIGHT_SPECULAR_RANGE_ERROR);
			}

			break;
		}
	}

	CurrentLightingState = OtherLightingState;
}

void
CXGLImport::MeshStateCharDataHandler(const char*s)
{
	switch(CurrentMeshState)
	{
	case MeshPositionState:
		{	
			float x, y, z;
			if(!Get3FloatVector(s, &x, &y, &z))
			{
				return;
			}

			if (CurrentVertexID != -1)
			{
				//this is a defined position
				CXGLVertex *v = new CXGLVertex;
				v->X = x;
				v->Y = y;
				v->Z = z;
				Resolver.AddVertexDefinition(CurrentContext, CurrentVertexID, v);
			}
			else
			{
				//we shouldn't have defined positions here
				assert(0);
			}

			break;
		}
	case MeshNormalState:
		{
			float x, y, z;
			if(!Get3FloatVector(s, &x, &y, &z))
			{
				return;
			}

			if (CurrentNormalID != -1)
			{
				float newX, newY, newZ;

				if(!GetNormalized3FloatVector(x, y, z, &newX, &newY, &newZ))
				{
					return;
				}
				
				//this is a defined normal
				CXGLNormal *normal = new CXGLNormal;
				normal->X = newX;
				normal->Y = newY;
				normal->Z = newZ;
				Resolver.AddNormalDefinition(CurrentContext, CurrentNormalID, normal);
			}

			break;
		}
	case MeshTextureCoordinateState:
		{
			float u, v;
			if(!Get2FloatVector(s, &u, &v))
			{
				return;
			}

			if (CurrentTextureCoordinateID != -1)
			{
				//this is a defined texture coordinate
				CXGLTextureCoordinate *tc = new CXGLTextureCoordinate;
				tc->U = u;
				tc->V = v;
				Resolver.AddTextureCoordinateDefinition(CurrentContext, CurrentTextureCoordinateID, tc);
			}
			break;
		}
	case MeshTextureRGBState:
		{
			unsigned int length = strlen(s);
			
			//make sure that the width and height are correct
			if (length != (CurrentTextureBuffer->TextureWidth*CurrentTextureBuffer->TextureHeight*6))
			{
				Error(TEXTUREBUFFER_ERROR5);
			}
			else
			{
				int numPixels = (int)((length*0.5)/3);
				CurrentTextureBuffer->RGBABuffer = new unsigned char[numPixels*4]; //for the alpha

				char string[3];

				int i;
				for (i = 0; i < numPixels; i ++)
				{
					memcpy(string, s+6*i+0, 2);
					string[2] = 0;
					int result = strtoul(string, NULL ,16); 
					CurrentTextureBuffer->RGBABuffer[4*i] = (unsigned char)result;

					memcpy(string, s+6*i+2, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+1] = (unsigned char)result;

					memcpy(string, s+6*i+4, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+2] = (unsigned char)result;

					//the alpha bit
					CurrentTextureBuffer->RGBABuffer[4*i+3] = (unsigned char)255;
				}
			}

			break;
		}
	case MeshTextureRGBAState:
		{
			unsigned int length = strlen(s);
			
			//make sure that the width and height are correct
			if (length != (CurrentTextureBuffer->TextureWidth*CurrentTextureBuffer->TextureHeight*8))
			{
				Error(TEXTUREBUFFER_ERROR8);
			}
			else
			{
				int numPixels = (int)((length*0.5)/4);
				CurrentTextureBuffer->RGBABuffer = new unsigned char[numPixels*4]; //for the alpha

				char string[3];

				int i;
				for (i = 0; i < numPixels; i ++)
				{
					memcpy(string, s+8*i+0, 2);
					string[2] = 0;
					int result = strtoul(string, NULL ,16); 
					CurrentTextureBuffer->RGBABuffer[4*i] = (unsigned char)result;

					memcpy(string, s+8*i+2, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+1] = (unsigned char)result;

					memcpy(string, s+8*i+4, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+2] = (unsigned char)result;

					memcpy(string, s+8*i+6, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+3] = (unsigned char)result;
				}
			}

			break;
		}
	case MeshExtGeomState:
		{
		}
	default:
		{
			break;
		}
	}
}


void
CXGLImport::FaceVertexStateCharDataHandler(const char* s)
{
	switch(CurrentFVState)
	{
	case FVPositionState:
		{			
			float x, y, z;

			if(!Get3FloatVector(s, &x, &y, &z))
			{
				return;
			}		
			
			if (CurrentVertexID != -1)
			{
				CXGLVertex *vertex = new CXGLVertex;
				vertex->X = x;
				vertex->Y = y;
				vertex->Z = z;
				Resolver.AddVertexDefinition(CurrentContext, CurrentVertexID, vertex);
			}
			else
			{
				CXGLVertex *v = Vertices.GetLastElementRef();
				v->XGLID = -1;
				v->X = x;
				v->Y = y;
				v->Z = z;

				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].VertexI = Vertices.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].VertexI = Vertices.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.VertexI = Vertices.GetNumElements() - 1;
				}
				else
				{
					assert(0);
				}
				
				Vertices.SetSize(Vertices.GetNumElements() + 1);
			}
			
			break;
		}
	case FVNormalState:
		{
			float x, y, z, newX, newY, newZ;
			if(!Get3FloatVector(s, &x, &y, &z))
			{
				return;
			}
			
			if(!GetNormalized3FloatVector(x, y, z, &newX, &newY, &newZ))
			{
				return;
			}
			
			
			if (CurrentNormalID != -1)
			{
				CXGLNormal *normal = new CXGLNormal;
				normal->X = newX;
				normal->Y = newY;
				normal->Z = newZ;
				//this is a defined normal

				Resolver.AddNormalDefinition(CurrentContext, CurrentNormalID, normal);
			}
			else
			{
				CXGLNormal *n = Normals.GetLastElementRef();
				n->XGLID = -1;
				n->X = x;
				n->Y = y;
				n->Z = z;

				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].NormalI = Normals.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].NormalI = Normals.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.NormalI = Normals.GetNumElements() - 1;
				}
				else
				{
					assert(0);
				}
	
				Normals.SetSize(Normals.GetNumElements() + 1);
			}

			break;
		}
	case FVTextureCoordinateState:
		{
			float u, v;
			if(!Get2FloatVector(s, &u, &v))
			{
				return;
			}

			if (CurrentTextureCoordinateID != -1)
			{
				CXGLTextureCoordinate *tc = new CXGLTextureCoordinate;
				tc->U = u;
				tc->V = v;
				//this is a defined texture coordinate

				Resolver.AddTextureCoordinateDefinition(CurrentContext, CurrentTextureCoordinateID, tc);
			}
			else
			{
				CXGLTextureCoordinate *tc = TextureCoordinates.GetLastElementRef();
				tc->XGLID = -1;
				tc->U = u;
				tc->V = v;

				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].TextureCoordinateI = TextureCoordinates.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].TextureCoordinateI = TextureCoordinates.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.TextureCoordinateI = TextureCoordinates.GetNumElements() - 1;
				}
				else
				{
					assert(0);
				}

				TextureCoordinates.SetSize(TextureCoordinates.GetNumElements() + 1);
			}
			break;
		}
	case FVNormalRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}
			
			if (referenceID != PrevNormalReferenceID)
			{
				PrevNormalReferenceID = referenceID;
				
				CXGLNormal *n = Normals.GetLastElementRef();
				n->XGLID = referenceID;
				
				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].NormalI = Normals.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].NormalI = Normals.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.NormalI = Normals.GetNumElements() - 1;
				}
				else
				{
					assert(0);
				}
	
				Normals.SetSize(Normals.GetNumElements() + 1);
			}
			else
			{
				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].NormalI = Normals.GetNumElements() - 2;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].NormalI = Normals.GetNumElements() - 2;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.NormalI = Normals.GetNumElements() - 2;
				}
				else
				{
					assert(0);
				}
			}

			break;
		}
	case FVVertexRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}
			
			if (referenceID != PrevVertexReferenceID)
			{
				PrevVertexReferenceID = referenceID;
				
				CXGLVertex *v = Vertices.GetLastElementRef();
				v->XGLID = referenceID;
				
				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].VertexI = Vertices.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].VertexI = Vertices.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.VertexI = Vertices.GetNumElements() - 1;
				}
				else
				{
					assert(0);
				}
	
				Vertices.SetSize(Vertices.GetNumElements() + 1);
			}
			else
			{
				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].VertexI = Vertices.GetNumElements() - 2;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].VertexI = Vertices.GetNumElements() - 2;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.VertexI = Vertices.GetNumElements() - 2;
				}
				else
				{
					assert(0);
				}
			}

			break;
		}
	case FVTextureCoordinateRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			if (referenceID != PrevTextureCoordinateReferenceID)
			{
				PrevTextureCoordinateReferenceID = referenceID;

				CXGLTextureCoordinate *tc = TextureCoordinates.GetLastElementRef();
				tc->XGLID = referenceID;

				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].TextureCoordinateI = TextureCoordinates.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].TextureCoordinateI = TextureCoordinates.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.TextureCoordinateI = TextureCoordinates.GetNumElements() - 1;
				}
				else
				{
					assert(0);
				}

				TextureCoordinates.SetSize(TextureCoordinates.GetNumElements() + 1);
			}
			else
			{
				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->FaceVertices[CurrentFaceVertex].TextureCoordinateI = TextureCoordinates.GetNumElements() - 2;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->LineVertices[CurrentFaceVertex].TextureCoordinateI = TextureCoordinates.GetNumElements() - 2;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					assert(CurrentFaceVertex == 0);
					p->PointVertex.TextureCoordinateI = TextureCoordinates.GetNumElements() - 2;
				}
				else
				{
					assert(0);
				}
			}
			
			break;
		}
	}
}


void 
CXGLImport::FaceStateCharDataHandler(const char *s)
{
	switch(CurrentFaceState)
	{
	case FaceTextureRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			CXGLTextureHandle *t = Resolver.CreateTextureReference(CurrentContext, referenceID);
			
			if (referenceID != PrevTextureReferenceID)
			{
				PrevTextureReferenceID = referenceID;

				CurrentMesh->TextureHandles.SetElement(CurrentMesh->TextureHandles.GetNumElements() - 1, t);
				
				CXGLFace *f = Faces.GetLastElementRef();
				f->TextureI = CurrentMesh->TextureHandles.GetNumElements() - 1;
			
				CurrentMesh->TextureHandles.SetSize(CurrentMesh->TextureHandles.GetNumElements() + 1);
			}
			else
			{
				CXGLFace *f = Faces.GetLastElementRef();
				f->TextureI = CurrentMesh->TextureHandles.GetNumElements() - 2;
			}

			break;
		}
			
	case FaceMaterialRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			if (referenceID != PrevMatReferenceID)
			{
				PrevMatReferenceID = referenceID;

				CXGLMaterial *m = Materials.GetLastElementRef();
				m->XGLID = referenceID;
				
				CXGLFace *f = Faces.GetLastElementRef();
				f->MaterialI = Materials.GetNumElements() - 1;
				
				Materials.SetSize(Materials.GetNumElements() + 1);
			}
			else
			{
				CXGLFace *f = Faces.GetLastElementRef();
				f->MaterialI = Materials.GetNumElements() - 2;
			}

			break;
		}
	case FaceSmoothState:
		{
			unsigned int groupNum;
			if(!GetUnsignedInt(s, &groupNum))
			{
				return;
			}
		
			CXGLFace *f = Faces.GetLastElementRef();
			
			if(!SetBit(groupNum, &(f->SmoothingGroups)))
			{
				Error(SMOOTHING_GROUP_OUT_OF_RANGE);
				return;
			}
		}
	default:
		{
			break;
		}
	}
}

void 
CXGLImport::LineStateCharDataHandler(const char *s)
{
	switch(CurrentLineState)
	{
	case LineTextureRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			CXGLTextureHandle *t = Resolver.CreateTextureReference(CurrentContext, referenceID);
			
			if (referenceID != PrevTextureReferenceID)
			{
				PrevTextureReferenceID = referenceID;

				CurrentMesh->TextureHandles.SetElement(CurrentMesh->TextureHandles.GetNumElements() - 1, t);
				
				CXGLLine *l = Lines.GetLastElementRef();
				l->TextureI = CurrentMesh->TextureHandles.GetNumElements() - 1;
			
				CurrentMesh->TextureHandles.SetSize(CurrentMesh->TextureHandles.GetNumElements() + 1);
			}
			else
			{
				CXGLLine *l = Lines.GetLastElementRef();
				l->TextureI = CurrentMesh->TextureHandles.GetNumElements() - 2;
			}

			break;
		}
			
	case LineMaterialRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			if (referenceID != PrevMatReferenceID)
			{
				PrevMatReferenceID = referenceID;

				CXGLMaterial *m = Materials.GetLastElementRef();
				m->XGLID = referenceID;
				
				CXGLLine *l = Lines.GetLastElementRef();
				l->MaterialI = Materials.GetNumElements() - 1;
				
				Materials.SetSize(Materials.GetNumElements() + 1);
			}
			else
			{
				CXGLLine *l = Lines.GetLastElementRef();
				l->MaterialI = Materials.GetNumElements() - 2;
			}

			break;
		}
	case LineStyleRefState:
		{
			unsigned int referenceID;
			if (!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			if (referenceID != PrevLineStyleReferenceID)
			{
				PrevLineStyleReferenceID = referenceID;

				CXGLLineStyle *ls = LineStyles.GetLastElementRef();
				ls->XGLID = referenceID;

				CXGLLine *l = Lines.GetLastElementRef();
				l->LineStyleI = LineStyles.GetNumElements() - 1;

				LineStyles.SetSize(LineStyles.GetNumElements() + 1);
			}
			else
			{
				CXGLLine *l = Lines.GetLastElementRef();
				l->LineStyleI = LineStyles.GetNumElements() - 2;
			}

			break;
		}
	default:
		{
			break;
		}
	}
}

void 
CXGLImport::PointStateCharDataHandler(const char *s)
{
	switch(CurrentPointState)
	{
	case PointTextureRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			CXGLTextureHandle *t = Resolver.CreateTextureReference(CurrentContext, referenceID);
			
			if (referenceID != PrevTextureReferenceID)
			{
				PrevTextureReferenceID = referenceID;

				CurrentMesh->TextureHandles.SetElement(CurrentMesh->TextureHandles.GetNumElements() - 1, t);
				
				CXGLPoint *p = Points.GetLastElementRef();
				p->TextureI = CurrentMesh->TextureHandles.GetNumElements() - 1;
			
				CurrentMesh->TextureHandles.SetSize(CurrentMesh->TextureHandles.GetNumElements() + 1);
			}
			else
			{
				CXGLPoint *p = Points.GetLastElementRef();
				p->TextureI = CurrentMesh->TextureHandles.GetNumElements() - 2;
			}

			break;
		}
			
	case PointMaterialRefState:
		{
			unsigned int referenceID;
			if(!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			if (referenceID != PrevMatReferenceID)
			{
				PrevMatReferenceID = referenceID;

				CXGLMaterial *m = Materials.GetLastElementRef();
				m->XGLID = referenceID;
				
				CXGLPoint *p = Points.GetLastElementRef();
				p->MaterialI = Materials.GetNumElements() - 1;
				
				Materials.SetSize(Materials.GetNumElements() + 1);
			}
			else
			{
				CXGLPoint *p = Points.GetLastElementRef();
				p->MaterialI = Materials.GetNumElements() - 2;
			}

			break;
		}
	case PointStyleRefState:
		{
			unsigned int referenceID;
			if (!GetUnsignedInt(s, &referenceID))
			{
				return;
			}

			if (referenceID != PrevPointStyleReferenceID)
			{
				PrevPointStyleReferenceID = referenceID;

				CXGLPointStyle *ps = PointStyles.GetLastElementRef();
				ps->XGLID = referenceID;

				CXGLPoint *p = Points.GetLastElementRef();
				p->PointStyleI = PointStyles.GetNumElements() - 1;

				PointStyles.SetSize(PointStyles.GetNumElements() + 1);
			}
			else
			{
				CXGLPoint *p = Points.GetLastElementRef();
				p->PointStyleI = PointStyles.GetNumElements() - 2;
			}

			break;
		}
	default:
		{
			break;
		}
	}
}

void 
CXGLImport::DataStateCharDataHandler(const char *s)
{
	switch (CurrentDataState)
	{
	case StringState:
		{
			//handle string data
			//get the length of the string and write it into the resizeable array
			unsigned int length = strlen(s) + 1;
			StringLengths.AddElement(length);

			//get and write the string into the resizeable array
			char *str = new char[length];
			memcpy(str, s, length);
			str[length-1] = 0;

			Strings.AddElement(str);
			break;
		}

	case BinState:
		{
			//handle binary data
			//get the length of the string and write it into the resizeable array
			unsigned int length = strlen(s);

			if((length % 2) != 0)
			{
				Error(BINARY_DATA_INVALID_LENGTH);
				break;
			}

			int dataLength = length/2;

			unsigned char *b = new unsigned char[dataLength];
			
			int success = GetBinaryDataFromHex(length, s, b);

			if(!success)
			{
				delete [] b;
				break;
			}

			BinLengths.AddElement(dataLength);

			//get and write the string into the resizeable array
			Bins.AddElement(b);
			break;
		}
	}
}

void 
CXGLImport::TextureStateCharDataHandler(const char *s)
{
	switch(CurrentTextureState)
	{
	case TextureRGBRefState:
		{
			int id;
			if(!GetInt(s, &id))
			{
				return;
			}

			CurrentTexture->TextureBufferHandle = Resolver.CreateTextureBufferReference(CurrentContext, id);
			break;
		}
	case TextureRGBARefState:
		{
			int id;
			if(!GetInt(s, &id))
			{		
				return;
			}	

			CurrentTexture->TextureBufferHandle = Resolver.CreateTextureBufferReference(CurrentContext, id);
			break;
		}
	case TextureBorderColorState:
		{
			float r, g, b, a;
			if(!Get4FloatVector(s, &r, &g, &b, &a))
			{
				return;
			}

			CurrentTexture->UseTextureBorderColor = 1;
			CurrentTexture->TextureBorderColor[0] = r;
			CurrentTexture->TextureBorderColor[1] = g;
			CurrentTexture->TextureBorderColor[2] = b;
			CurrentTexture->TextureBorderColor[3] = a;

			break;
		}
	case TextureRGBState:
		{
			unsigned int length = strlen(s);
			
			//make sure that the width and height are correct
			if (length != (CurrentTextureBuffer->TextureWidth*CurrentTextureBuffer->TextureHeight*6))
			{
				Error(TEXTUREBUFFER_ERROR8);
			}
			else
			{
				int numPixels = (int)((length*0.5)/3);
				CurrentTextureBuffer->RGBABuffer = new unsigned char[numPixels*4]; //for the alpha

				char string[3];
				
				int i;
				for (i = 0; i < numPixels; i ++)
				{
					memcpy(string, s+6*i+0, 2);
					string[2] = 0;
					int result = strtoul(string, NULL ,16); 
					CurrentTextureBuffer->RGBABuffer[4*i] = (unsigned char)result;

					memcpy(string, s+6*i+2, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+1] = (unsigned char)result;

					memcpy(string, s+6*i+4, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+2] = (unsigned char)result;

					//the alpha bit
					CurrentTextureBuffer->RGBABuffer[4*i+3] = (unsigned char)255;
				}
			}

			break;
		}
	case TextureRGBAState:
		{
			unsigned int length = strlen(s);
			
			//make sure that the width and height are correct
			if (length != (CurrentTextureBuffer->TextureWidth*CurrentTextureBuffer->TextureHeight*8))
			{
				Error(TEXTUREBUFFER_ERROR8);
			}
			else
			{
				int numPixels = (int)((length*0.5)/4);
				CurrentTextureBuffer->RGBABuffer = new unsigned char[numPixels*4]; //for the alpha

				char string[3];

				int i;
				for (i = 0; i < numPixels; i ++)
				{
					memcpy(string, s+8*i+0, 2);
					string[2] = 0;
					int result = strtoul(string, NULL ,16); 
					CurrentTextureBuffer->RGBABuffer[4*i] = (unsigned char)result;

					memcpy(string, s+8*i+2, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+1] = (unsigned char)result;

					memcpy(string, s+8*i+4, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+2] = (unsigned char)result;

					memcpy(string, s+8*i+6, 2);
					string[2] = 0;
					result = strtoul(string, NULL, 16);
					CurrentTextureBuffer->RGBABuffer[4*i+3] = (unsigned char)result;
				}
			}

			break;
		}
	}
}

void
CXGLImport::SpaceStateCharDataHandler(const char *s)
{


}

void 
CXGLImport::DataStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(DATA_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case StringTag:
		{
			//do nothing
			break;
		}

	case BinTag:
		{
			//do nothing
			break;
		}
	case DataTag:
		{
			int numBin = DataRequires.HowMany(BinTag);
			int numString = DataRequires.HowMany(StringTag);

			if((numBin + numString) < 1)
			{
				Error(DATA_REQUIRES);
				return;
			}
				
			//create the data from the resizeable arrays
			CreateCurrentDataFromResizeableArrays();

			//determine if this data is associated with a patch, object, or world
			//and then store the data handle in one of them
			if (CurrentPatchIndex == -1)
			{
				//let's send off the data we've accumulated to the object or world as there is no patch
				CObjectStackElem *top = ObjectStack.GetTop();
				if (top == NULL)
				{
					//send the data to the world
					CurrentWorld->DataHandles.SetSize(CurrentWorld->NumDatas + 1);

					//resolve the data and set the returned handle in the world's handle array
					CurrentWorld->DataHandles.SetElement(CurrentWorld->NumDatas, Resolver.AddData(CurrentContext, CurrentData));
					
					//increment the number of datas in the world
					CurrentWorld->NumDatas++;
				}
				else
				{
					//set the datahandles array to be one greater
					top->Object->DataHandles.SetSize(top->Object->NumDatas + 1);

					//resolve the data and set the returned handle in the object's handle array
					top->Object->DataHandles.SetElement(top->Object->NumDatas, Resolver.AddData(CurrentContext, CurrentData));

					//increment the number of datas in the object
					top->Object->NumDatas++;
				}
			}
			else
			{
				//let's send off the data we've accumulated to the patch
				CXGLPatch *patch = Patches.GetLastElement();

				//set the datahandles array to be one greater
				patch->DataHandles.SetSize(patch->NumDatas + 1);

				//resolve the data and set the returned handle in the patch's handle array
				patch->DataHandles.SetElement(patch->NumDatas, Resolver.AddData(CurrentContext, CurrentData));

				CurrentData = NULL;
			
				//increment the number of datas in the patch
				patch->NumDatas++;
			}

			//pop up to our parent state
			CurrentState = StateStack.Pop();

			//uninitialize the data
			UnInitializeCurrentData();
			break;
		}
	}

	CurrentDataState = OtherDataState;
}


void
CXGLImport::ExternalObjectStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(EXTERNAL_OBJECT_NEEDS_DATA);
		return;
	}

	switch(tag)
	{
	
	case ExternalObjectRefTag:
	case ExternalObjectStaticRefTag:
		{

			CElementRequires* objReq = &(ObjectStack.GetTop()->ObjectRequires);

			int numExtents = objReq->HowMany(ExtentsTag);
			
			int isStatic = ObjectStack.GetTop()->Object->IsStatic;

			if(!isStatic)
			{
				if(numExtents != 1)
				{
					Error(EXTERNAL_OBJECT_EXTENTS_REQUIRES);
					return;
				}
			}

			int numRefs = objReq->HowMany(RefTag);

			if(numRefs != 1)
			{
				Error(EXTERNAL_OBJECT_REF_REQUIRES);
				return;
			}

			int numRefTypes = objReq->HowMany(RefTypeTag);
			
			if(numRefTypes != 1)
			{
				Error(EXTERNAL_OBJECT_REFTYPE_REQUIRES);
				return;
			}

			int numTransforms = objReq->HowMany(TransformTag);

			if(numTransforms > 1)
			{
				Error(EXTERNAL_OBJECT_TRANSFORM_REQUIRES);
				return;
			}

			CXGLObject *object;
			object = ObjectStack.Pop();

			CXGLContext *parentContext = CurrentContext->GetParentContext();

			//give the object to the resolver		
			assert(object->XGLID == -1);

			CXGLObjectHandle *objectHandle;
			objectHandle = Resolver.AddObject(CurrentContext, object);
		
			// store the object handle
			CObjectStackElem *parent = ObjectStack.GetTop();
			
			//if there's no object on the stack, store the handle in the world
			if (parent == NULL)
			{
				if (CurrentWorld->MeshHandle != NULL)
				{
					Error(WORLD_MESH_SUB_OBJ_ERROR);
					return;
				}

				//if there is a user specified pathID
				if (objectHandle->ObjectPathID != -1)
				{
					int *value = TopObjectPathIDHash.GetMatchingValue(&objectHandle->ObjectPathID);
					if (value)
					{
						Error(UNIQUE_PATHID_ERROR);
						return;
					}

					//add the pathid to the hash
					TopObjectPathIDHash.Add(&objectHandle->ObjectPathID);
				}

				//add the new object to the world
				CurrentWorld->ObjectHandles.AddElement(objectHandle);
				CurrentWorld->NumObjects++;
			}
			else
			{
				if(parent->Object->MeshHandle != NULL)
				{
					Error(OBJECT_MESH_SUB_OBJ_ERROR);
					return;
				}

				//if there is a user specified pathID
				if (objectHandle->ObjectPathID != -1)
				{
					int *value = parent->PathIDHash.GetMatchingValue(&objectHandle->ObjectPathID);
					if (value)
					{
						Error(UNIQUE_PATHID_ERROR);
						return;
					}

					//add the pathid to the hash
					parent->PathIDHash.Add(&objectHandle->ObjectPathID);
				}

				//add the new object to the parent
				parent->Object->ObjectHandles.AddElement(objectHandle);
				parent->Object->NumObjects++;
			}
			
			CurrentState = StateStack.Pop();

			//pop up in the contexts
			CurrentContext = Resolver.ReleaseContext(CurrentContext);

			break;
		}
	case NameTag:
		{
			CObjectStackElem *parent = ObjectStack.GetTop();	
			assert(parent != NULL);

			//if we have an empty name tag put in an empty string
			if (parent->Object->Name == NULL)
			{
				parent->Object->Name = new char[strlen("") + 1];
				strcpy(parent->Object->Name, "");
			}
		
			break;
		}
	}

	CurrentExternalObjectState = OtherExternalState;
}

void
CXGLImport::TextureStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(TEXTURE_NEEDS_DATA);
		return;
	}


	switch(tag)
	{
	case TextureTag:
		{
			//verify that we have everything that we need

			int numImages = TextureRequires.HowMany(TextureRGBTag);
			numImages += TextureRequires.HowMany(TextureRGBATag);
			numImages += TextureRequires.HowMany(TextureRGBRefTag);
			numImages += TextureRequires.HowMany(TextureRGBARefTag);

			if(numImages != 1)
			{
				Error(TEXTURE_IMAGE_REQUIRES);
				return;
			}

			int numTexFuns = TextureRequires.HowMany(ReplaceTag);
			numTexFuns += TextureRequires.HowMany(ModulateTag);
			numTexFuns += TextureRequires.HowMany(DecalTag);

			if(numTexFuns != 1)
			{
				Error(TEXTURE_FUNCTION_REQUIRES);
				return;
			}

			int numWrapFuns = TextureRequires.HowMany(RepeatTag);
			numWrapFuns += TextureRequires.HowMany(ClampTag);

			if(numWrapFuns != 1)
			{
				Error(TEXTURE_WRAP_REQUIRES);
				return;
			}

			int numBorderColors = TextureRequires.HowMany(TextureBorderColorTag);

			if(numBorderColors > 1)
			{
				Error(TEXTURE_BORDER_COLOR);
				return;
			}

			if (CurrentTextureID != -1)
			{
				Resolver.AddTextureDefinition(CurrentContext, CurrentTextureID, CurrentTexture);
				CurrentTexture = NULL; //Uncommented this for cleaning purposes - achase 12/29/1999
			}
			else
			{
				CXGLTextureHandle *textureHandle = Resolver.AddTexture(CurrentContext, CurrentTexture);
				CurrentTexture = NULL;

				CXGLTextureHandle *t = CurrentMesh->TextureHandles.GetLastElement();
				memcpy(t, textureHandle, sizeof(*t));
				
				if (StateStack.GetTopValue() == FaceState)
				{
					CXGLFace *f = Faces.GetLastElementRef();
					f->MaterialI = CurrentMesh->TextureHandles.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == LineState)
				{
					CXGLLine *l = Lines.GetLastElementRef();
					l->MaterialI = CurrentMesh->TextureHandles.GetNumElements() - 1;
				}
				else if (StateStack.GetTopValue() == PointState)
				{
					CXGLPoint *p = Points.GetLastElementRef();
					p->MaterialI = CurrentMesh->TextureHandles.GetNumElements() - 1;
				}
				else
				{
					assert(0);
				}
		
				CurrentMesh->TextureHandles.SetSize(CurrentMesh->TextureHandles.GetNumElements() + 1);
				
				//we should never have CurrentContext of NULL, b/c that indicates
				//that we have a texture in the world that is not defined

				assert(CurrentContext->GetParentContext() != NULL);
			}

			//pop up to our parent state
			CurrentState = StateStack.Pop();
		
			break;
		}
	case TextureRGBTag:
		{
			if (CurrentTextureBufferID != -1)
			{
				Resolver.AddTextureBufferDefinition(CurrentContext, CurrentTextureBufferID, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
			else
			{
				CurrentTexture->TextureBufferHandle = Resolver.AddTextureBuffer(CurrentContext, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
				
			CurrentTextureBufferID = -1;
			
			break;
		}
	case TextureRGBATag:
		{
			if (CurrentTextureBufferID != -1)
			{
				Resolver.AddTextureBufferDefinition(CurrentContext, CurrentTextureBufferID, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
			else
			{
				CurrentTexture->TextureBufferHandle = Resolver.AddTextureBuffer(CurrentContext, CurrentTextureBuffer);
				CurrentTextureBuffer = NULL;
			}
				
			CurrentTextureBufferID = -1;
			
			break;
		}
	case TextureRGBRefTag:
		{
			assert(CurrentTexture->TextureBufferHandle != NULL);
			break;
		}
	case TextureRGBARefTag:
		{
			assert(CurrentTexture->TextureBufferHandle != NULL);
			break;
		}
	}

	CurrentTextureState = OtherTextureState;
}

extern "C"
void
CharElement(void* userData, const char *s, int len)
{

	CXGLImport* importer = (CXGLImport*) userData;

	importer->ImportCharElement(s, len);
}

void
CXGLImport::ImportCharElement(const char* s, int len)
{
	
	if(IsStopped())
	{
		return;
	}

	if(CurrentParserState == CharDataState)
	{
		int copyStringLength = CopyString.GetNumElements();
		int newLen = copyStringLength + len;

		//increase the size
		CopyString.SetSize(newLen);

		//so we can write over the null character
		char* endPtr = CopyString.GetElementRef(copyStringLength - 1);
		memcpy(endPtr, s, len);
		endPtr[len] = 0;

	}
	else
	{
		CurrentParserState = CharDataState;

		CopyString.SetSize(len + 1);

		char* string = CopyString.GetElementRef(0);

		memcpy(string, s, len);
		string[len] = 0;
	}

}


void
CXGLImport::AccumCharElement()
{
	NeedsData = 0;

	char *string = CopyString.GetElementRef(0);
#ifdef TRACE_PARSE
	CString s;
	s.Format("Char: %d\r\n", CurrentState);
	OutputDebugString(s);
#endif	


	switch(CurrentState)
	{
	case FaceVertexState:
		{
			FaceVertexStateCharDataHandler(string);
			break;
		}
	case FaceState:
		{
			FaceStateCharDataHandler(string);
			break;
		}
	case LineState:
		{
			LineStateCharDataHandler(string);
			break;
		}
	case PointState:
		{
			PointStateCharDataHandler(string);
			break;
		}
	case MaterialState:
		{
			MaterialStateCharDataHandler(string);
			break;
		}
	case LineStyleState:
		{
			LineStyleStateCharDataHandler(string);
			break;
		}
	case PointStyleState:
		{
			PointStyleStateCharDataHandler(string);
			break;
		}
	case SpaceState:
		{
			SpaceStateCharDataHandler(string);
			break;
		}
	case WorldState:
		{
			WorldStateCharDataHandler(string);
			break;
		}
	case LightingState:
		{
			LightingStateCharDataHandler(string);
			break;
		}
	case BackgroundState:
		{
			BackgroundStateCharDataHandler(string);
			break;
		}
	case ObjectState:
		{
			ObjectStateCharDataHandler(string);
			break;
		}
	case TransformState:
		{
			TransformStateCharDataHandler(string);
			break;
		}
	case MeshState:
		{
		 	MeshStateCharDataHandler(string);
			break;
		}
	case DataState:
		{
			DataStateCharDataHandler(string);
			break;
		}
	case TextureState:
		{
			TextureStateCharDataHandler(string);	
			break;
		}
	case ExternalObjectReferenceState:
		{
			ExternalObjectStateCharDataHandler(string);
			break;
		}
	case ExtgeomState:
		{
			ExtgeomStateCharDataHandler(string);
			break;
		}
	case OtherState:
		{
			OtherStateCharDataHandler(string);
			break;
		}
	}
}

CXGLImport*
CXGLImport::Import(CXGLStatusCallback* status)
{
	Status = status;
	State = STATE_IMPORTING;

	//initial the patch index state
	CurrentPatchIndex = -1;

	//set our initial states
	CurrentState = SpaceState;
	CurrentParserState = OtherState;

	Parser = NULL;

	//reset our internal state
	CurrentMaterialState = OtherMaterialState;
	CurrentLineStyleState = OtherLineStyleState;
	CurrentPointStyleState = OtherPointStyleState;
	CurrentTransformState = OtherTransformState;
	CurrentDataState = OtherDataState;
	CurrentTextureState = OtherTextureState;
	CurrentLightingState = OtherLightingState;
	CurrentBackgroundState = OtherBackgroundState;
	CurrentFaceState = OtherFaceState;
	CurrentLineState = OtherLineState;
	CurrentPointState = OtherPointState;
	CurrentFVState = OtherFVState;
	CurrentMeshState = OtherMeshState;
	CurrentObjectState = OtherObjectState;
	CurrentWorldState = OtherWorldState;
	CurrentExternalObjectState = OtherExternalState;
	CurrentExtgeomState = OtherExtgeomState;

	return this;
}

int 
CXGLImport::Import(const char *filename, int fileType, CXGLStatusCallback* status)
{
	if (filename == NULL)
		return 0;

	if (fileType == OtherFileType)
		return 0;

	FileType = fileType;

	Status = status;
	State = STATE_IMPORTING;

//  Begin parsing stuff!!!	

	//create the parser
	Parser = XML_ParserCreate(NULL);
	
	//initialize the parser
	XML_SetUserData(Parser, this);
	XML_SetElementHandler(Parser, StartElement, EndElement);
	XML_SetCharacterDataHandler(Parser, CharElement);

	//initial the patch index state
	CurrentPatchIndex = -1;

	//set our initial states
	CurrentState = SpaceState;
	CurrentParserState = OtherState;

	//reset our internal state
	CurrentMaterialState = OtherMaterialState;
	CurrentLineStyleState = OtherLineStyleState;
	CurrentPointStyleState = OtherPointStyleState;
	CurrentTransformState = OtherTransformState;
	CurrentDataState = OtherDataState;
	CurrentTextureState = OtherTextureState;
	CurrentLightingState = OtherLightingState;
	CurrentBackgroundState = OtherBackgroundState;
	CurrentFaceState = OtherFaceState;
	CurrentLineState = OtherLineState;
	CurrentPointState = OtherPointState;
	CurrentFVState = OtherFVState;
	CurrentMeshState = OtherMeshState;
	CurrentObjectState = OtherObjectState;
	CurrentWorldState = OtherWorldState;
	CurrentExternalObjectState = OtherExternalState;
	CurrentExtgeomState = OtherExtgeomState;

	//convert the filename to all lower case
	//so that we're not case sensitive
	char *lowFile = strlwr(strdup(filename));

	HANDLE handle;
	handle = CreateFile(lowFile, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_DELETE, (LPSECURITY_ATTRIBUTES) NULL, OPEN_EXISTING,    
						FILE_ATTRIBUTE_READONLY  | FILE_FLAG_SEQUENTIAL_SCAN,    
						(HANDLE) NULL);

	if (handle == INVALID_HANDLE_VALUE)
	{
		int err = GetLastError();
		if (err == ERROR_SHARING_VIOLATION)
		{
			Error(FILE_IN_USE);
			Status->SetRetry();
		}
		else
		{
			sprintf(ErrorString, "97: Unable to open file: %x\n", err);
			Error(ErrorString);
		}
	}
	else
	{
		NumBytesInFile = GetFileSize(handle, NULL);

		int fileNameLength = strlen(lowFile);
		if (FileType == XGLFileType)
		{
			//Read the XGL file
			int done;
	
			do
			{
				DWORD len;
				if (ReadFile(handle, Buf, sizeof(Buf), &len, NULL) == 0)
				{
					State = STATE_STOPPED;
					Error("Error Reading File\n");
				}
				else
				{
					done = len < sizeof(Buf);
					if (!XML_Parse(Parser, Buf, len, done)) 
					{
						sprintf(ErrorString, "%s: %s at line %d\n",PARSE_ERROR, XML_ErrorString(XML_GetErrorCode(Parser)),
							 XML_GetCurrentLineNumber(Parser));
						Error(ErrorString);
					}
				}

				if(State == STATE_STOPPED)
					done = 1;
			}
			while (!done);
		}
		else if (FileType == ZGLFileType)
		{
			CZLIBDecompress decompressor;

			bool ret = false;
			bool finished = false;

			char moduleFileName[MAX_PATH];
 			GetModuleFileName(GetModuleHandle(NULL), moduleFileName, sizeof(moduleFileName));

			char *lastSlash = strrchr(moduleFileName, '\\');
			if (lastSlash)
				*lastSlash = '\0';
		
			ret = decompressor.Initialize(moduleFileName);
			if (!ret)
			{
				strcpy(ErrorString, decompressor.GetLastErrorString());
				Error(ErrorString);

				//free the string create by strdup
				free(lowFile);

				//we're done let's free the parser
				XML_ParserFree(Parser);
				Parser = NULL;

				CloseHandle(handle);
				return 0;
			}

			char uncompressed[65536];
			char buf[65536];

			do 
			{
				DWORD len;
				if (ReadFile(handle, buf, sizeof(buf), &len, NULL) == 0)
				{
					sprintf(ErrorString, "ZGLToXGL: unable to read from file\n");
					Error(ErrorString);
					finished = true;
				}
				else
				{
					ret = decompressor.SetCompressedData(len, buf);
					assert(ret);
					
					bool needsMoreCompressedData = false;
					while (!needsMoreCompressedData && !finished)
					{
						unsigned int bytesRead;
						bool ret = decompressor.GetData(uncompressed, &bytesRead, &needsMoreCompressedData, &finished);
						if (!ret)
						{
							strcpy(ErrorString, decompressor.GetLastErrorString());
							Error(ErrorString);
							finished = true;
							break;
						}

						/*	These two lines are for compatiblity for zgl files that 
							were exported without flushing. */
						if ((len < sizeof(buf)) && needsMoreCompressedData)
							finished = true;
					
						if (bytesRead > 0)
						{
							if (!XML_Parse(Parser, uncompressed, bytesRead, finished)) 
							{
								sprintf(ErrorString, "%s: %s at line %d\n",PARSE_ERROR, XML_ErrorString(XML_GetErrorCode(Parser)),
									 XML_GetCurrentLineNumber(Parser));
								Error(ErrorString);
								break;
							}
						}
					}

					ret = decompressor.ReleaseCompressedData();
					assert(ret);



					if(State == STATE_STOPPED)
						finished = true;
				}
			}
			while (!finished);
			
			ret = decompressor.UnInitialize();
			if (!ret)
			{
				sprintf(ErrorString, "ZGL Error %s\n", decompressor.GetLastErrorString());
				Error(ErrorString);
			}
		}
		else
		{
			sprintf(ErrorString, "%s is not a valid XGL or ZGL file\n", lowFile);
			Error(ErrorString);
		}
	}

	//free the string create by strdup
	free(lowFile);

	//we're done let's free the parser
	XML_ParserFree(Parser);
	Parser = NULL;

	int success = (State == STATE_IMPORTING);
	State = STATE_STOPPED;

	if (handle != INVALID_HANDLE_VALUE)
		CloseHandle(handle);

	return success;
}

//transform manipulation functions
void
CXGLImport::StopImporting()
{
	State = STATE_STOPPED;
}

int
CXGLImport::IsStopped()
{
	return (State == STATE_STOPPED);
}

void
CXGLImport::SetTransformToIdentity(CXGLTransform *transform)
{
	transform->FX = 0.0;
	transform->FY = 0.0;
	transform->FZ = 1.0;
	transform->UX = 0.0;
	transform->UY = 1.0;
	transform->UZ = 0.0;
	transform->XP = 0.0;
	transform->YP = 0.0;
	transform->ZP = 0.0;
	transform->Scale = 1.0;
}

void
CXGLImport::InitializeExtents(CXGLExtents *extents)
{
	extents->MinX = 0.0;
	extents->MinY = 0.0;
	extents->MinZ = 0.0;
	extents->MaxX = 0.0;
	extents->MaxY = 0.0;
	extents->MaxZ = 0.0;
}


//data initialization functions
void 
CXGLImport::InitializeCurrentWorld()
{
	CurrentWorld = new CXGLWorld;
	CurrentWorld->NumObjects = 0;
	CurrentWorld->NumDatas = 0;
}

void 
CXGLImport::InitializeCurrentObject(int pathID, int xglID)
{
	CXGLObject *object = new CXGLObject;
	//create an empty object
	
	object->IsExternal = 0;
	object->MeshHandle = NULL;
	object->NumDatas = 0;
	object->NumObjects = 0;
	object->PathID = pathID;
	object->XGLID = xglID;
	object->IsStatic = 0;

	InitializeExtents(&(object->Extents));

	SetTransformToIdentity(&(object->Transform));
	
	//push the beginnings of this object onto our object stack
	ObjectStack.Push(object);

	//Initialize the requires

	ObjectStack.GetTop()->ObjectRequires.Init();
	//later we'll fill in this object more fully
}

void
CXGLImport::InitializeCurrentObjectExternal(int pathID, int isStatic)
{
	CXGLObject *object = new CXGLObject;

	object->IsExternal = 1;
	object->MeshHandle = NULL;
	object->NumDatas = 0;
	object->PathID = pathID;
	object->XGLID = -1;
	object->IsStatic = isStatic;

	InitializeExtents(&(object->Extents));

	SetTransformToIdentity(&(object->Transform));

	//push the beginnings of this object onto our object stack
	ObjectStack.Push(object);

	ObjectStack.GetTop()->ObjectRequires.Init();
}

void 
CXGLImport::InitializeCurrentData(const char *organization)
{
	//we need to allocate our organization array to be one bigger than the size of it
	//so we can store the trailing zero
	CurrentData = new CXGLData;
	CurrentData->Organization = new char[strlen(organization) + 1];
	strcpy(CurrentData->Organization, organization);
	CurrentData->NumBins = 0;
	CurrentData->NumStrings = 0;

	//initialize the resizeable arrays
	StringNames.SetSize(0);
	StringNameLengths.SetSize(0);
	Strings.SetSize(0);
	StringLengths.SetSize(0);
	BinNames.SetSize(0);
	BinNameLengths.SetSize(0);
	Bins.SetSize(0);
	BinLengths.SetSize(0);
}

void
CXGLImport::InitializeCurrentLight()
{
	CurrentLight = CurrentLighting->Lights + CurrentLighting->NumLights;

	CurrentLight->DX = 8.0f;
	CurrentLight->DY = -14.0f;
	CurrentLight->DZ = -20.0f;
	CurrentLight->DiffuseR = 1.0f;
	CurrentLight->DiffuseG = 1.0f;
	CurrentLight->DiffuseB = 1.0f;
	CurrentLight->SpecularR = 0.2f;
	CurrentLight->SpecularG = 0.2f;
	CurrentLight->SpecularB = 0.2f;
}

void 
CXGLImport::InitializeCurrentLighting()
{
	CurrentLighting = new CXGLLighting;
	CurrentLighting->AmbientR = 0.0f;
	CurrentLighting->AmbientG = 0.0f;
	CurrentLighting->AmbientB = 0.0f;
	CurrentLighting->NumLights = 0;
}

void 
CXGLImport::InitializeCurrentBackground()
{
	CurrentBackground = new CXGLBackground;
	CurrentBackground->Red = 0.0f;
	CurrentBackground->Green = 0.0f;
	CurrentBackground->Blue = 0.0f;
}

void
CXGLImport::InitializeCurrentMaterial()
{
	//Initialize to good values in case not everything is specified!!
	//don't change these values as they conform to the XGL spec
	
	CurrentMaterial.Alpha = 1.0;

	CurrentMaterial.AmbientR = 0.0;
	CurrentMaterial.AmbientG = 0.0;
	CurrentMaterial.AmbientB = 0.0;

	CurrentMaterial.DiffuseR = 0.0;
	CurrentMaterial.DiffuseG = 0.0;
	CurrentMaterial.DiffuseB = 0.0;

	CurrentMaterial.EmissiveR = 0.0;
	CurrentMaterial.EmissiveG = 0.0;
	CurrentMaterial.EmissiveB = 0.0;

	CurrentMaterial.SpecularR = 0.0;
	CurrentMaterial.SpecularG = 0.0;
	CurrentMaterial.SpecularB = 0.0;

	CurrentMaterial.Shininess = 0.0;
}

void 
CXGLImport::InitializeCurrentLineStyle()
{
	//Initialize to good values in case not everything is specified!!
	//don't change these values as they conform to the XGL spec

	CurrentLineStyle.Width = 1.0;
	CurrentLineStyle.Pattern = 0xffff;
	CurrentLineStyle.PatternFactor = 1;
}

void 
CXGLImport::InitializeCurrentPointStyle()
{
	//Initialize to good values in case not everything is specified!!
	//don't change these values as they conform to the XGL spec

	CurrentPointStyle.Size = 1.0;
}
	
void
CXGLImport::InitializeCurrentMesh()
{
	//allocate the mesh we'll eventually pass to the resolver
	CurrentMesh = new CXGLMesh;

	//initialize our resizeable arrays
	Faces.SetSize(1);
	Lines.SetSize(1);
	Points.SetSize(1);
	Vertices.SetSize(1);
	Normals.SetSize(1);
	Materials.SetSize(1);
	LineStyles.SetSize(1);
	PointStyles.SetSize(1);
	TextureCoordinates.SetSize(1);
	CurrentMesh->TextureHandles.SetSize(1);
	Patches.SetSize(0);
	
	//initialize the prev reference ids
	PrevTextureCoordinateReferenceID = -1;
	PrevTextureReferenceID = -1;
	PrevMatReferenceID = -1;
	PrevLineStyleReferenceID = -1;
	PrevPointStyleReferenceID = -1;
	PrevVertexReferenceID = -1;
	PrevNormalReferenceID = -1;

	//initialize our current patch
	CurrentPatchIndex = -1;
	CurrentPatchHigh = 0; //reserved

	CurrentFaceState = OtherFaceState;
	CurrentLineState = OtherLineState;
	CurrentPointState = OtherPointState;
}

void 
CXGLImport::InitializeCurrentTexture()
{
	//allocate the texture that we'll eventually pass to the resolver
	CurrentTexture = new CXGLTexture;

	//set some defaults in the Texture
	CurrentTexture->UseTextureBorderColor = 0;
	CurrentTexture->TextureFunction = TextureReplace;
	CurrentTexture->TextureWrap = TextureRepeat;
	CurrentTexture->TextureBufferHandle = NULL;
}

void 
CXGLImport::InitializeCurrentTextureBuffer()
{
	//set some defaults
	CurrentTextureBuffer = new CXGLTextureBuffer;
	CurrentTextureBuffer->TextureBufferFormat = TextureRGBA;
	CurrentTextureBuffer->TextureWidth = 0;
	CurrentTextureBuffer->TextureHeight = 0;
	CurrentTextureBuffer->RGBABuffer = NULL;
}

void 
CXGLImport::InitializeCurrentPatch(int patchID)
{
	//get the current index
	CurrentPatchIndex = Patches.GetNumElements();
	
	//this patch will be delete when the mesh is created
	CXGLPatch *patch = new CXGLPatch;
	patch->NumDatas = 0;
	patch->PatchID = patchID;

	Patches.AddElement(patch);
}

//uninitialization functions
void 
CXGLImport::UnInitializeCurrentMesh()
{
	CurrentMesh = NULL;
	
	//uninitialize our resizeable arrays
	Faces.SetSize(0);
	Lines.SetSize(0);
	Points.SetSize(0);
	Vertices.SetSize(0);
	Normals.SetSize(0);
	Materials.SetSize(0);
	LineStyles.SetSize(0);
	PointStyles.SetSize(0);
	TextureCoordinates.SetSize(0);
	Patches.SetSize(0);

	PrevTextureCoordinateReferenceID = -1;
	PrevTextureReferenceID = -1;
	PrevMatReferenceID = -1;
	PrevLineStyleReferenceID = -1;
	PrevPointStyleReferenceID = -1;
	PrevVertexReferenceID = -1;
	PrevNormalReferenceID = -1;

	CurrentPatchIndex = -1;
}

void 
CXGLImport::UnInitializeCurrentData()
{
	CurrentData = NULL;

	//uninitialize our resizeable arrays
	StringNames.SetSize(0);
	StringNameLengths.SetSize(0);
	Strings.SetSize(0);
	StringLengths.SetSize(0);
	BinNames.SetSize(0);
	BinNameLengths.SetSize(0);
	Bins.SetSize(0);
	BinLengths.SetSize(0);
}

void
CXGLImport::CreateCurrentMeshFromResizeableArrays()
{
	//let's figure out the size of all of the mesh arrays
	CurrentMesh->NumFaces = Faces.GetNumElements() - 1;
	CurrentMesh->NumLines = Lines.GetNumElements() - 1;
	CurrentMesh->NumPoints = Points.GetNumElements() - 1;
	CurrentMesh->NumMaterials = Materials.GetNumElements() - 1;
	CurrentMesh->NumLineStyles = LineStyles.GetNumElements() - 1;
	CurrentMesh->NumPointStyles = PointStyles.GetNumElements() - 1;
	CurrentMesh->NumNormals = Normals.GetNumElements() - 1;
	CurrentMesh->NumVertices = Vertices.GetNumElements() - 1;
	CurrentMesh->NumTextureCoordinates = TextureCoordinates.GetNumElements() - 1;
	CurrentMesh->NumTextures = CurrentMesh->TextureHandles.GetNumElements() - 1;
	CurrentMesh->NumPatches = Patches.GetNumElements();

	//allocate the new arrays and copy the data
	if (CurrentMesh->NumFaces > 0)
	{
		CurrentMesh->Faces = new CXGLFace[CurrentMesh->NumFaces];
		memcpy(CurrentMesh->Faces, Faces.GetElementRef(0), sizeof(Faces.GetElement(0)) * CurrentMesh->NumFaces);
	}

	if (CurrentMesh->NumLines > 0)
	{
		CurrentMesh->Lines = new CXGLLine[CurrentMesh->NumLines];
		memcpy(CurrentMesh->Lines, Lines.GetElementRef(0), sizeof(Lines.GetElement(0)) * CurrentMesh->NumLines);
	}

	if (CurrentMesh->NumPoints > 0)
	{
		CurrentMesh->Points = new CXGLPoint[CurrentMesh->NumPoints];
		memcpy(CurrentMesh->Points, Points.GetElementRef(0), sizeof(Points.GetElement(0)) * CurrentMesh->NumPoints);
	}

	if (CurrentMesh->NumMaterials > 0)
	{
		CurrentMesh->Materials = new CXGLMaterial[CurrentMesh->NumMaterials];
		memcpy(CurrentMesh->Materials, Materials.GetElementRef(0), sizeof(Materials.GetElement(0)) * CurrentMesh->NumMaterials);
	}

	if (CurrentMesh->NumLineStyles > 0)
	{
		CurrentMesh->LineStyles = new CXGLLineStyle[CurrentMesh->NumLineStyles];
		memcpy(CurrentMesh->LineStyles, LineStyles.GetElementRef(0), sizeof(LineStyles.GetElement(0)) * CurrentMesh->NumLineStyles);
	}

	if (CurrentMesh->NumPointStyles > 0)
	{
		CurrentMesh->PointStyles = new CXGLPointStyle[CurrentMesh->NumPointStyles];
		memcpy(CurrentMesh->PointStyles, PointStyles.GetElementRef(0), sizeof(PointStyles.GetElement(0)) * CurrentMesh->NumPointStyles);
	}

	if (CurrentMesh->NumNormals > 0)
	{
		CurrentMesh->Normals = new CXGLNormal[CurrentMesh->NumNormals];
		memcpy(CurrentMesh->Normals, Normals.GetElementRef(0), sizeof(Normals.GetElement(0)) * CurrentMesh->NumNormals);
	}

	if (CurrentMesh->NumVertices > 0)
	{
		CurrentMesh->Vertices = new CXGLVertex[CurrentMesh->NumVertices];
		memcpy(CurrentMesh->Vertices, Vertices.GetElementRef(0), sizeof(Vertices.GetElement(0)) * CurrentMesh->NumVertices);
	}

	if (CurrentMesh->NumTextureCoordinates > 0)
	{
		CurrentMesh->TextureCoordinates = new CXGLTextureCoordinate[CurrentMesh->NumTextureCoordinates];
		memcpy(CurrentMesh->TextureCoordinates, TextureCoordinates.GetElementRef(0), sizeof(TextureCoordinates.GetElement(0)) * CurrentMesh->NumTextureCoordinates);
	}


	if (CurrentMesh->NumPatches > 0)
		CurrentMesh->Patches = new CXGLPatch[CurrentMesh->NumPatches];

	//copy the patches into the mesh and delete the patches we allocated
	unsigned int i, d;
	for (i = 0; i < CurrentMesh->NumPatches; i++)
	{
		//copy over the patch
		CXGLPatch *patch = Patches.GetElement(i);
		CurrentMesh->Patches[i].NumDatas = patch->NumDatas;
		CurrentMesh->Patches[i].PatchID = patch->PatchID;

		
		//set the size of the datahandles array
		CurrentMesh->Patches[i].DataHandles.SetSize(patch->NumDatas);

		//copy the datahandles array
		for (d = 0; d < CurrentMesh->Patches[i].NumDatas; d++)
		{
			CurrentMesh->Patches[i].DataHandles.SetElement(d, patch->DataHandles.GetElement(d));
		}
	}

	for (i = 0; i < CurrentMesh->NumPatches; i++)
	{
		CXGLPatch *patch = Patches.GetElement(i);
		delete patch;
	}
	Patches.SetSize(0);
}

void
CXGLImport::CreateCurrentDataFromResizeableArrays()
{
	//let's figure out the size of all of the data arrays
	CurrentData->NumStrings = Strings.GetNumElements();
	CurrentData->NumBins = Bins.GetNumElements();

	if(CurrentData->NumStrings > 0)
	{
		//allocate the new arrays
		CurrentData->StringNames = new char*[CurrentData->NumStrings];
		CurrentData->Strings = new char*[CurrentData->NumStrings];
	}

	if(CurrentData->NumBins > 0)
	{
		CurrentData->BinNames = new char*[CurrentData->NumBins];
		CurrentData->Bins = new unsigned char*[CurrentData->NumBins];
		CurrentData->BinLengths = new unsigned int[CurrentData->NumBins];
		//copy the information into this data

		//copy the lengths first
		memcpy(CurrentData->BinLengths, BinLengths.GetElementRef(0), sizeof(int)*CurrentData->NumBins);
	}


	unsigned int i;
	for (i = 0; i < CurrentData->NumStrings; i++)
	{
		/***************
			String name
		****************/
		//get the length of the stringname
		int stringNameLength = StringNameLengths.GetElement(i);

		//allocate the new stringname
		CurrentData->StringNames[i] = new char[stringNameLength];

		//copy the stringname from the resizeable array
		memcpy(CurrentData->StringNames[i], StringNames.GetElement(i), sizeof(char)*stringNameLength);
		
		//zero terminate the stringname
		char *stringName = CurrentData->StringNames[i];
		stringName[stringNameLength - 1] = 0;

		//delete the stringname stored in the resizeable array
		char *toDelete = StringNames.GetElement(i);
		delete[] toDelete;


		/***************
			String
		****************/
		//get the length of the string
		int stringLength = StringLengths.GetElement(i);

		//allocate the new string
		CurrentData->Strings[i] = new char[stringLength];

		//copy the string from the resizeable array
		memcpy(CurrentData->Strings[i], Strings.GetElement(i), sizeof(char)*stringLength);
		
		//zero terminate the string
		char *string = CurrentData->Strings[i];
		string[stringLength - 1] = 0;

		//delete the string stored in the resizeable array
		toDelete = Strings.GetElement(i);
		delete[] toDelete;
	}

	for (i = 0; i < CurrentData->NumBins; i++)
	{
		/***************
			Bin name
		****************/
		//get the length of the binname
		int binNameLength = BinNameLengths.GetElement(i);

		//allocate the new binname
		CurrentData->BinNames[i] = new char[binNameLength];

		//copy the binname from the resizeable array
		memcpy(CurrentData->BinNames[i], BinNames.GetElement(i), sizeof(char)*binNameLength);
		
		//zero terminate the binname
		char *binName = CurrentData->BinNames[i];
		binName[binNameLength - 1] = 0;

		//delete the binname stored in the resizeable array
		char *toDelete = BinNames.GetElement(i);
		delete[] toDelete;


		/***************
			Bin
		****************/
		//get the length of the bin
		int binLength = BinLengths.GetElement(i);

		//allocate the new bin
		CurrentData->Bins[i] = new unsigned char[binLength];

		//copy the bin from the resizeable array
		memcpy(CurrentData->Bins[i], Bins.GetElement(i), sizeof(unsigned char)*binLength);

		//delete the bin stored in the resizeable array
		unsigned char *unsignedToDelete = Bins.GetElement(i);
		delete[] unsignedToDelete;
	}
}

int
CXGLImport::VerifyCurrentWorldResolved()
{
	char buffer[250];
	unsigned int i;
	for (i = (CurrentWorld->NumObjects - 1); i < 0xffffffff; i--)
	{
		if (VerifyObjectResolved(CurrentWorld->ObjectHandles.GetElement(i)) == 0)
		{			
			sprintf(buffer, "%s : %u", UNRESOLVED_OBJECT_ERROR, i);
			Error(buffer);
			return 0;
		}
	}

	return 1;
}

int
CXGLImport::VerifyObjectResolved(CXGLObjectHandle *handle)
{
	if ((handle->CallbackHandle != -1) &&
		(handle->Object == NULL))
	{
		//this object is resolved
	}
	else
	{
		//this object is unresolved
		return 0;
	}

	return 1;
}

void
CXGLImport::InitializeCurrentExtgeom(const char *organization)
{
	//we need to allocate our organization array to be one bigger than the size of it
	//so we can store the trailing zero
	CurrentExtgeom = new CXGLExtgeom;
	CurrentExtgeom->Organization = new char[strlen(organization) + 1];
	strcpy(CurrentExtgeom->Organization, organization);
	strlwr(CurrentExtgeom->Organization);
	//initialize the resizeable arrays
	CurrentExtgeom->Extgeomelements.SetSize(0);
}

void
CXGLImport::ExtgeomStateBeginTagHandler(int tag, const char** atts)
{
	switch(tag)
	{
	case ExtGeomElemTag:
		{
			char **currentAtt = (char **)atts;
			char *typestring = NULL;

			if (*currentAtt == NULL)
			{
				Error(EXTGEOMELEM_NO_TYPE);
			}

			while (*currentAtt != NULL)
			{
				if (stricmp(*currentAtt, EXTGEOMELEMTYPE) != 0)
				{
					Error(EXTGEOMELEM_ATTR_NOT_CALLED_TYPE);
				}
				else
				{
					CurrentExtgeomState = ExtgeomelemState;
					
					//get the length of the type
					unsigned int length = strlen(*(currentAtt + 1)) + 1;

					//get the string name and set it in the resizeable array
					if(typestring != NULL) delete typestring;
					typestring = new char[length];
					strcpy(typestring, *(currentAtt + 1));
					strlwr(typestring);
				}

				currentAtt += 2;
			}

			// initiation of current Extgeomelem:
			CurrentExtgeomelem = new CXGLExtgeomelem;
			CurrentExtgeomelem->Type = typestring;
				
			break;
		}
	case ExtGeomElemDataTag:
		{
			CurrentExtgeomState = ExtgeomelemdataState;
			NeedsData = 1;
			break;
		}
	case ExtGeomElemExtentsTag:
		{
			CurrentExtgeomState = ExtgeomelemextentsState;
			NeedsData = 1;
			break;
		}
	case ExtGeomElemPriorityTag:
		{
			CurrentExtgeomState = ExtgeomelempriorityState;
			NeedsData = 1;
			break;
		}
	case OtherTag:
		{
			StateStack.Push(CurrentState);
			CurrentState = OtherState;
			OtherStateLevel = 1;
			CurrentExtgeomState = OtherExtgeomState;
			break;
		}
	default:
		{
			Error(IMPROPER_TAG_ERROR);
			break;
		}
	}
}

void 
CXGLImport::ExtgeomStateCharDataHandler(const char *s)
{
	switch (CurrentExtgeomState)
	{
	case ExtgeomelemdataState:
		{
			if(CurrentExtgeomelem->Extgeomelemdata != NULL)
			{ 
				Error(EXTGEOMELEMDATA_MORE_THAN_ONE_DATA_TAG);
				break;
			}
			//get the length of the string (really binary data)
			char *smod;
			unsigned int length;
			
			length = strlen(s);
			smod = new char [length+1];
			strcpy(smod,s);
			CleanStringFromSpacesAndNL(smod); // removing spaces and NL
			
			length = strlen(smod);
			if((length % 2) != 0)
			{
				delete [] smod;
				Error(EXTGEOMELEMDATA_INVALID_LENGTH);
				break;
			}
			int dataLength = length/2;
			unsigned char *b = new unsigned char[dataLength];
			int success = GetBinaryDataFromHex(length, smod, b);
			delete [] smod;
			if(!success)
			{
				delete [] b;
				Error(EXTGEOMELEMDATA_INVALID_BIN_DATA);
				break;
			}
			CurrentExtgeomelem->LengthExtgeomelemdata = dataLength;
			CurrentExtgeomelem->Extgeomelemdata = b;
			break;
		}

	case ExtgeomelemextentsState:
		{
			if(CurrentExtgeomelem->Extgeomelemextents != NULL)
			{ 
				Error(EXTGEOMELEMEXTENTS_MORE_THAN_ONE_EXTENTS_TAG);
				break;
			}
            CXGLExtents *extents = new CXGLExtents;
            if(!Get6DoubleVector(s,&(extents->MinX),&(extents->MinY),&(extents->MinZ),
								   &(extents->MaxX),&(extents->MaxY),&(extents->MaxZ)))
			{
				delete extents;
				Error(EXTGEOMELEMEXTENTS_INVALID_VECTOR);
				break;
			}
			CurrentExtgeomelem->Extgeomelemextents = extents;
			break;
		}

	case ExtgeomelempriorityState:
		{
			if(CurrentExtgeomelem->Extgeomelempriority != -1.)
			{ 
				Error(EXTGEOMELEMPRIORITY_MORE_THAN_ONE_PRIORITY_TAG);
				break;
			}
			float priority;
			if(!GetFloat(s,&priority))
			{
				Error(EXTGEOMELEMPRIORITY_INVALID_PRIORITY_TAG);
				break;
			}
			if(priority < 0. || priority > 1.)
			{
				Error(EXTGEOMELEMPRIORITY_VALUE_OUTSIDE_0_1);
				break;
			}
			CurrentExtgeomelem->Extgeomelempriority = priority;
			break;
		}
	}
}

void 
CXGLImport::ExtgeomStateEndTagHandler(int tag)
{
	if(NeedsData)
	{
		Error(EXTGEOM_NEEDS_DATA);
		return;
	}
	
	switch(tag)
	{
	case ExtGeomElemDataTag:
		{
			//do nothing
			break;
		}
	case ExtGeomElemExtentsTag:
		{
			//do nothing
			break;
		}
	case ExtGeomElemPriorityTag:
		{
			//do nothing
			break;
		}
	case ExtGeomElemTag:
		{
			if(CurrentExtgeomelem->Extgeomelemdata == NULL || 
			   CurrentExtgeomelem->Extgeomelemextents == NULL)
			{
				Error(EXTGEOMELEM_NO_DATA_OR_EXTENTS);
				return;
			}
			if(CurrentExtgeomelem->Extgeomelempriority == -1.)
				CurrentExtgeomelem->Extgeomelempriority = 1.;  // default value is 1
			// attaching current extended element to ext geom class:
			CurrentExtgeom->Extgeomelements.AddElement(CurrentExtgeomelem);
			break;
		}

	case ExtGeomTag:
		{
			int numExtgeomelem = CurrentExtgeom->Extgeomelements.GetNumElements();
			if(numExtgeomelem == 0)
			{
				Error(EXTGEOM_NO_EXTGEOMELEM_TAG);
				break;
			}
			//pop up to our parent state
			CurrentState = StateStack.Pop();
			if(CurrentState == MeshState)
			{
				if(CurrentMesh != 0)
				{
					int numExtGeom = CurrentMesh->NumExtgeoms;
					CurrentMesh->ExtgeomHandles.SetSize(numExtGeom+1);
					CurrentMesh->ExtgeomHandles.SetElement(numExtGeom,Resolver.AddExtgeom(CurrentContext, CurrentExtgeom));
					CurrentMesh->NumExtgeoms++;
				} 

				CurrentExtgeom = NULL;
				CurrentExtgeomelem = NULL;
				CurrentMeshState = OtherMeshState;
			}
			else
			{	// This must be object
				//let's send off the data we've accumulated to the object
				CObjectStackElem *top = ObjectStack.GetTop();
				//set the datahandles array to be one greater
				top->Object->ExtgeomHandles.SetSize(top->Object->NumExtgeoms + 1);
				//resolve the data and set the returned handle in the object's handle array
				top->Object->ExtgeomHandles.SetElement(top->Object->NumExtgeoms, Resolver.AddExtgeom(CurrentContext, CurrentExtgeom));
				//increment the number of Extgeoms in the object
				top->Object->NumExtgeoms++;
				CurrentExtgeom = NULL;
				CurrentExtgeomelem = NULL;
			}
			break;
		}
	}

	CurrentExtgeomState = OtherExtgeomState;
}

void CleanStringFromSpacesAndNL(char *s)
{
	unsigned int ind1 = 0, ind2 = 0;
	char c;
	c = *(s+ind1);
	while (c != 0)
	{ 
		if(!(c == ' ' || c == '\n'))
		{
			if(ind1 != ind2) *(s+ind2) = c;
		    ind2++;
		}
		ind1++; c = *(s+ind1);
	}
	*(s+ind2)=0;
}
